/**********************************************************************/
/*      Licensed Materials - Property of IBM                          */
/*      "Restricted Materials of IBM"                                 */
/*      5748-T15                                                      */
/*      Copyright IBM Corp. 2016, 2021                                */
/*      All Rights Reserved                                           */
/*      US Government Users Restricted Rights - Use, duplication      */
/*      or dislosure restricted by GSA ADP Schedule Contract          */
/*      with IBM Corp.                                                */
/**********************************************************************/
/* This file is provided on an "as is" basis.  There are no           */
/* warranties, express or implied, including the implied warranties   */
/* of merchantability and fitness for a particular purpose.           */
/* IBM has no obligation to provide service, defect correction, or    */
/* any maintenance for the file.  IBM has no obligation to supply     */
/* any updates or enhancements for the file to you even if such are   */
/* or later become available.                                         */
/**********************************************************************/
/* FUNCTION NAME.....QFNL                                             */
/* PURPOSE...........Generate FIND/FILE activity                      */
/* FUNCTION TYPE.....TPF                                              */
/* DOCUMENT NAME.....NA                                               */
/*                                                                    */
/* INPUT.......ZTEST i FFACT parameters                               */
/*             OR ZTEST FFACT HELP or ? to display the correct        */
/*                syntax of the command.                              */
/*  NOTE: In addition to the parameters documented in the Driver's    */
/*    user guide, I include the following in case we want to add      */
/*    "canned" cases in the future:                                   */
/*   ZTEST i FFACT CASE-nn|ALL to run "canned" case by number         */
/*           or ALL                                                   */
/*   ZTEST i FFACT START to loop running "canned" cases               */
/*   ZTEST i FFACT STOP  to stop running "canned" cases               */
/*   ZTEST i FFACT STATUS status of "canned" case run                 */
/*                                                                    */
/* Options:                                                           */
/* i : I-stream number (optional parameter)                           */
/*                                                                    */
/* OUTPUT......Some programs are self checking the passing parameters */
/*                                                                    */
/* DESCRIPTION:                                                       */
/*                                                                    */
/* QFNL is the "main" segment for the Logical Record Cache Driver,    */
/*      ZTEST FFACT or ZTEST FFAC                                     */
/* It determines which test case to run based on the number entered   */
/* by the user.                                                       */
/*                                                                    */
/*                                                                    */
/**********************************************************************/
/* Change History                                                     */
/*   Defect           Date               Description                  */
/*   FFACT           02/01/16            Created                      */
/*   dynami3         05/26/17            Enhanced for Dynamic CPU     */
/*                                        Utlization project (PJ44591)*/
/*   d26352          20180810            Compiler compatibility updates  */
/*   d26441          20190110            CTL-3 on ZTEST FFACT FIXED   */
/*                                       KILL GROUP-1                 */
/*   d26488          20190516            Format of output line for    */
/*                                         ZTEST FFACT FIXED STATUS   */
/*                                         ZTEST FFACT POOL STATUS    */
/*   d26633          20200916            Add support for multiple     */
/*                                         DECBs for FIXED BUILD      */
/*   d26696          20210401            Terminology replacement      */
/*                                                                    */
/**********************************************************************/

#include <tpf/tpfparse.h>
#include <qfnl.h>                                                               //d26352
#include <tpf/tpfapi.h>
#include <fcntl.h>
#include <fstream>
using namespace std;

int qfnl_modify_fixed_record(void * record_ptr,
                             TPF_FA8 file_address,
                             qfnl_fixed_info * group_info_ptr);
int qfnl_modify_pool_record(void * record_ptr,
                             TPF_FA8 file_address,
                             qfnl_pool_info * group_info_ptr,
                             tpf_TOD_ext_type * caller_tod_ptr);
boolean qfnl_is_recordOK(void * record_ptr,
                         TPF_FA8 file_address,
                         qfnl_fixed_info * group_info_ptr);
static void qfnl_fixed_build(struct IPRSE_output *parse,
                             int num_parms);
int qfnl_parse_fixed_build(struct IPRSE_output *parse,
                           int num_parms);
static void qfnl_fixed_start(struct IPRSE_output *parse,
                             int num_parms);
int qfnl_parse_fixed_start(struct IPRSE_output *parse,
                             int num_parms);
boolean qfnl_first_record_ok(qfnl_pool_info_Ptr group_info_ptr, char *rec_ID_ptr);
boolean qfnl_first_record_ok(qfnl_fixed_info_Ptr group_info_ptr, char *rec_ID_ptr);
void qfnl_fixed_child_ecb (qfnl_child_ecb_parm_list_Ptr incoming_ecb_parm_ptr);
boolean qfnl_get_fixed_fileaddr(qfnl_fixed_info_Ptr group_info_ptr,
                                unsigned int current_ordinal,
                                TPF_DECB *decb,
                                TPF_FA8 *fileaddr_ptr,
                                char * FARFstring,
                                int FARFstring_size,
                                boolean doDump);
boolean qfnl_get_pool_fileaddr(qfnl_pool_info_Ptr group_info_ptr,
                                unsigned int index,
                                TPF_DECB *decb,
                                TPF_FA8 *fileaddr_ptr,
                                char * FARFstring,
                                int FARFstring_size,
                                boolean doDump);
void qfnl_pool_set_expired(qfnl_pool_info_Ptr group_info_ptr);
void qfnl_pool_child_ecb (qfnl_child_ecb_parm_list_Ptr incoming_ecb_parm_ptr);

static void qfnl_fixed_status(struct IPRSE_output *parse,
                             int num_parms);
int qfnl_parse_fixed_status(struct IPRSE_output *parse,
                             int num_parms);
void qfnl_check_fixed_status(char * statusString, qfnl_fixed_info_Ptr group_info_ptr);
void qfnl_fixed_general_status(void);
static void qfnl_fixed_stop(struct IPRSE_output *parse,
                             int num_parms);
int qfnl_parse_fixed_stop(struct IPRSE_output *parse,
                          int num_parms);
void qfnl_fixed_stop_all_groups(void);
static void qfnl_fixed_kill(struct IPRSE_output *parse,
                             int num_parms);
int qfnl_parse_fixed_kill(struct IPRSE_output *parse,
                             int num_parms);
static void qfnl_fixed_clear(struct IPRSE_output *parse,
                             int num_parms);
int qfnl_parse_fixed_clear(struct IPRSE_output *parse,
                             int num_parms);
static void qfnl_pool_build(struct IPRSE_output *parse,
                             int num_parms);
int qfnl_parse_pool_build(struct IPRSE_output *parse,
                           int num_parms);
static void qfnl_pool_start(struct IPRSE_output *parse,
                             int num_parms);
int qfnl_parse_pool_start(struct IPRSE_output *parse,
                            int num_parms);
static void qfnl_pool_status(struct IPRSE_output *parse,
                             int num_parms);
int qfnl_parse_pool_status(struct IPRSE_output *parse,
                            int num_parms);
void qfnl_check_pool_status(char * statusString, qfnl_pool_info_Ptr group_info_ptr);
void qfnl_pool_general_status(void);
static void qfnl_pool_stop(struct IPRSE_output *parse,
                             int num_parms);
int qfnl_parse_pool_stop(struct IPRSE_output *parse,
                          int num_parms);
void qfnl_pool_stop_all_groups(void);
static void qfnl_pool_kill(struct IPRSE_output *parse,
                             int num_parms);
int qfnl_parse_pool_kill(struct IPRSE_output *parse,
                            int num_parms);
static void qfnl_pool_clear(struct IPRSE_output *parse,
                             int num_parms);
int qfnl_parse_pool_clear(struct IPRSE_output *parse,
                             int num_parms);
static void qfnl_pool_list(struct IPRSE_output *parse,
                             int num_parms);
int qfnl_parse_pool_list(struct IPRSE_output *parse,
                            int num_parms);

static void qfnl_help(struct IPRSE_output *parse,
                        int num_parms);

qfnl_fixed_info_Ptr qfnl_locate_fixed_group (int group,
                                             boolean beNoisy,
                                             boolean lookInFileSystem);

qfnl_pool_info_Ptr qfnl_locate_pool_group (int group,
                                           boolean beNoisy);

qfnl_fixed_ctl_tbl_Ptr get_fixed_table();
qfnl_pool_ctl_tbl_Ptr get_pool_table();
int qfnl_random (int min, int max);
unsigned int convert_input_data_to_hex(char *, char *);

boolean qfnl_check_if_groupAlreadyBuilt(char * file_name,
                                        qfnl_fixed_build_parm_list_Ptr fixed_build_ptr,
                                        boolean * is_error_ptr);

int qfnl_FARF6_to_hex_string(char * FARFstring,
                             int input_size,
                             TPF_FA8 file_address);

int convertBinToHexStr(char *dst, int *dstlen,
                        const char *src, int srclen);

/* External functions */
// TODO: Add your own here or modify 'runApi'.
//       You'll have to tinker with the canned mode and/or
//       parsing sections.
int runApi(int casenum);

/* Canned case mode functions */
static void cannedStart(int force);
static void cannedStop(int force);
static void cannedStatus();
static void cannedRun();

/* Global info flags */
int verbose = 0;
int debug = 0;
int warnings = 1;
int repeat_canned = 0;

int one = 1;

/* Variables used throughout */

/*-------------------------------------------------------------------------------------------------------------*/
// Used to keep track of parameters for the various commands:

qfnl_fixed_build_parm_list my_fixed_build;
qfnl_fixed_build_parm_list already_fixed_build;  // used to read parameters for something already built
int                        decb_build_count;     // Not written to file, runtime variable only //d26633

qfnl_fixed_start_parm_list my_fixed_start;

qfnl_fixed_status_parm_list my_fixed_status;

qfnl_fixed_stop_parm_list my_fixed_stop;

qfnl_fixed_kill_parm_list my_fixed_kill;

qfnl_fixed_clear_parm_list my_fixed_clear;


qfnl_pool_build_parm_list my_pool_build;

qfnl_pool_start_parm_list my_pool_start;

qfnl_pool_stop_parm_list my_pool_stop;

qfnl_pool_status_parm_list my_pool_status;

qfnl_pool_list_parm_list my_pool_list;

qfnl_pool_kill_parm_list my_pool_kill;

qfnl_pool_clear_parm_list my_pool_clear;


/*-------------------------------------------------------------------------------------------------------------*/

char                      stpoc_value[10];
FILE * FIXED_GROUP_FILE;
char FIXED_file_name[2+6+sizeof(QFNL_DIR)+sizeof(QFNL_PFX)]; // used to remember fixed file groups
FILE * FIXED_GROUP_LOCK;
char FIXED_lock_name[2+6+sizeof(QFNL_DIR)+sizeof(QFNL_LCK)]; // used to prevent simultaneous builds of the same group

#define GFS_INACTIVE     0x01

extern "C" int QFNL(void)
{
 // if(levtest(D0)) {  // If entered by functional message
 if ( (levtest(D0)) &&
      (strncmp((char *)&ecbptr()->ebw000, "EXPR", 4) != 0) ) { // If entered by functional message

  // char *sys_state = (char *) cinfc_fast(CINFC_CMMSTI);
  // struct mi0mi *core_block;
  char *input_text;
  struct IPRSE_output parse_result;
  struct IPRSE_output *parse;
  int num_parms;

  int input_casenumber, internal_casenumber;     /* used for test case number */

  /************************************************************************************************/
  /* Grammar                                                                                      */
  /* Using IPRSE to check the syntax of the input function message:                               */
  /*   ZTEST FFAC fixed Build GROUP-1 TYPE-#WAARI RID-C1C2 BEGIN-0 END-2710                       */
  /*   ZTEST FFACT FIXED BUILD GROUP-2 TYPE-#WAARI RID-C1C9 BEGIN-2711 END-4E20                   */
  /*   ztest ffact FIXED START G-1 SEED-33 R-3 W-2 ECBS-10                                        */
  /*   ztest ffact FIXED START G-1 SEED-33 R-3 W-2 ECBS-10 LOCK-N                                 */
  /*   ZTEST FFACT FIXED STAT G-3                                                                 */
  /*   ZTEST FFACT FIXED STAT                                                                     */
  /*   ZTEST FFACT FIXED STOP G-1                                                                 */
  /*   ZTEST FFACT FIXED KILL G-1                                                                 */
  /*   ZTEST FFACT FIXED CLEAR G-3                                                                */
  /*                                                                                              */
  /*   ZTEST FFACT  < CASE-xx|CASE-ALL|START|STOP|STATus >                                        */
  /*                           xx is the test case number                                         */
  /************************************************************************************************/
  // char *grammar = IPRSE_STRICT_GRAMMAR IPRSE_MIXED_CASE_GRAMMAR
  // char *grammar = IPRSE_MIXED_CASE_GRAMMAR
  // char *grammar = "<IPRSE_STRICT><IPRSE_MIXED_CASE>"
  char *grammar =
    "{ FFACt Fixed BUild"
    " Group-d++++"
    " Type-n++++++++<"
    " RID-xxxx"
    " Begin-x+++++++++"
    " End-x+++++++++"
    " [Decb-d++]"                                                               //d26633
    "| FFACt Fixed STARt"
    " Group-d++++"
    " SEED-d++"
    " Read-d++"
    " Write-d++"
    " ECBs-d++"
    " [Lock-a]"
    " [LOw-a++++++]"
    " [SWitch-a++++++]"
    " [UTILclass-a+++++++<]"
    " [RESourceclass-a]"
    "| FFACt Fixed STATus"
    " [Group-d++++]"
    "| FFACt Fixed STOP"
    " [Group-d++++]"
    "| FFACt Fixed KILL"
    " Group-d++++"
    "| FFACt Fixed CLEAR"
    " Group-d++++"
    "| FFACt Pool BUild"
    " Group-d++++"
    " Num-d+++++"
    " STPool-aaaa"
    " RID-xxxx"
    " Expire-d+++"
    "| FFACt Pool STARt"
    " Group-d++++"
    " SEED-d++"
    " Read-d++"
    " Write-d++"
    " ECBs-d++"
    " [Lock-a]"
    " [LOw-a++++++]"
    " [SWitch-a++++++]"
    " [UTILclass-a+++++++<]"
    " [RESourceclass-a]"
    "| FFACt Pool STATus"
    " [Group-d++++]"
    "| FFACt Pool LIST"
    " Group-d++++"
    " Num-d+++++"
    "| FFACt Pool STOP"
    " [Group-d++++]"
    "| FFACt Pool KILL"
    " Group-d++++"
    "| FFACt Pool CLEAR"
    " Group-d++++"
    "| FFACt START [Force]"
    "| FFACt STOP [Force]"
    "| FFACt STATus"
    "| FFACt CASE-d++"
    "| FFACt CASE-ALL"
    "| FFACt HELP"
    "| FFACt HELP Fixed"
    "| FFACt HELP Fixed BUild"
    "| FFACt HELP Fixed STARt"
    "| FFACt HELP Fixed STATus"
    "| FFACt HELP Fixed STOP"
    "| FFACt HELP Fixed CLEAR"
    "| FFACt HELP Fixed KILL"
    "| FFACt HELP Pool"
    "| FFACt HELP Pool BUild"
    "| FFACt HELP Pool STARt"
    "| FFACt HELP Pool STATus"
    "| FFACt HELP Pool LIST"
    "| FFACt HELP Pool STOP"
    "| FFACt HELP Pool CLEAR"
    "| FFACt HELP Pool KILL"
    "| FFACt HELP CASE"
    "| FFACt HELP START"
    "| FFACt HELP STOP"
    "| FFACt HELP STATus"
    "}";

  // TODO: modify the parse grammar

  /********************************************************************/
  /* System is required to be in NORM state.                          */
  /********************************************************************/
  // TODO: Uncoment out the following code if and only if
  //    you want driver to require NORM state
  // if (*sys_state != _CCNORM)
  // {
  //  printErr(104, "%s DRIVER MUST BE RUN IN NORM STATE", DRVR_HDR);
  //  return 0;
  // }

  /********************************************************************/
  /* Check for launch of repeated CANNED cases...                     */
  /********************************************************************/
  if (strncmp( (char *)&ecbptr()->ebw000, "canned", 6) == 0)
  {
    cannedRun();
    return 0;
  }

  /********************************************************************/
  /* Parse Input                                                      */
  /********************************************************************/
  // core_block = ecbptr()->ce1cr0;
  // *((char *) &core_block->mi0ln0 + core_block->mi0cct - 1) = '\0';
  // input_text = strpbrk(core_block->mi0acc, DRVR_NAME) + strlen(DRVR_NAME);
  input_text = ecbp2()->ce2crsmsg;
  input_text = input_text + strlen("ZTEST");

  // num_parms = IPRSE_parse(input_text, grammar, &parse_result,
  //                        IPRSE_STRICT+IPRSE_MIXED_CASE+
  //                        IPRSE_ALLOC+IPRSE_NOPRINT+IPRSE_EOM, NULL);
  num_parms = IPRSE_parse(input_text, grammar, &parse_result,
                          IPRSE_NOEOM + IPRSE_PRINT + IPRSE_ALLOC + IPRSE_MIXED_CASE + IPRSE_QUOTE,
                          "QFNL");
  parse = &parse_result;

  /********************************************************************/
  /* Check for invalid parsing                                        */
  /********************************************************************/
  if (num_parms == 0)
  {
    printErr(100, "UNKNOWN SYNTAX\n"
                  "Enter ZTEST %s HELP to display syntax", DRVR_NAME);
    exit(0);
  }

  if (strcmp(parse->IPRSE_parameter, "FFACt") == 0) {
    parse = parse->IPRSE_next;
    num_parms--;
  } else {
    printErr(111, "UNKNOWN SYNTAX\n"
                  "Enter ZTEST %s HELP to display syntax", DRVR_NAME);
    exit(0);
  }

  /********************************************************************/
  /* Check for help command                                           */
  /********************************************************************/
  if (strcmp(parse->IPRSE_parameter, "HELP") == 0)
  {
     // insert help code here
  }


  /********************************************************************/
  /* Handling of CANNED cases...                                      */
  /********************************************************************/
  if (strcmp(parse->IPRSE_parameter, "START") == 0)
  {
    parse = parse->IPRSE_next;
    if (parse && strcmp(parse->IPRSE_parameter, "Force") == 0)
    {
      cannedStart(CANNED_FORCE);
    }
    else
    {
      cannedStart(CANNED_DEFAULT);
    }
  }
  else if (strcmp(parse->IPRSE_parameter, "STOP") == 0)
  {
    parse = parse->IPRSE_next;
    if (parse && strcmp(parse->IPRSE_parameter, "Force") == 0)
    {
      cannedStop(CANNED_FORCE);
    }
    else
    {
      cannedStop(CANNED_DEFAULT);
    }
  }
  else if (strcmp(parse->IPRSE_parameter, "STATus") == 0)
  {
    cannedStatus();
  }

  /********************************************************************/
  /* Processing for ZTEST FFACT Fixed or Pool commands...             */
  /********************************************************************/

  else if (strcmp(parse->IPRSE_parameter, "Fixed") == 0)
  {  // begin If ZTEST FFACT Fixed

   /*******************************************************************/
   /*  ...Process ZTEST FFACT Fixed commands...                       */
   /*******************************************************************/

    if (debug) printf("\n !!! Hooray!! We made it to parse ZTEST FFACT FIXED command !!! \n");

    /******************************************************************/
    /* Examine all FIXED parameters                                   */
    /******************************************************************/

    parse = parse->IPRSE_next;
    num_parms--;
    // printf("\n !!!! DEBUG  parse->IPRSE_parameter is %s !!!\n", parse->IPRSE_parameter);

    if (strcmp(parse->IPRSE_parameter, "BUild") == 0) {

     /*******************************************************************/
     /*  ...Process ZTEST FFACT FIXED BUILD command..                   */
     /*******************************************************************/

     qfnl_fixed_build(parse,num_parms);

    } else if (strcmp(parse->IPRSE_parameter, "STARt") == 0) {

     /*******************************************************************/
     /*  ...Process ZTEST FFACT FIXED START command..                   */
     /*******************************************************************/

     qfnl_fixed_start(parse,num_parms);

    } else if (strcmp(parse->IPRSE_parameter, "STOP") == 0) {

     /*******************************************************************/
     /*  ...Process ZTEST FFACT FIXED STOP command..                    */
     /*******************************************************************/

     qfnl_fixed_stop(parse,num_parms);

    } else if (strcmp(parse->IPRSE_parameter, "STATus") == 0) {

     /*******************************************************************/
     /*  ...Process ZTEST FFACT FIXED STATUS command..                  */
     /*******************************************************************/

     qfnl_fixed_status(parse,num_parms);

    } else if (strcmp(parse->IPRSE_parameter, "KILL") == 0) {

     /*******************************************************************/
     /*  ...Process ZTEST FFACT FIXED KILL command..                    */
     /*******************************************************************/

     qfnl_fixed_kill(parse,num_parms);

    } else if (strcmp(parse->IPRSE_parameter, "CLEAR") == 0) {

     /*******************************************************************/
     /*  ...Process ZTEST FFACT FIXED CLEAR command..                   */
     /*******************************************************************/

     qfnl_fixed_clear(parse,num_parms);

    }

  }  // end If ZTEST FFACT Fixed

  else if (strcmp(parse->IPRSE_parameter, "Pool") == 0)
  {  // begin If ZTEST FFACT Pool

   /*******************************************************************/
   /*  ...Process ZTEST FFACT Pool commands...                        */
   /*******************************************************************/

    if (debug) printf("\n !!! Hooray!! We made it to parse ZTEST FFACT POOL command !!! \n");

    /******************************************************************/
    /* Examine all POOL parameters                                    */
    /******************************************************************/

    parse = parse->IPRSE_next;
    num_parms--;
    // printf("\n !!!! DEBUG  parse->IPRSE_parameter is %s !!!\n", parse->IPRSE_parameter);

    if (strcmp(parse->IPRSE_parameter, "BUild") == 0) {

     /*******************************************************************/
     /*  ...Process ZTEST FFACT POOL BUILD command..                    */
     /*******************************************************************/

     qfnl_pool_build(parse,num_parms);

    } else if (strcmp(parse->IPRSE_parameter, "STARt") == 0) {

     /*******************************************************************/
     /*  ...Process ZTEST FFACT POOL START command..                    */
     /*******************************************************************/

     qfnl_pool_start(parse,num_parms);

    } else if (strcmp(parse->IPRSE_parameter, "STOP") == 0) {

     /*******************************************************************/
     /*  ...Process ZTEST FFACT POOL STOP command..                     */
     /*******************************************************************/

     qfnl_pool_stop(parse,num_parms);

    } else if (strcmp(parse->IPRSE_parameter, "STATus") == 0) {

     /*******************************************************************/
     /*  ...Process ZTEST FFACT POOL STATUS command..                   */
     /*******************************************************************/

     qfnl_pool_status(parse,num_parms);

    } else if (strcmp(parse->IPRSE_parameter, "LIST") == 0) {

     /*******************************************************************/
     /*  ...Process ZTEST FFACT POOL LIST command..                     */
     /*******************************************************************/

     qfnl_pool_list(parse,num_parms);

    } else if (strcmp(parse->IPRSE_parameter, "KILL") == 0) {

     /*******************************************************************/
     /*  ...Process ZTEST FFACT POOL KILL command..                     */
     /*******************************************************************/

     qfnl_pool_kill(parse,num_parms);

    } else if (strcmp(parse->IPRSE_parameter, "CLEAR") == 0) {

     /*******************************************************************/
     /*  ...Process ZTEST FFACT POOL CLEAR command..                    */
     /*******************************************************************/

     qfnl_pool_clear(parse,num_parms);

    }

  }  // end If ZTEST FFACT Pool

  else if (strcmp(parse->IPRSE_parameter, "HELP") == 0)
  { // if ZTEST FFACT HELP

     /*******************************************************************/
     /*  ...Process ZTEST FFACT HELP commands..                         */
     /*******************************************************************/

     qfnl_help(parse,num_parms);

  } // end if ZTEST FFACT HELP

  /********************************************************************/
  /* Specific Test Case Logic                                         */
  /********************************************************************/
  // TODO: place your logic here
  else if (strcmp(parse->IPRSE_parameter, "CASE-d++") == 0)
  {
    // TODO: put your logic here
    //       for example, this calls an "external" function that is
    //       defined in 'QFNL_test.c'. In the sample, this will execute
    //       the test case through runApi().
    input_casenumber = atoi(parse->IPRSE_value);
    if ((input_casenumber < QFNL_TC_BEGIN) | (input_casenumber > QFNL_TC_END)) {
      printErr(11, "Valid test case numbers are %d - %d. Try again.", QFNL_TC_BEGIN, QFNL_TC_END);
      exit(0);
    }
    internal_casenumber = input_casenumber - 1;  // the internal cases are zero based
    runApi(internal_casenumber);
  }
  else if (strcmp(parse->IPRSE_parameter, "CASE-ALL") == 0)
  {
    runApi(-1); // Will run all cases
  }

  else {
    printMsg(12,'I',WTOPC_NO_CHAIN,"ECB Exiting because unrecognized parameter!!!!");
    exit(0);
  }


 } else  if ( (levtest(D0)) &&
              (strncmp((char *)&ecbptr()->ebw000, "EXPR", 4) == 0) ) { // If created by cretc_list

   qfnl_pool_build_parm_list_Ptr group_built = (qfnl_pool_build_parm_list_Ptr) ecbptr()->ce1cr0;  // get input parameter list

   if (debug) printf("\n !!! ENTERED by cretc_level for POOL group %d !!!!\n", group_built->group);

   if (group_built->expiry != 0) { // should be nonzero if we got here, but if
                                   //   for some reason it isn't, we will quietly exit and stop
                                   //   creating more ecbs

     qfnl_pool_info_Ptr my_group = qfnl_locate_pool_group (group_built->group, FALSE);

     if (my_group != NULL) { // only process if group found

       if (debug) printf("\n !!! cretc_level ECB found POOL group %d !!!!\n", group_built->group);

       if (memcmp(&my_group->tod_built,
                  &group_built->tod_built,
                  sizeof(group_built->tod_built)) == 0) {  // only proceed if TOD group was built matches mine

         // process this group to mark all VALID pool entries as NOT VALID
         qfnl_pool_set_expired(my_group);

         // create the next ecb to process expiry again after the next expiry interval has ellapsed...
         cretc_level (CRETC_SECONDS|CRETC_1052,
                      QFNL( ),
                      group_built->expiry,
                      "EXPR",
                      D0);

       } // end only proceed if TOD group was built matches mine
       else if (debug) printMsg(1,'I',WTOPC_NO_CHAIN,"POOL GROUP %d WAS FOUND BY EXPIRY BUT TOD NO LONGER MATCHES",
                                group_built->group);

     } // end only process if group found
     else if (debug) printMsg(1,'I',WTOPC_NO_CHAIN,"POOL GROUP %d NO LONGER FOUND BY EXPIRY PROCESSING",
                              group_built->group);

   } // end should be nonzero if we got here
   else if (debug) printMsg(1,'I',WTOPC_NO_CHAIN,"POOL GROUP %d WILL BE IGNORED BY EXPIRY AS INPUT EXPIRY IS NOW ZERO",
                              group_built->group);

   if (levtest(D0))
      relcc(D0);           // We are done with the block passed to us, so release it
   return 0;

 } else {  // else Entered by swisc_create


  if (verbose) printMsg(13,'I',WTOPC_NO_CHAIN,"ECB Activated by swisc_create!!!!");

  if (strncmp( (char *)&ecbptr()->ebw000, "canned", 6) == 0)
  {

    cannedRun();
    return 0;

  } else { // not CANNED

    struct ebmaxc_parmlist myparm[] = {
    {MAX64HEAP, MAXVALUE},
    {0, 0} };

    tpf_ebmaxc(myparm);

    qfnl_child_ecb_parm_list_Ptr incoming_ecb_parm_ptr = (qfnl_child_ecb_parm_list_Ptr) &ecbptr()->ebw000;
    if (verbose) printf("\n !!!!! The child ECB number %d will operate on group %d !!!!\n",
                        incoming_ecb_parm_ptr->ecbNum, incoming_ecb_parm_ptr->group);

  //  printf("\n!!!!! CHILD ecb Received SEED=%x GROUP=%d READS=%d WRITES=%d !!!!!\n",
  //                   incoming_ecb_parm_ptr->child_seed,
  //                   incoming_ecb_parm_ptr->group,
  //                   incoming_ecb_parm_ptr->numberOfReads,
  //                   incoming_ecb_parm_ptr->numberOfWrites);

    if (incoming_ecb_parm_ptr->group_type == QFNL_FIXED)
      qfnl_fixed_child_ecb(incoming_ecb_parm_ptr);
    else
      if (incoming_ecb_parm_ptr->group_type == QFNL_POOL)
         qfnl_pool_child_ecb(incoming_ecb_parm_ptr);

    if (verbose) printMsg(15,'I',WTOPC_NO_CHAIN,"SWISC ECB Exiting !!!");
    exit(0);

  } // end else not CANNED

 }         // end of else Entered by swisc_create

 return 0;
}

/**********************************************************************/
/*                                                                    */
/*  Generic functions used for common tasks that the driver functions */
/*   must perform to service requests.                                */
/**********************************************************************/

int qfnl_random (int min, int max)
{

   int r = 0;
   const int range = 1 + max - min;
   const int buckets = RAND_MAX / range;
   const int limit = buckets * range;

  // Create equal size buckets all in a row, then fire randomly towards
  // the buckets until you land in one of them. All buckes are equally
  // likely.  If you land off the end of the line of buckets, try again

    do
    {
       r = rand();
    } while (r >= limit);

  int ret_val = min + (r / buckets);

  // when min is zero, it seems that ret_val will never be assigned
  //    a value less than 1.  For this reason, when ret_val is 1,
  //    check the latest value of r that was returned from rand().
  //    If r was even, set ret_val to 0 instead of 1.
  //
  //    NOTE: This is to fix a problem where the driver never
  //    seemed to randomly update record ordinal 0.

  if ( (min == 0) && (ret_val == 1) && (!(r % 2)) )
      ret_val = 0;

  return ret_val;

} // end of qfnl_random

qfnl_fixed_ctl_tbl_Ptr get_fixed_table()
{

 qfnl_fixed_ctl_tbl_Ptr  fixed_table_ptr = NULL;

 struct uatbc_area uatbcParms; /* Parameters for UATBC function   */
 uint   frames = QFNL_NUM_FRAMES;
 uint   flags = GSYSC_UNIQUE + GSYSC_64BIT;
 char token[ITOKENLEN] = QFNL_FIXED_TOKEN;
 char owner [IOWNERLEN] = QFNL_FIXED_OWNER;

/* ------------------------------------------------------------------- */
/* Create the unique name for the 64-bit system heap storage by        */
/* combining the characters "QFIX" and the 4 character name of the     */
/* subsystem on which the driver command is running.  Call UATBC to    */
/* determine the name of this subsystem.  Then update the owner name   */
/* with the unique token.                                              */
/* ------------------------------------------------------------------- */

   uatbcParms.uat_call = UAT_ESSU;
   uatbc (&uatbcParms);
   memcpy(&token[4],&uatbcParms.uat_ss_name,sizeof(uatbcParms.uat_ss_name));
   memcpy(&owner[10],&uatbcParms.uat_ss_name,sizeof(uatbcParms.uat_ss_name));

   fixed_table_ptr = (qfnl_fixed_ctl_tbl_Ptr) tpf_fsysc(token,NULL);

   if (fixed_table_ptr == NULL) { // if the control table does not exist,
                                  //   create and initialize it now

    /* ------------------------------------------------------------------- */
    /* Using the GSYSC (tpf_gsysc) function, allocate 1-4K frame of system */
    /* heap storage.  Set the token name to the unique name QFIXxxxx       */
    /* (created in the previous step) and specify the options to use       */
    /* 64-bit heap and unique naming.                                      */
    /* ------------------------------------------------------------------- */

    fixed_table_ptr =  (qfnl_fixed_ctl_tbl_Ptr) tpf_gsysc(frames, token, owner, flags);

    tpf_stpoc(TPF_STPOC_ON, stpoc_value);  // allow ourselves to modify key 9 storage gotten by tpf_gsysc

    /* ------------------------------------------------------------------- */
    /*  Initialize the block of heap storage that was allocated to keep    */
    /*   track of Fixed groups for this subsystem.                         */
    /* ------------------------------------------------------------------- */

    memcpy(fixed_table_ptr->eye_catcher,token,ITOKENLEN);

    fixed_table_ptr->number_of_groups = 0;

    qfnl_fixed_info * group_info_ptr = fixed_table_ptr->my_fixed;

    for(int i=0; i < QFNL_MAX_FIXED; i++) {
      memset(group_info_ptr,0x00,sizeof(qfnl_fixed_info));
      group_info_ptr->group_flag = QFNL_GROUP_STAT_FREESLOT;
      group_info_ptr++;
    }

    tpf_stpoc(TPF_STPOC_OFF, stpoc_value); // done modifying for now

   } // end if the control table does not exist

 return(fixed_table_ptr);

}


qfnl_pool_ctl_tbl_Ptr get_pool_table()
{

 qfnl_pool_ctl_tbl_Ptr  pool_table_ptr = NULL;

 struct uatbc_area uatbcParms; /* Parameters for UATBC function   */
 uint   frames = QFNL_NUM_FRAMES;
 uint   flags = GSYSC_UNIQUE + GSYSC_64BIT;
 char token[ITOKENLEN] = QFNL_POOL_TOKEN;
 char owner [IOWNERLEN] = QFNL_POOL_OWNER;

/* ------------------------------------------------------------------- */
/* Create the unique name for the 64-bit system heap storage by        */
/* combining the characters "QPOO" and the 4 character name of the     */
/* subsystem on which the driver command is running.  Call UATBC to    */
/* determine the name of this subsystem.  Then update the owner name   */
/* with the unique token.                                              */
/* ------------------------------------------------------------------- */

   uatbcParms.uat_call = UAT_ESSU;
   uatbc (&uatbcParms);
   memcpy(&token[4],&uatbcParms.uat_ss_name,sizeof(uatbcParms.uat_ss_name));
   memcpy(&owner[10],&uatbcParms.uat_ss_name,sizeof(uatbcParms.uat_ss_name));

   pool_table_ptr = (qfnl_pool_ctl_tbl_Ptr) tpf_fsysc(token,NULL);

   if (pool_table_ptr == NULL) { // if the control table does not exist,
                                  //   create and initialize it now

    /* ------------------------------------------------------------------- */
    /* Using the GSYSC (tpf_gsysc) function, allocate 1-4K frame of system */
    /* heap storage.  Set the token name to the unique name QPOOxxxx       */
    /* (created in the previous step) and specify the options to use       */
    /* 64-bit heap and unique naming.                                      */
    /* ------------------------------------------------------------------- */

    pool_table_ptr =  (qfnl_pool_ctl_tbl_Ptr) tpf_gsysc(frames, token, owner, flags);

    tpf_stpoc(TPF_STPOC_ON, stpoc_value);  // allow ourselves to modify key 9 storage gotten by tpf_gsysc

    /* ------------------------------------------------------------------- */
    /*  Initialize the block of heap storage that was allocated to keep    */
    /*   track of Fixed groups for this subsystem.                         */
    /* ------------------------------------------------------------------- */

    memcpy(pool_table_ptr->eye_catcher,token,ITOKENLEN);

    pool_table_ptr->number_of_groups = 0;

    qfnl_pool_info * group_info_ptr = pool_table_ptr->my_pool;

    for(int i=0; i < QFNL_MAX_FIXED; i++) {
      memset(group_info_ptr,0x00,sizeof(qfnl_pool_info));
      group_info_ptr->group_flag = QFNL_GROUP_STAT_FREESLOT;
      group_info_ptr++;
    }

    tpf_stpoc(TPF_STPOC_OFF, stpoc_value); // done modifying for now

   } // end if the control table does not exist

 return(pool_table_ptr);

}

/**********************************************************************/
/*                                                                    */
/* This function will modify the contents of a fixed file record,     */
/*   belonging to a fixed file group, that was read into memory.      */
/*   Currently the contents will include the time stamp, the CPU ID   */
/*   that modified the record, and the file address of the record     */
/*   itself, as well as an incremental counter within the record.     */
/*                                                                    */
/*   This funciton returns a 1 if the modify was a success.           */
/*   This function returns a 0 if there was any sort of failure.      */
/* NOTE: It is the responsibility of the caller of this function to   */
/*   actually file the modified record back on the z/TPF database.    */
/*                                                                    */
/* !!!! NOTE ALSO :                !!!!                               */
/*   When modifying this function, you should consider modifying      */
/*    functin qfnl_is_recordOK as it verifies the contents of the     */
/*    record.                                                         */
/*   ALSO consider modifying qfnl_modify_pool_record if the change    */
/*    is applicable to pool file records.                             */
/*                                                                    */
/**********************************************************************/
int qfnl_modify_fixed_record(void * record_ptr,
                             TPF_FA8 file_address,
                             qfnl_fixed_info * group_info_ptr)
{ // qfnl_modify_fixed_record

    tpf_TOD_ext_type tod;
    int myrc = tpf_STCKE(&tod);
    if (myrc != STCKE_OK) {
       printMsg(16,'E',WTOPC_NO_CHAIN,"tpf_STCKE FAILED MODIFYING FA %.16lX FOR FIXED GROUP %d",
             file_address,
             group_info_ptr->group_number);
       tpf_stpoc(TPF_STPOC_ON, stpoc_value);  // allow modify key 9
       tpf_serialized_update8(one, (tsug_t *)&group_info_ptr->group_stats.errCount);
       tpf_serialized_update8(one, (tsug_t *)&group_info_ptr->group_stats.envErr);
       tpf_stpoc(TPF_STPOC_OFF, stpoc_value);
       return(0);
    } else { // store clock worked
       // The store of the extended TOD worked...
       //  so proceed with modifying the contents of the record that was read into memory, this includes
       //   the record ID, file address of the record itself, CPU ID modifying the record, record size,
       //   timestamp, and an incremental counter...

       int tod_size = sizeof(tod);
       char * entry_chunk = (char *) record_ptr;
       qfnl_data_header data_header;
       qfnl_data_header_Ptr my_header = (qfnl_data_header_Ptr) record_ptr;

       int lengthLeft = group_info_ptr->record_size;

       // first set up the entry header to contain record ID, file address,
       //    timestamp, record size, CPU ID as eyecatcher, anything else...

       int header_size = sizeof(data_header);
       if (header_size <= group_info_ptr->record_size) {
          memcpy(&(my_header->rec_ID), group_info_ptr->rec_ID, QFNL_RID_SIZE);
          memset(&(my_header->rcc), 0x00, QFNL_RCC_SIZE);
          memcpy(&(my_header->FA), &file_address,8);
          // printf("\n !!! Size of file address is %d !!!", sizeof(file_address));
          memcpy(&(my_header->tod), &tod, tod_size);
          my_header->record_size = group_info_ptr->record_size;
          my_header->eyecatcher.cpuID = thisCPUid();
          for (size_t index = 0; index < (sizeof(data_header.eyecatcher.repeat_cpuID)); index++)
            my_header->eyecatcher.repeat_cpuID[index] = thisCPUid();
          entry_chunk = entry_chunk + header_size;
          my_header++;
          lengthLeft = lengthLeft - header_size;
       }

       // now set all the items in the data entry

       if (lengthLeft > 0) {

          qfnl_data_item data_item;
          qfnl_data_item_Ptr itemPtr = (qfnl_data_item_Ptr) entry_chunk;

          int item_size = sizeof(data_item);

          int number_of_entries = lengthLeft / item_size;

          for (int j = 1; j <= number_of_entries; j++) {
            memcpy(&(itemPtr->tod), &tod, tod_size);
            memcpy(&(itemPtr->FA), &file_address, sizeof(file_address));
            // itemPtr->file_address = file_address;
            itemPtr->counter = j;
            itemPtr++;
          } // end for

       }

    } // end else store clock worked

    return(1);

} // qfnl_modify_fixed_record

/**********************************************************************/
/*                                                                    */
/* This function will modify the contents of a pool record,           */
/*   belonging to a pool group, that was read into memory.            */
/*   Currently the contents will include the time stamp, the CPU ID   */
/*   that modified the record, and the file address of the record     */
/*   itself, as well as an incremental counter within the record.     */
/*                                                                    */
/*   This function also returns the tod (time stamp) that was put in  */
/*    the record so that the caller can update the time that the pool */
/*    was added and/or updated in the group.                          */
/*                                                                    */
/*   This funciton returns a 1 if the modify was a success.           */
/*   This function returns a 0 if there was any sort of failure.      */
/* NOTE: It is the responsibility of the caller of this function to   */
/*   actually file the modified record back on the z/TPF database.    */
/*                                                                    */
/* !!!! NOTE ALSO :                !!!!                               */
/*   When modifying this function, you should consider modifying      */
/*    functin qfnl_is_recordOK as it verifies the contents of the     */
/*    record.                                                         */
/*   ALSO consider modifying qfnl_modify_fixedrecord if the change    */
/*    is applicable to fixed file records.                            */
/*                                                                    */
/**********************************************************************/
int qfnl_modify_pool_record(void * record_ptr,
                             TPF_FA8 file_address,
                             qfnl_pool_info * group_info_ptr,
                             tpf_TOD_ext_type * caller_tod_ptr)
{ // qfnl_modify_pool_record

    tpf_TOD_ext_type tod;  // local variable used to temporarily store TOD
    int myrc = tpf_STCKE(&tod);
    if (myrc != STCKE_OK) {
       printMsg(16,'E',WTOPC_NO_CHAIN,"tpf_STCKE FAILED MODIFYING FA %.16lX FOR POOL GROUP %d",
             file_address,
             group_info_ptr->group_number);
       tpf_stpoc(TPF_STPOC_ON, stpoc_value);  // allow modify key 9
       tpf_serialized_update8(one, (tsug_t *)&group_info_ptr->group_stats.errCount);
       tpf_serialized_update8(one, (tsug_t *)&group_info_ptr->group_stats.envErr);
       tpf_stpoc(TPF_STPOC_OFF, stpoc_value);
       return(0);
    } else { // store clock worked
       // The store of the extended TOD worked...
       //  so proceed with modifying the contents of the record that was read into memory, this includes
       //   the record ID, file address of the record itself, CPU ID modifying the record, record size,
       //   timestamp, and an incremental counter...

       int tod_size = sizeof(tod);
       char * entry_chunk = (char *) record_ptr;
       qfnl_data_header data_header;
       qfnl_data_header_Ptr my_header = (qfnl_data_header_Ptr) record_ptr;

       int lengthLeft = group_info_ptr->record_size;

       // first set up the entry header to contain record ID, file address,
       //    timestamp, record size, CPU ID as eyecatcher, anything else...

       int header_size = sizeof(data_header);
       if (header_size <= group_info_ptr->record_size) {
          memcpy(&(my_header->rec_ID), group_info_ptr->hex_rid, QFNL_RID_SIZE);
          memset(&(my_header->rcc), 0x00, QFNL_RCC_SIZE);
          memcpy(&(my_header->FA), &file_address,8);
          // printf("\n !!! Size of file address is %d !!!", sizeof(file_address));
          memcpy(&(my_header->tod), &tod, tod_size);
          my_header->record_size = group_info_ptr->record_size;
          my_header->eyecatcher.cpuID = thisCPUid();
          for (size_t index = 0; index < (sizeof(data_header.eyecatcher.repeat_cpuID)); index++)
            my_header->eyecatcher.repeat_cpuID[index] = thisCPUid();
          entry_chunk = entry_chunk + header_size;
          my_header++;
          lengthLeft = lengthLeft - header_size;
       }

       // now set all the items in the data entry

       if (lengthLeft > 0) {

          qfnl_data_item data_item;
          qfnl_data_item_Ptr itemPtr = (qfnl_data_item_Ptr) entry_chunk;

          int item_size = sizeof(data_item);

          int number_of_entries = lengthLeft / item_size;

          for (int j = 1; j <= number_of_entries; j++) {
            memcpy(&(itemPtr->tod), &tod, tod_size);
            memcpy(&(itemPtr->FA), &file_address, sizeof(file_address));
            // itemPtr->file_address = file_address;
            itemPtr->counter = j;
            itemPtr++;
          } // end for

       }

       // Also return the TOD to the caller to store in the table item that keeps track of this pool

       memcpy(caller_tod_ptr, &tod, tod_size);

    } // end else store clock worked

    return(1);

} // qfnl_modify_pool_record

boolean qfnl_is_recordOK(void * record_ptr,
                         TPF_FA8 file_address,
                         qfnl_fixed_info * group_info_ptr)
{ // qfnl_is_recordOK

/********************************************************************/
/*  This function will perform a minimal verify of the contents of  */
/*     the record, mostly by making sure that the file address      */
/*     is contained throughout the record, as well as that the      */
/*     record ID and RCC, record size, and that the integer value   */
/*     is incremented uniformly throughout the record.              */
/*  It returns TRUE if the record is OK                             */
/********************************************************************/

  boolean recordOK = true;  // assume it is ok until we find a problem

  char zeroRCC[2] = {'\0','\0'}; // used for zero record ID check

  char * entry_chunk = (char *) record_ptr;
  qfnl_data_header data_header;
  qfnl_data_header_Ptr my_header = (qfnl_data_header_Ptr) record_ptr;

  int lengthLeft = group_info_ptr->record_size;

// first verify that the header contains the record ID, RCC,
//  and record size that match the information for this group

  if ( memcmp(&(my_header->rec_ID), group_info_ptr->rec_ID, QFNL_RID_SIZE) !=0 ){
    recordOK = false;
    return(recordOK);
  }

  if ( memcmp(&(my_header->rcc), zeroRCC, QFNL_RCC_SIZE) !=0 ){
    recordOK = false;
    return(recordOK);
  }

  if ( my_header->record_size != group_info_ptr->record_size ) {
    recordOK = false;
    return(recordOK);
  }

// also verify that the header contains the file address of the record
//  itself

  if ( memcmp(&(my_header->FA), &file_address, sizeof(file_address)) !=0 ){
    recordOK = false;
    return(recordOK);
  }

// now verify the rest of the record contents

  int header_size = sizeof(data_header);
  entry_chunk = entry_chunk + header_size;
  lengthLeft = group_info_ptr->record_size - header_size;

  if (lengthLeft > 0) { // If there is anything left

     qfnl_data_item data_item;
     qfnl_data_item_Ptr itemPtr = (qfnl_data_item_Ptr) entry_chunk;

     int item_size = sizeof(data_item);

     int number_of_entries = lengthLeft / item_size;

     for (int j = 1; (j <= number_of_entries)&&recordOK; j++) {
       if ( memcmp(&(itemPtr->FA), &file_address, sizeof(file_address)) != 0 )
          recordOK = false;
       if ( itemPtr->counter != j )
          recordOK = false;
       itemPtr++;
     } // end for

  } // end If there is anything left

  return(recordOK);

} // qfnl_is_recordOK

boolean qfnl_is_recordOK(void * record_ptr,
                         TPF_FA8 file_address,
                         qfnl_pool_info * group_info_ptr)
{ // qfnl_is_recordOK

/********************************************************************/
/*  This function will perform a minimal verify of the contents of  */
/*     the record, mostly by making sure that the file address      */
/*     is contained throughout the record, as well as that the      */
/*     record ID and RCC, record size, and that the integer value   */
/*     is incremented uniformly throughout the record.              */
/*  It returns TRUE if the record is OK                             */
/********************************************************************/

  boolean recordOK = true;  // assume it is ok until we find a problem

  char zeroRCC[2] = {'\0','\0'}; // used for zero record ID check

  char * entry_chunk = (char *) record_ptr;
  qfnl_data_header data_header;
  qfnl_data_header_Ptr my_header = (qfnl_data_header_Ptr) record_ptr;

  int lengthLeft = group_info_ptr->record_size;

// first verify that the header contains the record ID, RCC,
//  and record size that match the information for this group

  if ( memcmp(&(my_header->rec_ID), group_info_ptr->hex_rid, QFNL_RID_SIZE) !=0 ){
    recordOK = false;
    return(recordOK);
  }

  if ( memcmp(&(my_header->rcc), zeroRCC, QFNL_RCC_SIZE) !=0 ){
    recordOK = false;
    return(recordOK);
  }

  if ( my_header->record_size != group_info_ptr->record_size ) {
    recordOK = false;
    return(recordOK);
  }

// also verify that the header contains the file address of the record
//  itself

  if ( memcmp(&(my_header->FA), &file_address, sizeof(file_address)) !=0 ){
    recordOK = false;
    return(recordOK);
  }

// now verify the rest of the record contents

  int header_size = sizeof(data_header);
  entry_chunk = entry_chunk + header_size;
  lengthLeft = group_info_ptr->record_size - header_size;

  if (lengthLeft > 0) { // If there is anything left

     qfnl_data_item data_item;
     qfnl_data_item_Ptr itemPtr = (qfnl_data_item_Ptr) entry_chunk;

     int item_size = sizeof(data_item);

     int number_of_entries = lengthLeft / item_size;

     for (int j = 1; (j <= number_of_entries)&&recordOK; j++) {
       if ( memcmp(&(itemPtr->FA), &file_address, sizeof(file_address)) != 0 )
          recordOK = false;
       if ( itemPtr->counter != j )
          recordOK = false;
       itemPtr++;
     } // end for

  } // end If there is anything left

  return(recordOK);

} // qfnl_is_recordOK

/**********************************************************************/
/*                                                                    */
/* Functions to handle each option of the ZTEST FFACT command.        */
/*                                                                    */
/**********************************************************************/
static void qfnl_fixed_build(struct IPRSE_output *parse,
                             int num_parms)
{

 int i = 0;  // index

 qfnl_fixed_ctl_tbl_Ptr  fixed_table_ptr = NULL;

 // parse the options specified for the ZTEST FFACT FIXED BUILD command...
 if ( qfnl_parse_fixed_build(parse,num_parms) == 0 )
   exit(0);   // exit if parse function did not like something

 // Make sure that that the file system is available by checking if GFS is active.
 //  Even though this is fixed and other ZTEST FFACT FIXED functions should work
 //  without GFS, build may need to create a file in the file system to communicate
 //  the creation of a new fixed file group to other processors in the complex and
 //  to "remember" this group across IPLs on the current processor.  Without GFS
 //  active, this file cannot be created in the file system and the ZTEST FFACT FIXED BUILD
 //  command would fail.

 /*******************************************************************/
 /*   Make sure filesystem is available                             */
 /*******************************************************************/

 char *state_ptr = (char*)cinfc_fast_ss(CINFC_CMMOPS, ecbptr()->ce1dbi);
 if (*state_ptr & GFS_INACTIVE) { // GFS_INACTIVE - BIT7 in CMMOPS
 // GFS is not active - error
   printMsg(80,'E',WTOPC_CHAIN,"FILE SYSTEM MUST BE ACTIVE FOR ZTEST FFACT FIXED BUILD");
   printMsg(80,0,WTOPC_CHAIN,"       BECAUSE BUILD OF A NEW FIXED FILE GROUP");
   printMsg(80,0,WTOPC_CHAIN,"       CREATES A FILE IN THE FILE SYSTEM.");
   printMsg(80,0,WTOPC_CHAIN,"       ONCE A FIXED FILE GROUP HAS BEEN");
   printMsg(80,0,WTOPC_CHAIN,"       BUILT, OTHER ZTEST FFACT FIXED");
   printMsg(80,0,WTOPC_CHAIN_END,"       COMMANDS WILL WORK WITHOUT GFS.");
   exit(0);
 }

 fixed_table_ptr = get_fixed_table();

 if (fixed_table_ptr == NULL) {
   printMsg(17,'E',WTOPC_NO_CHAIN,"CANNOT FIND OR ALLOCATE FIXED FILE CONTROL TABLE - BUILD FAILED");
   return;
 } else {  // have table for fixed

   printMsg(3,'I',WTOPC_NO_CHAIN,"ADDRESS OF FIXED CONTROL TABLE IS %.16lX", (long) fixed_table_ptr);

/*----------------------------------------------------------------------------------------------------*/
    // Before we add a group number for the current group we are building to our table, make sure
    //  it is not already there.
    //  If it is already there and is active, do not perform another build but inform the user that if they
    //  wish to do a build, they must either specify another group number or kill the current group
    //  number so that it can be re-used.

    qfnl_fixed_info * group_info_ptr = fixed_table_ptr->my_fixed;

    // Lock the file whose name is based on this group just to prevent more than one processor from trying to build a group with the
    //   same number at the same time

    int fixed_lock_file;  // will be used to lock building a fixed group with this number
    mode_t             mode = S_IRUSR | S_IWUSR | S_IXUSR |
                              S_IRGRP | S_IWGRP | S_IXGRP |
                              S_IROTH | S_IWOTH | S_IXOTH;


    fixed_lock_file = open(FIXED_lock_name, O_RDWR+O_CREAT, mode);

    if (fixed_lock_file == -1) {

        printMsg(18,'E',WTOPC_NO_CHAIN,"Cannot Open File %s for WRITE, errno = %d", FIXED_lock_name, errno);
        tpf_stpoc(TPF_STPOC_ON, stpoc_value);  // allow modify key 9
        tpf_serialized_update8(1, (tsug_t *)&group_info_ptr->group_stats.errCount);
        tpf_serialized_update8(1, (tsug_t *)&group_info_ptr->group_stats.envErr);
        tpf_stpoc(TPF_STPOC_OFF, stpoc_value);
        return;

    }

    struct   flock          lock_it;
    fstream                 my_file;

    lock_it.l_type = F_WRLCK;      /* write lock this file   */
    lock_it.l_whence = SEEK_SET;   /* set to start of file   */
    lock_it.l_start = 0;           /* lock from start of file*/
    lock_it.l_len = 0;             /* ...to the end of file. */

    // lock the file to serialize creating a fixed file group of this number across all processors in the complex

    while ( fcntl(fixed_lock_file,F_SETLKW,&lock_it) == -1 ) {
       printMsg(19,'W',WTOPC_NO_CHAIN,"TEMPOARILY FAILED TO GET LOCK on KEY FILE %s errno = %d", FIXED_lock_name, errno);
       sleep (2);
    }

    my_file.open (FIXED_lock_name, fstream::in);
    if(!my_file.is_open()) {

    // If file will not open, return error
    // errno is set on the open

      tpf_stpoc(TPF_STPOC_ON, stpoc_value);  // allow modify key 9
      tpf_serialized_update8(1, (tsug_t *)&group_info_ptr->group_stats.errCount);
      tpf_serialized_update8(1, (tsug_t *)&group_info_ptr->group_stats.envErr);
      tpf_stpoc(TPF_STPOC_OFF, stpoc_value);

      printMsg(20,'E',WTOPC_NO_CHAIN,"qfnl_fixed_build CANNOT OPEN FILE %s errno= %d",FIXED_lock_name,errno);
      return;
    }

    // FIXED_GROUP_LOCK = fopen(FIXED_lock_name, "a");
    // char say_lock[] = "LOCK";
    // fwrite(say_lock, 1, sizeof(say_lock), FIXED_GROUP_LOCK);

    // Also lock the table so no other ECB on the same processor adds or deletes an element from under us

    corhc(fixed_table_ptr); // hold table before search

    for(i=0; (i < QFNL_MAX_FIXED); i++) { // for all items in table until we find a match
      if ( (my_fixed_build.group == group_info_ptr->group_number) &&
           (group_info_ptr->group_flag != QFNL_GROUP_STAT_FREESLOT) )
      {

        printMsg(21,'E',WTOPC_NO_CHAIN,"FIXED GROUP %d HAS ALREADY BEEN BUILT.  USE A DIFFERENT GROUP NUMBER OR KILL FIXED GROUP %d",
                   my_fixed_build.group, my_fixed_build.group);

        // unlock and close the lock file as we are bailing out

        my_file.close();

        lock_it.l_type = F_UNLCK;      //Unlock
        lock_it.l_whence = SEEK_SET;   //Set to start of file
        lock_it.l_start = 0;           //Lock from start of file
        lock_it.l_len = 0;             //Lock to the end of file
        fcntl(fixed_lock_file, F_SETLKW, &lock_it);
        close(fixed_lock_file);
        remove(FIXED_lock_name);       // we no longer need the file, so get rid of it

        // Also unhold the table of fixed groups so other ECBs can get at it
        coruc(fixed_table_ptr); // unhold table before exit

        exit(0);

      } else

        group_info_ptr++;  // advance to next slot in my_fixed array

    } // end for all items in table

    // Now that we know the group is not already in our table, we can add it there...


         // Make sure our table is big enough to keep track of this group

         if (fixed_table_ptr->number_of_groups >= QFNL_MAX_FIXED) {
             printMsg(22,'E',WTOPC_NO_CHAIN,"WE CANNOT KEEP TRACK OF MORE THAN %d FIXED FILE GROUPS", fixed_table_ptr->number_of_groups);
             printMsg(23,'I',WTOPC_NO_CHAIN,"FOR MORE FIXED FILE GROUPS, CHANGE QFNL_MAX_FIXED IN qfnl.h, REBUILD,RELOAD QFNL and re-IPL");

             // unlock and close the lock file as we are bailing out
             my_file.close();

             lock_it.l_type = F_UNLCK;      //Unlock
             lock_it.l_whence = SEEK_SET;   //Set to start of file
             lock_it.l_start = 0;           //Lock from start of file
             lock_it.l_len = 0;             //Lock to the end of file
             fcntl(fixed_lock_file, F_SETLKW, &lock_it);
             close(fixed_lock_file);
             remove(FIXED_lock_name);       // we no longer need the file, so get rid of it

             // Also unhold the table of fixed groups so other ECBs can get at it
             coruc(fixed_table_ptr); // unhold table before exit

             exit(0);
         }

         // Look for an item in the table that is not in use

         boolean found = false;
         group_info_ptr = fixed_table_ptr->my_fixed;
         for(i=0; (i < QFNL_MAX_FIXED)&&(!found); i++) { // for all items in table until we find a free item
             if (group_info_ptr->group_flag == QFNL_GROUP_STAT_FREESLOT) {
                found = true;
             }

             if (!found) group_info_ptr++;  // only if not found, advance to next slot in my_fixed array

         } // end for all items in table until we find a free item

         if (!found) {
              printMsg(24,'E',WTOPC_NO_CHAIN,"LOGIC ERROR: NO FREE ITEM IN FIXED FILE GROUP TABLE OF %d ENTRIES FOUND", QFNL_MAX_FIXED);

              // unlock and close the lock file as we are bailing out
              my_file.close();

              lock_it.l_type = F_UNLCK;      //Unlock
              lock_it.l_whence = SEEK_SET;   //Set to start of file
              lock_it.l_start = 0;           //Lock from start of file
              lock_it.l_len = 0;             //Lock to the end of file
              fcntl(fixed_lock_file, F_SETLKW, &lock_it);
              close(fixed_lock_file);
              remove(FIXED_lock_name);       // we no longer need the file, so get rid of it

              // Also unhold the table of fixed groups so other ECBs can get at it
              coruc(fixed_table_ptr); // unhold table before exit

              exit(0);
         }

         // A free item in the table was found.  If there is no file in the file system indicating that a group with the same number
         //   was already built, we will store information for the current fixed group build request in it.
         // Otherwise (if there is a file in the file system indicating that a group with the same number was already built), we
         //   reject this build request, and copy information from the file system for that group number to the table so that the
         //   user can start activity for that group number on this processor.

         // Check if there is already a file in the file system indicating a group of the same number was already created.

         boolean failure = false;        // initially assume no failure
         boolean already_built = false;  // tells if we already had built a group with the same number

         already_built = qfnl_check_if_groupAlreadyBuilt(FIXED_file_name, &already_fixed_build, &failure);

         if (failure) {

            // the file is out there indicating a build was done for this group, but a problem was found with the file
            //  So inform the user and bail out
            //   (NOTE: They will have to kill the group and enter the build command again if they really want to build
            //      this fixed group.)

            printMsg(25, 'E', WTOPC_NO_CHAIN, "BUILD FOR FIXED GROUP %d FAILED", my_fixed_build.group);

            // unlock and close the lock file as we are bailing out
            my_file.close();

            lock_it.l_type = F_UNLCK;      //Unlock
            lock_it.l_whence = SEEK_SET;   //Set to start of file
            lock_it.l_start = 0;           //Lock from start of file
            lock_it.l_len = 0;             //Lock to the end of file
            fcntl(fixed_lock_file, F_SETLKW, &lock_it);
            close(fixed_lock_file);
            remove(FIXED_lock_name);       // we no longer need the file, so get rid of it

            // Also unhold the table of fixed groups so other ECBs can get at it
            coruc(fixed_table_ptr); // unhold table before exit

            return;

         }

         if (already_built) { // if already built, copy the old build values instead

            tpf_stpoc(TPF_STPOC_ON, stpoc_value);  // allow ourselves to modify key 9 storage gotten by tpf_gsysc

            // copy the values from the build that was alread done ...

            group_info_ptr->group_number = already_fixed_build.group;
            group_info_ptr->begin_ordinal = already_fixed_build.begin_ordinal;
            group_info_ptr->end_ordinal = already_fixed_build.end_ordinal;
            memcpy(group_info_ptr->rec_ID,already_fixed_build.hex_rid,QFNL_RID_SIZE);
            memcpy(group_info_ptr->rec_type,already_fixed_build.rtype,QFNL_RTYPE_SIZE);
            group_info_ptr->record_size = already_fixed_build.record_size;

            // Mark this group as ready to do work ...

            group_info_ptr->group_flag = QFNL_GROUP_STAT_STOPPED;   // mark group stopped, as build now complete
                                                                     // (i.e. no longer in progress)
            fixed_table_ptr->number_of_groups++;  // increment count of groups in memory

            tpf_stpoc(TPF_STPOC_OFF, stpoc_value); // done modifying for now

            // Notify the user that we are not using his values...
            printMsg(26,'E',WTOPC_NO_CHAIN,"GROUP %d PREVIOUSLY BUILT BY CPU-%c. USING THOSE VALUES INSTEAD",
                already_fixed_build.group, already_fixed_build.cpuID);

            // unlock and close the lock file as we are bailing out
            my_file.close();

            lock_it.l_type = F_UNLCK;      //Unlock
            lock_it.l_whence = SEEK_SET;   //Set to start of file
            lock_it.l_start = 0;           //Lock from start of file
            lock_it.l_len = 0;             //Lock to the end of file
            fcntl(fixed_lock_file, F_SETLKW, &lock_it);
            close(fixed_lock_file);
            remove(FIXED_lock_name);       // we no longer need the file, so get rid of it

            // Also unhold the table of fixed groups so other ECBs can get at it
            coruc(fixed_table_ptr); // unhold table before exit

            return;


         } else { // not already built, copy in new build request values

            tpf_stpoc(TPF_STPOC_ON, stpoc_value);  // allow ourselves to modify key 9 storage gotten by tpf_gsysc

            group_info_ptr->group_flag = QFNL_GROUP_STAT_BUILDING;  // build is now in progress for this group
            group_info_ptr->group_number = my_fixed_build.group;     // ... copy in values for new build request ...
            group_info_ptr->begin_ordinal = my_fixed_build.begin_ordinal;
            group_info_ptr->end_ordinal = my_fixed_build.end_ordinal;
            memcpy(group_info_ptr->rec_ID,my_fixed_build.hex_rid,QFNL_RID_SIZE);
            memcpy(group_info_ptr->rec_type,my_fixed_build.rtype,QFNL_RTYPE_SIZE);

            // Update count of fixed file groups that the driver will remember

            fixed_table_ptr->number_of_groups++;

            tpf_stpoc(TPF_STPOC_OFF, stpoc_value); // done modifying for now

            coruc(fixed_table_ptr); // unhold the table now that we are done adding to it

         } // end else not already built

         /*-------------------------------------------------------------------------------------------*/
         /* Starting with the beginning ordinal, get the file address of each ordinal for this        */
         /*  fixed record type, all the way up to and including the ending ordinal, filing out        */
         /*  initialized records along the way.                                                       */
         /*  NOTE:                                                                                    */
         /*    If either a FACE error or an I/O error is experienced for the beginning ordinal,       */
         /*  then abort the build for this group and mark the entire slot that would have been used   */
         /*  for this group in the table as unused.  If an error is experienced for one of the other  */
         /*  ordinals, then issue an error message that the build terminated permaturely and update   */
         /*  the ending ordinal for this group to be the ordinal number that preceded the error.      */
         /*-------------------------------------------------------------------------------------------*/

         TPF_FAC8 fac;
         TPF_FA8 fileaddr;
         DECBC_RC decbRC;

         /* get storage to hold the DECB list */
         int decb_index;                                                        //d26633
         TPF_DECB **decb_list =                                                 //d26633
         (TPF_DECB **) calloc(decb_build_count, sizeof(decb_list));             //d26633


         char FARFstring[QFNL_FARF_STRING_SIZE];

         unsigned int current_ordinal = 0;
         unsigned int prev_ordinal = 0;
         int success_count = 0;  // we will count how many records successfully processed
         void ** record_ptr_list = (void **) calloc(decb_build_count, sizeof(record_ptr_list)); //d26633
         char errMsg[100] = ""; // use to generate error messages

         // allocate DECBs to work with the records in this group               //d26633
         for (decb_index = 0; decb_index<decb_build_count; decb_index++)        //d26633
         {                                                                      //d26633
            decb_list[decb_index] = tpf_decb_create(NULL, &decbRC);             //d26633
              if (decb_list[decb_index]  == NULL)                               //d26633
              { // if could not allocate a DECB, build failed
              failure = true;
              sprintf(errMsg,"ALLOCATE DECB FAILED FOR BUILD OF FIXED GROUP %d", my_fixed_build.group);
              printMsg(27, 'E', WTOPC_NO_CHAIN, errMsg);
              } // end if could not allocate a DECB
         }                                                                      //d26633

         decb_index = 0; /* reset index */                                      //d26633


         for ( current_ordinal = group_info_ptr->begin_ordinal;
               (current_ordinal <= group_info_ptr->end_ordinal) && (!failure);
               current_ordinal++ ) 

           { // for every ordinal in the group unless there is a failure

           fac.ifacord = current_ordinal;

           memcpy(fac.ifacrec, "        ", sizeof(fac.ifacrec));
           memcpy(fac.ifacrec, group_info_ptr->rec_type, strlen(group_info_ptr->rec_type));
           fac.ifactyp = IFAC8FCS;
           tpf_fac8c(&fac);

           if (fac.ifacret != TPF_FAC8_NRM) { // FACS_ERROR

             failure = true;           // flag failure
             // put out an appropriate error message depending on the nature of the FACS error
             switch (fac.ifacret)
             {
             case TPF_FAC8_NIU:
               sprintf(errMsg,"FACS ERROR PROCESSING ORDINAL %x, RECORD TYPE %s : NOT IN USE",
                       fac.ifacord, group_info_ptr->rec_type);
                       // current_ordinal, group_info_ptr->rec_type);
               break;
             case TPF_FAC8_RTH:
               sprintf(errMsg,"FACS ERROR PROCESSING ORDINAL %x, RECORD TYPE %s : RECORD TYPE NONEXISTENT OR EXCEEDS LIMIT",
                       fac.ifacord, group_info_ptr->rec_type);
                       // current_ordinal, group_info_ptr->rec_type);
               break;
             case TPF_FAC8_ROR:
               sprintf(errMsg,"FACS ERROR PROCESSING ORDINAL %x, RECORD TYPE %s : ORDINAL OUT OF RANGE",
                       current_ordinal, group_info_ptr->rec_type);
               break;
             case TPF_FAC8_NSP:
               sprintf(errMsg,"FACS ERROR PROCESSING ORDINAL %x, RECORD TYPE %s : NO SPLIT CHAIN FOR RECORD",
                       current_ordinal, group_info_ptr->rec_type);
               break;
             case TPF_FAC8_POR:
               sprintf(errMsg,"FACS ERROR PROCESSING ORDINAL %x, RECORD TYPE %s : INPUT PARAMETER NOT VALID",
                       current_ordinal, group_info_ptr->rec_type);
               break;
             default:
               sprintf(errMsg,"FACS ERROR PROCESSING ORDINAL %x, RECORD TYPE %s : UNKNOWN ERROR",
                       current_ordinal, group_info_ptr->rec_type);
               break;
             }  // end switch
             printMsg(28, 'E', WTOPC_NO_CHAIN, errMsg);
             failure = true;

           // end FACS_ERROR

           } else { // file address retrieved from FACS

             decb_list[decb_index]->idecfa = fac.ifacadr;  // get file address retrieved from FACS //d26633
             fileaddr = fac.ifacadr;      // ditto

             int convertOK = qfnl_FARF6_to_hex_string(FARFstring, sizeof(FARFstring), fileaddr);
             if ( convertOK == 1 ) {  // converted ok
                if (debug) printf("\n !!! Hooray.... For REC TYPE %s Ordinal %x File address is %s !!\n",
                            group_info_ptr->rec_type, current_ordinal, FARFstring);
             }

             if (success_count == 0) { // if this is the first time a DECB is used, try to read it to get its size   //d26633
                                       // NOTE:   we will use the core block gotten on the read to initialize all the//d26633
                                       //     other records in the group                                             //d26633
                // read the record if all the DECBs have not yet been initialized                                    //d26633

                record_ptr_list[decb_index] = find_record_ext(decb_list[decb_index],  // decb //d26633
                                             &fileaddr,                          // file address of record to read
                                             RECID_RESET,                        // bypass record ID check
                                             '\0',                               // bypass RCC
                                             NOHOLD_WAIT,                        // find without hold and wait
                                             FIND_DEFEXT);                       // use default non GDS

                if (record_ptr_list[decb_index] == NULL) { // if failure finding beginning ordinal   //d26633

                  failure = true;
                  sprintf(errMsg, "ERROR READING RECORD TYPE %s ORDINAL %x, SUD = %2.2X",
                                   group_info_ptr->rec_type, current_ordinal,  decb_list[decb_index]->idecsud); //d26633
                  printMsg(29, 'E', WTOPC_NO_CHAIN, errMsg);


                } // end if failure finding ordinal to initialize the DECB

                if (!failure) {
                  tpf_stpoc(TPF_STPOC_ON, stpoc_value);  // allow ourselves to modify key 9 storage gotten by tpf_gsysc
                  group_info_ptr->record_size = decb_list[decb_index]->cbrw.cb_size; // get the record size  //d26633
                  tpf_stpoc(TPF_STPOC_OFF, stpoc_value);  // done for now
                  if (debug) printf("\n !!! Hooray .... Record SIZE first record in group is %d", group_info_ptr->record_size);

                  // also save the record size in the structure that will get saved in the file system
                  //   so other processors will know the record size in advance in case you want to start
                  //   an instance with only writes (only valid for LOCK-N)

                  my_fixed_build.record_size = decb_list[decb_index]->cbrw.cb_size; //d26633
                }


             }  // end if this is the first time a DECB is used ...             //d26633

             // file out the record for the build
             if (!failure) {

               // Modify the contents of the record in memory

               if (qfnl_modify_fixed_record(record_ptr_list[decb_index],        //d26633
                                            fileaddr,
                                            group_info_ptr) == 0) {  // if modify fails quit now
                 failure = true;
               }
             }

             if (!failure) {  // if no failures yet, try to file the record

               file_record_ext(decb_list[decb_index],                           //d26633
                               &fileaddr,
                               group_info_ptr->rec_ID,
                               '\0',
                               NOREL,
                               FILE_DEFEXT);

              /* only wait if all the decb's are pending writes or we're at the end of the ordinal range */   //d26633
              if ((decb_index == (decb_build_count-1)) || (current_ordinal == group_info_ptr->end_ordinal ))  //d26633
              {                                                                                               //d26633
 
               if ( waitc() ) {  // if error on wait
                  failure = true;
                  if (convertOK) {
                    sprintf(errMsg, "ERROR FILING RECORD TYPE %s ORDINAL %x, FA=%s SUD=%2.2X",
                                     group_info_ptr->rec_type, current_ordinal, FARFstring, decb_list[decb_index]->idecsud); //d26633

                  } else {
                    sprintf(errMsg, "ERROR FILING RECORD TYPE %s ORDINAL %x, SUD=%2.2X",
                                     group_info_ptr->rec_type, current_ordinal, decb_list[decb_index]->idecsud); //d26633
                  }

                  printMsg(30, 'E', WTOPC_NO_CHAIN, errMsg);
               }
               decb_index = 0;                                                  //d26633
               success_count++;                                                 //d26633
              }
               else decb_index++;                                               //d26633

             }

             if (!failure) {
               prev_ordinal = current_ordinal;  // save as the last ordinal that worked
             }

           }        // end else file address retrieved from FACS

           if (group_info_ptr->group_flag == QFNL_GROUP_STAT_STOPPED) { // If group suddenly stopped, terminate the build
                                                                         //   as someone most likely entered a ZTEST FFACT FIXED STOP
                                                                         //   command...

             failure = true;
             printMsg(31,'E',WTOPC_NO_CHAIN,"BUILD FOR GROUP %d TERMINATING: STOPPED STATUS DETECTED",
                                group_info_ptr->group_number);

           } // end If group suddenly stopped

         } // end for every ordinal in the group unless there is a failure

         if (failure) {

            //  If there was a failure, determine if the failure was total
            //     or if the build just terminated prematurely, and update the fixed group table
            //     accordingly...

            if (success_count == 0) { // if no records were successfully processed, consider it a total failure

               printMsg(25, 'E', WTOPC_NO_CHAIN, "BUILD FOR FIXED GROUP %d FAILED", my_fixed_build.group);

               // On a total failure, clean up by removing all record of the failed build from our fixed group control table...

               tpf_stpoc(TPF_STPOC_ON, stpoc_value);  // allow ourselves to modify key 9 storage gotten by tpf_gsysc

               group_info_ptr->group_flag = QFNL_GROUP_STAT_FREESLOT;  // nothing built so make table slot free
               group_info_ptr->group_number = 0;
               group_info_ptr->begin_ordinal = 0;
               group_info_ptr->end_ordinal = 0;
               memset(group_info_ptr->rec_ID,0x00,QFNL_RID_SIZE);
               memset(group_info_ptr->rec_type,0x00,QFNL_RTYPE_SIZE);

               // Decrement count of fixed file groups that the driver will remember

               fixed_table_ptr->number_of_groups--;

               tpf_stpoc(TPF_STPOC_OFF, stpoc_value); // done modifying for now

               // unlock and close the lock file as we are bailing out
               my_file.close();

               lock_it.l_type = F_UNLCK;      //Unlock
               lock_it.l_whence = SEEK_SET;   //Set to start of file
               lock_it.l_start = 0;           //Lock from start of file
               lock_it.l_len = 0;             //Lock to the end of file
               fcntl(fixed_lock_file, F_SETLKW, &lock_it);
               close(fixed_lock_file);
               remove(FIXED_lock_name);       // we no longer need the file, so get rid of it

               return;             // Now that we have cleaned up, bail out on total failure!!!!

            } else {  // partially successful

               printMsg(35, 'E', WTOPC_NO_CHAIN, "BUILD FOR FIXED GROUP %d TERMINATED AT ORDINAL %x",
                        group_info_ptr->group_number, prev_ordinal);
               tpf_stpoc(TPF_STPOC_ON, stpoc_value);  // allow ourselves to modify key 9 storage gotten by tpf_gsysc
               group_info_ptr->group_flag = QFNL_GROUP_STAT_STOPPED;   // build no longer in progress
               group_info_ptr->end_ordinal = prev_ordinal;
               my_fixed_build.end_ordinal = prev_ordinal;
               tpf_stpoc(TPF_STPOC_OFF, stpoc_value); // done modifying for now

            }         // end else partially successful

         // end if failure
         } else { // else successful build

            tpf_stpoc(TPF_STPOC_ON, stpoc_value);  // allow ourselves to modify key 9 storage gotten by tpf_gsysc
            group_info_ptr->group_flag = QFNL_GROUP_STAT_STOPPED;   // mark group stopped, as build now complete
                                                                     // (i.e. no longer in progress)
            tpf_stpoc(TPF_STPOC_OFF, stpoc_value); // done modifying for now


            printMsg(36, 'I', WTOPC_NO_CHAIN, "BUILD FOR FIXED GROUP %d COMPLETED", group_info_ptr->group_number);
         }        // end else successful build


/*----------------------------------------------------------------------------------------------------*/

   // At this point we write the file to the file system that describes the build that was done, so other
   //    processors in the complex can know about this fixed file group...


   size_t my_fixed_build_size = sizeof(my_fixed_build);
   FIXED_GROUP_FILE = fopen(FIXED_file_name, "w");
   fwrite(&my_fixed_build, 1, my_fixed_build_size, FIXED_GROUP_FILE);
   fclose(FIXED_GROUP_FILE);

   // unlock the file that was used to serialize build of this group across all processors

   char say_lock[] = "LOCK";
   write(fixed_lock_file, &say_lock, (size_t) sizeof(say_lock));

   // unlock and close the file as we are done

   my_file.close();

   lock_it.l_type = F_UNLCK;      //Unlock
   lock_it.l_whence = SEEK_SET;   //Set to start of file
   lock_it.l_start = 0;           //Lock from start of file
   lock_it.l_len = 0;             //Lock to the end of file
   fcntl(fixed_lock_file, F_SETLKW, &lock_it);
   close(fixed_lock_file);
   remove(FIXED_lock_name);       // we no longer need the file, so get rid of it

 } // end have table for fixed

}

int qfnl_parse_fixed_build(struct IPRSE_output *parse,
                             int num_parms)
{

//
// This function parses the build command and if an error is found it returns 0, otherwise
//   it returns 1.
//
// NOTE: This function parses the input parameters and uses them to assign values to
//   the my_fixed_build structure that represents the fixed group build that is being
//   requested.  It also uses the group number to generate the name of the file in
//   the file system that will be created to represent this group, if the fixed group
//   build request is accepted and completes successfully.
//

  my_fixed_build.group = 0;
  my_fixed_build.begin_ordinal = 0;
  my_fixed_build.end_ordinal = 0;
  memset (my_fixed_build.hex_rid,'\0',sizeof(my_fixed_build.hex_rid));
  memset (my_fixed_build.rtype,' ',sizeof(my_fixed_build.rtype));
  char rid_zeroes[QFNL_RID_SIZE];  // used to check if all zeroes specified for record ID
  memset (rid_zeroes, 0x00, QFNL_RID_SIZE);

  unsigned int input_rid_length = 0;
  char input_rid[5];
  memset (input_rid,'\0',sizeof(input_rid));
  memset (my_fixed_build.groupString, '\0', sizeof(my_fixed_build.groupString));

  if (debug) printf("\n!! Hooray!! Recognized FIXED BUILD !!\n");
  /******************************************************************/
  /* Loop through all BUILD parameters                              */
  /******************************************************************/
  decb_build_count = 10; // set DECB default                                    //d26633
  parse = parse->IPRSE_next;
  num_parms--;

  while ((num_parms > 0) && parse)
  { // while there are more parameters

   if (debug) printMsg(9, 'I', WTOPC_NO_CHAIN, "!!! PARSE PARM IS %s", parse->IPRSE_parameter);

   if (parse && (strncmp(parse->IPRSE_parameter, "Group-",6) == 0))
   { // if Group specified

    // get the group number and save it both as an integer and as a string

    my_fixed_build.group = atoi(parse->IPRSE_value);
    memcpy(my_fixed_build.groupString, parse->IPRSE_value, strlen(parse->IPRSE_value));

    if (my_fixed_build.group <=0) {
      printMsg(37,'E',WTOPC_NO_CHAIN,"GROUP %d NOT DEFINED:  GROUP NUMBER MUST BE GREATER THAN 0", my_fixed_build.group);
      return(0);
    }

    //  The parser should prevent an integer greater than 5 digits, but I coded a check anyway just in case

    if (my_fixed_build.group > 99999) {
      printMsg(37,'E',WTOPC_NO_CHAIN,"GROUP %d NOT DEFINED:  MAXIMUM GROUP NUMBER IS 99999", my_fixed_build.group);
      return(0);
    }

    if (debug) printMsg(1,'I',WTOPC_NO_CHAIN,"!!! input Group = %d",my_fixed_build.group);
    if (debug) printMsg(1,'I',WTOPC_NO_CHAIN,"!!! input Group String = %s",my_fixed_build.groupString);

    // generate the file name that will be used to communicate the existence of this group to other processors in the complex

    sprintf(FIXED_file_name,"%s%s%s",QFNL_DIR,QFNL_PFX,my_fixed_build.groupString);

    // generate the file name that will be used to serialize builds of this given group number

    sprintf(FIXED_lock_name,"%s%s%s",QFNL_DIR,QFNL_LCK,my_fixed_build.groupString);

   } // end if Group specified

   if (parse && (strncmp(parse->IPRSE_parameter, "Type-",5) == 0))
   { // if Type specified

    memcpy(my_fixed_build.rtype, parse->IPRSE_value, strlen(parse->IPRSE_value));
    if (debug) printMsg(1,'I',WTOPC_NO_CHAIN,"!!! my_fixed_build.rtype initially set to %s!!!",my_fixed_build.rtype);
    if (debug) printMsg(1,'I',WTOPC_NO_CHAIN,"!!! parse->IPRSE_value at this point is %s!!!",parse->IPRSE_value);

    // If the record type entered by the user is not valid, try prefixing it with # and see if
    //    that record type is valid.

    if (tpf_rtcuc_cton(my_fixed_build.rtype) < 0) {  // if input record type is not valid

      // If there is room for the #, try adding it as a prefix and see if this makes a valid
      //    record type.  If so, this will be the record type that we pass to the build function.

      if (strlen(parse->IPRSE_value) < 8) {

        memset (my_fixed_build.rtype,' ',sizeof(my_fixed_build.rtype));
        my_fixed_build.rtype[0] = '#';
        memcpy(&my_fixed_build.rtype[1], parse->IPRSE_value, strlen(parse->IPRSE_value));
        if (debug) printMsg(1,'I',WTOPC_NO_CHAIN,"!!! my_fixed_build.rtype is NOW %s!!!",my_fixed_build.rtype);
        if (debug) printMsg(1,'I',WTOPC_NO_CHAIN,"!!! parse->IPRSE_value is NOW %s!!!",parse->IPRSE_value);

        // If the record type prefixed with # is no good, put it back to what the user had specified
        //    and let the build routine identify the error

        if (tpf_rtcuc_cton(my_fixed_build.rtype) < 0) {
           memset (my_fixed_build.rtype,' ',sizeof(my_fixed_build.rtype));
           memcpy(my_fixed_build.rtype, parse->IPRSE_value, strlen(parse->IPRSE_value));
           if (debug) printMsg(1,'I',WTOPC_NO_CHAIN,"!!! my_fixed_build.rtype SET BACK TO %s!!!",my_fixed_build.rtype);
        }

      }

      //   otherwise leave it as is and the build will tell the user that the rtype is no good

    } // end if input recrod type is not valid

    if (debug) printMsg(1,'I',WTOPC_NO_CHAIN,"!!! my_fixed_build.rtype set to %s!!!",my_fixed_build.rtype);


   } // end if Type specified

   if (parse && (strncmp(parse->IPRSE_parameter, "RID-",4) == 0))
   { // if RID specified

    input_rid_length = convert_input_data_to_hex(parse->IPRSE_value,
                                        input_rid);
    if (debug) printMsg(1,'I',WTOPC_NO_CHAIN,"!!! input RID length = %d", input_rid_length);
    memcpy(my_fixed_build.hex_rid,input_rid,input_rid_length);
    if (debug) printMsg(1,'I',WTOPC_NO_CHAIN,"!!! input RID in character is = %s", my_fixed_build.hex_rid);
    if ( memcmp(my_fixed_build.hex_rid, rid_zeroes, QFNL_RID_SIZE) == 0 ) {  // if RID specified is all zeroes
      printMsg(81,'E',WTOPC_NO_CHAIN,"RECORD ID ALL ZEROES NOT ALLOWED WHEN BUILDING A GROUP");
      return(0);
    }

   } // end if RID specified

   if (parse && (strncmp(parse->IPRSE_parameter, "Begin-",6) == 0))
   { // if Begin specified

    // get input beginning ordinal value
    char temp_begin[17];        // temp begin ordinal + '\0'

    strcpy(temp_begin,parse->IPRSE_value);
    my_fixed_build.begin_ordinal  = strtoul(temp_begin, (char**) NULL, HEX_BASE);

    if (debug) printMsg(1,'I',WTOPC_NO_CHAIN,"!!! BEGIN ORDINAL IN HEX = %x IN DECIMAL = %d !!!",
                 my_fixed_build.begin_ordinal, my_fixed_build.begin_ordinal);

   } // end if Begin specified

   if (parse && (strncmp(parse->IPRSE_parameter, "End-",4) == 0))
   { // if End specified

    // get input ending ordinal value
    char temp_end[17];        // temp end ordinal + '\0'

    strcpy(temp_end,parse->IPRSE_value);
    my_fixed_build.end_ordinal  = strtoul(temp_end, (char**) NULL, HEX_BASE);

    if (debug) printMsg(1,'I',WTOPC_NO_CHAIN,"!!! END ORDINAL IN HEX = %x IN DECIMAL = %d !!!",
                 my_fixed_build.end_ordinal, my_fixed_build.end_ordinal);

   } // end if End specified

   parse = parse->IPRSE_next;
   num_parms--;

   if (parse && (strncmp(parse->IPRSE_parameter, "Decb-",5) == 0))              //d26633
   { // if DECB count specified                                                 //d26633

    decb_build_count = atoi(parse->IPRSE_value);                                //d26633

    if (debug) printMsg(1,'I',WTOPC_NO_CHAIN,"decb_build_count = %d",decb_build_count); //d26633

    if ( (decb_build_count < 0) ||                                              //d26633
         (decb_build_count > 99) ) {                                            //d26633
      printMsg(46,'E',WTOPC_NO_CHAIN,"GROUP %d NOT BUILT:  DECB COUNT MUST BE LESS THAN 100 AND GREATER THAN ZERO", //d26633 
         my_fixed_start.group);                                                 //d26633
      return(0);                                                                //d26633
    }                                                                           //d26633
    } // End of DECB count specified                                            //d26633


  } // end while there are more parameters

 // Even though the CPU ID that does the build is not an explicit parameter, determine the
 //   CPU ID on which the build command was entered and save it as one of the build
 //   parameters.  This way, when the build information is later written to the file system,
 //   we can know which CPU in the complex did the build.

 my_fixed_build.cpuID = thisCPUid();

 return(1);

}

boolean qfnl_first_record_ok(qfnl_pool_info_Ptr group_info_ptr, char *rec_ID_ptr)
{ // begin qfnl_first_record_ok (for pools)

//   return(true);
// This function will read the first record from the group using zero record ID
//   and will then obtain the record ID that was read for that record.
// If the record ID matches the group record ID and the contents of the record
//   have been properly initialized, it will return TRUE that the first record is OK
//   and it will store the record ID read at the address referenced by
//   rec_ID_ptr.
// If there is any error reading the record, or the record ID does not match the group,
//   or the record is not properly initialized, FALSE is returned that the first record is
//   NOT OK.

  boolean record_ok = true;  // assume the record is ok until we find a problem

  char errMsg[100] = "";

  // We need the following in order to do i/o:

  TPF_DECB *decb = NULL;
  DECBC_RC decbRC;
  TPF_FA8 fileaddr;
  void * record_ptr = NULL;  // will eventually point to record read or initialized
  enum t_find_decb findtype=NOHOLD_WAIT; // used for find without hold

  char FARFstring[QFNL_FARF_STRING_SIZE];


  // allocate a DECB

  decb = tpf_decb_create(NULL, &decbRC);

  if (decb == NULL) { // if could not allocate a DECB

    record_ok = false;
    sprintf(errMsg,"ALLOCATE DECB FAILED FOR START COMMAND FOR POOL GROUP %d",
                         group_info_ptr->group_number);
    serrc_op(SERRC_ALL_HEAP_RETURN,0x5555,errMsg, NULL);

    // increment error counter(s)
    tpf_stpoc(TPF_STPOC_ON, stpoc_value);   // allow ourselves to modify key 9 storage gotten by tpf_gsysc
    tpf_serialized_update8(1, (tsug_t *)&group_info_ptr->group_stats.errCount);
    tpf_serialized_update8(1, (tsug_t *)&group_info_ptr->group_stats.envErr);
    tpf_stpoc(TPF_STPOC_OFF, stpoc_value);  // done with modify for now

    return(record_ok);  // just bail out now

  } // end if could not allocate a DECB

  // get the file address of the first pool in this group

  if (qfnl_get_pool_fileaddr(group_info_ptr,
                              1,
                              decb,
                              &fileaddr,
                              FARFstring,
                              QFNL_FARF_STRING_SIZE,
                              true)) {  // if cannot get file address

     // increment error counter(s)
     tpf_stpoc(TPF_STPOC_ON, stpoc_value);   // allow ourselves to modify key 9 storage gotten by tpf_gsysc
     tpf_serialized_update8(1, (tsug_t *)&group_info_ptr->group_stats.errCount);
     tpf_serialized_update8(1, (tsug_t *)&group_info_ptr->group_stats.envErr);
     tpf_stpoc(TPF_STPOC_OFF, stpoc_value);  // done with modify for now

     record_ok = false;

    // end if cannot get file address
  } else { // else got file address, so read the record

     // read the record and then verify its contents

     record_ptr = find_record_ext(decb,                               // decb
                                  &fileaddr,                          // file address of record to read
                                  RECID_RESET,                        // bypass use of record ID
                                  '\0',                               // bypass RCC
                                  findtype,                           // type of find controls hold or not
                                  FIND_DEFEXT);                       // use default non GDS

     if (record_ptr == NULL) { // if failure finding beginning ordinal

       sprintf(errMsg, "ERROR READING RECORD NUMBER %d OF POOL GROUP %d, FA=%s SUD = %2.2X",
               1, group_info_ptr->group_number, FARFstring, decb->idecsud);
       serrc_op(SERRC_ALL_HEAP_RETURN,0x5555,errMsg, NULL);

       // increment error counter(s)
       tpf_stpoc(TPF_STPOC_ON, stpoc_value);   // allow ourselves to modify key 9 storage gotten by tpf_gsysc
       tpf_serialized_update8(1, (tsug_t *)&group_info_ptr->group_stats.errCount);
       tpf_serialized_update8(1, (tsug_t *)&group_info_ptr->group_stats.ioErr);
       tpf_stpoc(TPF_STPOC_OFF, stpoc_value);  // done with modify for now

       record_ok = false;

       // end if failure finding beginning ordinal
     } else if ( qfnl_is_recordOK(record_ptr, fileaddr, group_info_ptr) ) { // if record contents ok (i.e. initialized and has same rec ID as group)

                memcpy(rec_ID_ptr, group_info_ptr->hex_rid, QFNL_RID_SIZE);          // copy record ID from group which was verified to match the one in the record

                // increment count of successful reads

                tpf_stpoc(TPF_STPOC_ON, stpoc_value);   // allow ourselves to modify key 9 storage gotten by tpf_gsysc
                tpf_serialized_update8(1, (tsug_t *)&group_info_ptr->group_stats.readCount);
                tpf_stpoc(TPF_STPOC_OFF, stpoc_value);  // done with modify for now

            } else { // end else record read is NOT OK

                sprintf(errMsg, "QUESTIONABLE CONTENTS IN POOL GROUP %d, %dTH RECORD WITH FA=%s",
                       group_info_ptr->group_number, 1, FARFstring);
                serrc_op(SERRC_ALL_HEAP_RETURN,0x5555,errMsg, NULL);

                // increment error counter(s)
                tpf_stpoc(TPF_STPOC_ON, stpoc_value);   // allow ourselves to modify key 9 storage gotten by tpf_gsysc
                tpf_serialized_update8(1, (tsug_t *)&group_info_ptr->group_stats.errCount);
                tpf_serialized_update8(1, (tsug_t *)&group_info_ptr->group_stats.contentErr);
                tpf_stpoc(TPF_STPOC_OFF, stpoc_value);  // done with modify for now
                record_ok = false;

            }   // end else record read is NOT OK

  } // end else got file address...

  // free resources

  if ( levtest(decb) ) {  // test if need free record buffer
    relcc(decb);
    record_ptr = NULL;
  }

  if (decb)
    tpf_decb_release(decb);  // release decb if it exists

  return(record_ok);

} // end qfnl_first_record_ok (for pools)

boolean qfnl_first_record_ok(qfnl_fixed_info_Ptr group_info_ptr, char *rec_ID_ptr)
{

// This function will read the first record from the group using zero record ID
//   and will then obtain the record ID that was read for that record.
// If the record ID matches the group record ID and the contents of the record
//   have been properly initialized, it will return TRUE that the first record is OK
//   and it will store the record ID read at the address referenced by
//   rec_ID_ptr.
// If there is any error reading the record, or the record ID does not match the group,
//   or the record is not properly initialized, FALSE is returned that the first record is
//   NOT OK.

  boolean record_ok = true;  // assume the record is ok until we find a problem

  char errMsg[100] = "";

  // We need the following in order to do i/o:

  TPF_DECB *decb = NULL;
  DECBC_RC decbRC;
  TPF_FA8 fileaddr;
  void * record_ptr = NULL;  // will eventually point to record read or initialized
  enum t_find_decb findtype=NOHOLD_WAIT; // used for find without hold

  char FARFstring[QFNL_FARF_STRING_SIZE];


  // allocate a DECB

  decb = tpf_decb_create(NULL, &decbRC);

  if (decb == NULL) { // if could not allocate a DECB

    record_ok = false;
    sprintf(errMsg,"ALLOCATE DECB FAILED FOR START COMMAND FOR FIXED GROUP %d",
                         group_info_ptr->group_number);
    serrc_op(SERRC_ALL_HEAP_RETURN,0x5555,errMsg, NULL);

    // increment error counter(s)
    tpf_stpoc(TPF_STPOC_ON, stpoc_value);   // allow ourselves to modify key 9 storage gotten by tpf_gsysc
    tpf_serialized_update8(1, (tsug_t *)&group_info_ptr->group_stats.errCount);
    tpf_serialized_update8(1, (tsug_t *)&group_info_ptr->group_stats.envErr);
    tpf_stpoc(TPF_STPOC_OFF, stpoc_value);  // done with modify for now

    return(record_ok);  // just bail out now

  } // end if could not allocate a DECB

  // get the file address of the first ordinal in this group

  if (qfnl_get_fixed_fileaddr(group_info_ptr,
                              group_info_ptr->begin_ordinal,
                              decb,
                              &fileaddr,
                              FARFstring,
                              QFNL_FARF_STRING_SIZE,
                              true)) {  // if cannot get file address

     // increment error counter(s)
     tpf_stpoc(TPF_STPOC_ON, stpoc_value);   // allow ourselves to modify key 9 storage gotten by tpf_gsysc
     tpf_serialized_update8(1, (tsug_t *)&group_info_ptr->group_stats.errCount);
     tpf_serialized_update8(1, (tsug_t *)&group_info_ptr->group_stats.envErr);
     tpf_stpoc(TPF_STPOC_OFF, stpoc_value);  // done with modify for now

     record_ok = false;

    // end if cannot get file address
  } else { // else got file address, so read the record

     // read the record and then verify its contents

     record_ptr = find_record_ext(decb,                               // decb
                                  &fileaddr,                          // file address of record to read
                                  RECID_RESET,                        // bypass use of record ID
                                  '\0',                               // bypass RCC
                                  findtype,                           // type of find controls hold or not
                                  FIND_DEFEXT);                       // use default non GDS

     if (record_ptr == NULL) { // if failure finding beginning ordinal

       sprintf(errMsg, "ERROR READING RECORD TYPE %s ORDINAL %x, FA=%s SUD = %2.2X",
               group_info_ptr->rec_type, group_info_ptr->begin_ordinal, FARFstring, decb->idecsud);
       serrc_op(SERRC_ALL_HEAP_RETURN,0x5555,errMsg, NULL);

       // increment error counter(s)
       tpf_stpoc(TPF_STPOC_ON, stpoc_value);   // allow ourselves to modify key 9 storage gotten by tpf_gsysc
       tpf_serialized_update8(1, (tsug_t *)&group_info_ptr->group_stats.errCount);
       tpf_serialized_update8(1, (tsug_t *)&group_info_ptr->group_stats.ioErr);
       tpf_stpoc(TPF_STPOC_OFF, stpoc_value);  // done with modify for now

       record_ok = false;

       // end if failure finding beginning ordinal
     } else if ( qfnl_is_recordOK(record_ptr, fileaddr, group_info_ptr) ) { // if record contents ok (i.e. initialized and has same rec ID as group)

                memcpy(rec_ID_ptr, group_info_ptr->rec_ID, QFNL_RID_SIZE);          // copy record ID from record

                // increment count of successful reads

                tpf_stpoc(TPF_STPOC_ON, stpoc_value);   // allow ourselves to modify key 9 storage gotten by tpf_gsysc
                tpf_serialized_update8(1, (tsug_t *)&group_info_ptr->group_stats.readCount);
                tpf_stpoc(TPF_STPOC_OFF, stpoc_value);  // done with modify for now

            } else { // end else record read is NOT OK

                sprintf(errMsg, "QUESTIONABLE RECORD CONTENTS RECORD TYPE %s ORDINAL %x, FA=%s",
                       group_info_ptr->rec_type, group_info_ptr->begin_ordinal, FARFstring);
                serrc_op(SERRC_ALL_HEAP_RETURN,0x5555,errMsg, NULL);

                // increment error counter(s)
                tpf_stpoc(TPF_STPOC_ON, stpoc_value);   // allow ourselves to modify key 9 storage gotten by tpf_gsysc
                tpf_serialized_update8(1, (tsug_t *)&group_info_ptr->group_stats.errCount);
                tpf_serialized_update8(1, (tsug_t *)&group_info_ptr->group_stats.contentErr);
                tpf_stpoc(TPF_STPOC_OFF, stpoc_value);  // done with modify for now
                record_ok = false;

            }   // end else record read is NOT OK

  } // end else got file address...

  // free resources

  if ( levtest(decb) ) {  // test if need free record buffer
    relcc(decb);
    record_ptr = NULL;
  }

  if (decb)
    tpf_decb_release(decb);  // release decb if it exists

  return(record_ok);

} // end qfnl_first_record_ok

static void qfnl_fixed_start(struct IPRSE_output *parse,
                             int num_parms)
{ // begin qfnl_fixed_start

 if (debug) printf("\n!! Hooray!! Recognized FIXED START !!\n");

 qfnl_child_ecb_parm_list ecb_parameters;       // parameters to pass to children ECBs
 struct swisc_create_input create_parameters;   // used for swisc parameters

 struct ev0bk event_blk;    // for event processing

 // parse the options specified for the ZTEST FFACT FIXED START command...
 if ( qfnl_parse_fixed_start(parse,num_parms) == 0 )
   return;    // return if parser did not like something

 // At this point, we know that the ZTEST FFACt START command passed preliminary syntax validation.
 //  The next thing to do is see if we can find or put information for that group in the
 //  Fixed File Control Table

 qfnl_fixed_info_Ptr found_fixed_info_ptr = qfnl_locate_fixed_group(my_fixed_start.group, TRUE, TRUE);
 if (found_fixed_info_ptr == NULL) {  // group not found
   printMsg(39,'E',WTOPC_NO_CHAIN,"START OF GROUP %d FAILED: GROUP INFO FROM BUILD NOT FOUND",
            my_fixed_start.group);
   return;
 } else {  // group found

   printMsg(3,'I',WTOPC_NO_CHAIN,"ADDRESS OF fixed group found is %.16lX", (long) found_fixed_info_ptr);
   if (debug) printf("\n !!!! Start Command Preliminary Syntax Checks Accepted !!!");

   // First check that a build for this group is not still in progress.  If that is the case, we cannot
   //  start an instance of activity on this group until the build completes.  So reject the request.

   if (found_fixed_info_ptr->group_flag == QFNL_GROUP_STAT_BUILDING) {
     printMsg(40,'E',WTOPC_NO_CHAIN,"START COMMAND FOR GROUP %d REJECTED: BUILD IS IN PROGRESS",
              my_fixed_start.group);
     return;
   }

   // If this instance of the driver is to be started with a utilization class, ensure that the utilization class
   //   is valid (defined on file and loaded into memory) before proceeding...

   // if ( (my_fixed_start.useUtilClass)&&(lodic_isUtilClass(my_fixed_start.utilClass)==0) ) {
   if (my_fixed_start.useUtilClass) { // if we are using a utilization class, validate it
     if (lodic_isUtilClass(my_fixed_start.utilClass)==0) { // if utilizaiton class not valid
       printMsg(110,'E',WTOPC_NO_CHAIN,"START COMMAND FOR GROUP %d REJECTED: UTILCLASS %s NOT IN MEMORY",
                my_fixed_start.group, my_fixed_start.utilClass);
       return;
     }
   } // end if we are using a utilization class, validate it

   /* ---------------------------  Beginning of fix to allow ZTEST i FFACT FIXED START to work when called via ZEXEC ------------------ */
   //
   //  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
   //  NOTE !!! It is important for this to code, which contains a fork to a child ECB who will create the grandchildren
   //         that do the actual work, to be done BEFORE setting of any ECB attributes via tpf_easetc
   //  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
   //
   // Instead of doing the work ourselves, we fork another ECB that will serve as the potential long-running ECB to do the
   //     work.  The reason for this is to allow ZEXEC to execute a ZTEST FFACT FIXED START command which loops
   //     forever.  If the ECB for the command itself loops forever, ZEXEC will never get control back and will timeout.
   //     So we want the ZTEST FFACT FIXED START parent ECB to exit, and have the child created by fork do the looping and possibly
   //     be a longrunning ECB that continuously generates work.

   pid_t pid = fork();
   if (pid == -1) {  // if fork failed...
     // Access profile definition in memory to flag error and stop profile

     tpf_stpoc(TPF_STPOC_ON, stpoc_value);  // allow ourselves to modify key 9 storage gotten by tpf_gsysc
     tpf_serialized_update8(1, (tsug_t *)&found_fixed_info_ptr->group_stats.envErr);  // consider this an environmental error
     found_fixed_info_ptr->group_flag = QFNL_GROUP_STAT_STOPPED; // stop all activity this group
     tpf_stpoc(TPF_STPOC_OFF, stpoc_value); // done modifying for now

     // take a dump and exit
     char errMsg[100] = "";
     sprintf(errMsg,"START COMMAND FOR GROUP %d ATTEMPT TO EXECUTE fork() FAILED",
   		     found_fixed_info_ptr->group_number);
     serrc_op(SERRC_ALL_HEAP,0x5555,errMsg,NULL);
     return;  // return to exit
   }  // end if fork failed...

   if (pid) { // if parent of fork, do nothing but exit
     if (debug) printMsg(1,'I',WTOPC_NO_CHAIN,"!!!!! PARENT of START of GROUP %d exiting!!!!!", found_fixed_info_ptr->group_number);
     return; // return to exit
   } // end if parent of fork

   // We only arrive here if we are the child ECB of the fork.  This child will loop the number of times (or indefinitely) on the profile,
   //   creating its own children to operate on that profile.

   /* ----------------------------     end   of fix to allow ZTEST i FFACT FIXED START to work when called via ZEEXC ------------------ */

   // Now check if this group is already started.  If it is, then make sure that the LOCK-Y or N
   //  matches the locking status for this group.
   // Otherwise if the group is not already started, this is the first instance for the group
   //  so set the locking status for the group to match the locking status specified for this instance.

   boolean lockBasedOnInstance = false;  // initially assume this instance will not determine locking status for whole group

   if (found_fixed_info_ptr->group_flag == QFNL_GROUP_STAT_STARTED ) { // group already started

     if ( (found_fixed_info_ptr->group_stats.lockingActive) && (my_fixed_start.doLocking == 0) ) {
       printMsg(41,'E',WTOPC_NO_CHAIN,"START COMMAND FOR GROUP %d REJECTED: ACTIVITY ALREADY RUNNING WITH LOCKING",
                my_fixed_start.group);
       return;
     }

     if ( (found_fixed_info_ptr->group_stats.lockingActive == 0) && (my_fixed_start.doLocking) ) {
       printMsg(42,'E',WTOPC_NO_CHAIN,"START COMMAND FOR GROUP %d REJECTED: ACTIVITY ALREADY RUNNING WITHOUT LOCKING",
                my_fixed_start.group);
       return;
     }

   } else { // group not started, so set its locking status to match this instance
     lockBasedOnInstance = true;   // we will set locking status for group based on this instance
   }

   // Before actually starting activity for this instance, check whether or not we need to use a LODIC utilization class
   //    and/or resource class...
   // NOTE: The driver will use LODIC_IBMBATCH for its resource class

   int rc = 1;  // assume success until failure

   if (my_fixed_start.useUtilClass) {

     // if we want to use both a utilization class and resource class, then do so...

     if (my_fixed_start.useResClass) {

       printMsg(1,'I',WTOPC_NO_CHAIN,"calling lodic_ext(LODIC_ECBCREATE+LODIC_IBMBATCH+LODIC_CPU_UTIL, NULL, my_fixed_start.utilClass);");
       rc = lodic_ext(LODIC_ECBCREATE+LODIC_IBMBATCH+LODIC_CPU_UTIL, NULL, my_fixed_start.utilClass);

     } else { // otherwise use just a utilization class...

       printMsg(1,'I',WTOPC_NO_CHAIN,"calling lodic_ext(LODIC_ECBCREATE+LODIC_NONE+LODIC_CPU_UTIL, NULL,  my_fixed_start.utilClass);");
       rc = lodic_ext(LODIC_ECBCREATE+LODIC_NONE+LODIC_CPU_UTIL, NULL, my_fixed_start.utilClass);

     } // end else otherwise use just a utilization class...

   } else {

      if (my_fixed_start.useResClass) { // if use just resource class

       printMsg(1,'I',WTOPC_NO_CHAIN,"calling lodic_ext(LODIC_ECBCREATE+LODIC_IBMBATCH, NULL);");
         rc = lodic_ext(LODIC_ECBCREATE+LODIC_IBMBATCH, NULL);

      } // end if use just resource class

   }

   // if there was an error from lodic_ext, abort the start and inform the user

   char errMsg[100] = ""; // use to generate error messages

   if (rc == 0) {  // if error form lodic

      switch (errno) {
         case ETPFLODIC_LOWRES:
            sprintf(errMsg, "GROUP %d NOT STARTED: LODIC FAILED DUE TO LOW RESOURCES",
                    my_fixed_start.group);
            break;
         case ETPFLODIC_MAXSUSP:
            sprintf(errMsg, "GROUP %d NOT STARTED: LODIC FAILED DUE TO MAX SUSPEND (NOT EXPECTED)",
                    my_fixed_start.group);
            break;
         case ETPFLODIC_NOTFOUND:
            sprintf(errMsg, "GROUP %d NOT STARTED: UTIL CLASS %s DOES NOT EXIST",
                    my_fixed_start.group, my_fixed_start.utilClass);
            break;
         default:
            sprintf(errMsg, "GROUP %d NOT STARTED: LODIC FAILED WITH UNKNOWN ERROR",
                    my_fixed_start.group);
            break;

      printMsg(112, 'E', WTOPC_NO_CHAIN, errMsg);
      return; // bail out

      } // end switch

   } // end if error form lodic

   // ALSO before actually starting activity for this instance, check and perform any tpf_easetc that was
   //    requested on the START command...

   if (my_fixed_start.lowPriority) { // if low priority requested
      if (my_fixed_start.inheritLow) {
        tpf_easetc(TPF_EASETC_LOWPRIORITY, TPF_EASETC_INHERIT_YES+TPF_EASETC_SET_ON);
      } else {
        tpf_easetc(TPF_EASETC_LOWPRIORITY, TPF_EASETC_INHERIT_NO+TPF_EASETC_SET_ON);
      }
   }

   if (my_fixed_start.switchable) { // if switchable requested
      if (my_fixed_start.inheritSW) {
        tpf_easetc(TPF_EASETC_SWITCHABLE, TPF_EASETC_INHERIT_YES+TPF_EASETC_SET_ON);
      } else {
        tpf_easetc(TPF_EASETC_SWITCHABLE, TPF_EASETC_INHERIT_NO+TPF_EASETC_SET_ON);
      }
   }

   //
   // Now that we know that the group is not currently being built, and there there is no conflict
   //   with this instance and previously started instances still running in terms of locking,
   //   and any lodic_ext and/or tpf_easetc that was requested has been performed,
   //   prepare to spawn child ECBs to perform activity against this group:
   //
   //     Set status indicator for this group to show that activity has been started for it
   //       and increment the total count of active instances for this group.
   //       ALSO, if this is the first instance to be started for this group,
   //        set status of whether or not locking should be done for the instances operating on
   //        this group.
   //

   //     Rather than blindly using the record ID for the group, read the first record in the group using record ID zero
   //       to determine the record ID.  Verify that the record ID read from the first record in this fashion matches the
   //       record ID that had been stored in the group information for this group during BUILD.  (Also verify that the
   //       first record is initalized properly.)
   //
   //     Then pass the record ID that was read to the child ECBs as a parameter.
   //
   //     NOTE: This is only done for purposes of testing find with record ID zero.

   if (!qfnl_first_record_ok(found_fixed_info_ptr, ecb_parameters.rec_ID)) { // if problem getting record ID from first record
                                                                             //   or the record is not properly initialized
     printMsg(43,'E',WTOPC_NO_CHAIN,"START OF FIXED GROUP %d ABORTED DUE TO PROBLEM WITH FIRST RECORD",
                                     my_fixed_start.group);
     return; // !!! Important to bail out before setting any started indicator for this instance !!!
   }

   tpf_stpoc(TPF_STPOC_ON, stpoc_value);  // allow ourselves to modify key 9 storage gotten by tpf_gsysc
   found_fixed_info_ptr->group_flag = QFNL_GROUP_STAT_STARTED;
   if (lockBasedOnInstance)
     found_fixed_info_ptr->group_stats.lockingActive = my_fixed_start.doLocking;
   tpf_serialized_update8(1, (tsug_t *)&found_fixed_info_ptr->group_stats.instanceCount);
   tpf_stpoc(TPF_STPOC_OFF, stpoc_value); // done modifying for now

   //
   //     Set up child ECB parameters for the instance being started.
   //        NOTE: The only parameter we do not set is the child_seed.  That will be different for each
   //        ECB that is created.
   //

   ecb_parameters.group = my_fixed_start.group;  // group number ECB is to operate on
   ecb_parameters.group_type = QFNL_FIXED;       // dealing with ZTEST FFACt FIXED START ...
   ecb_parameters.numberOfReads = my_fixed_start.numberOfReads;  // number of reads to be done by each child ECB
   ecb_parameters.numberOfWrites = my_fixed_start.numberOfWrites;  //number of writes to be done by each child ECB
   ecb_parameters.doLocking = my_fixed_start.doLocking;  // whether ECB is to lock the records prior to writing to them



   //
   //     Set up SWISC parameters that will be used evey time a new ECB is created for this instance
   //

   create_parameters.program = "QFNL";
   create_parameters.istream = ecbptr()->ce1isn;         // create ECB on current i-stream for starters
   create_parameters.cpu_list = SWISC_LIST_READY;        // and put it on ready list
   create_parameters.bypass = SWISC_ACTNUM_CURRENT;
   create_parameters.ebw_data = &ecb_parameters;
   create_parameters.ebw_data_length = sizeof(ecb_parameters);
   create_parameters.data_level = NO_LVL;


   //
   // While activity for this group has been STARTed, continue to generate the specified number of ECBs on the
   //   input list, waiting a bit after generating each batch of ECBs
   //

   int j; // index of inner loop

   srand(my_fixed_start.seed);  // seed the random generator

   // int K = 1;  // TEMP PATCH!!!!!

   // printf("\n!!!! my_fixed_start.numECBs = %d !!!!\n", my_fixed_start.numECBs);
   unsigned int ecbNum = 1;  // pass a sequential number to each child ECB useful just for debugging messages

   // for (K=1; K <= 10; K++ ) { // TEMP PATCH!!! temporary loop of 10 just to print random numbers for first 10 batches of ECBs

   while (found_fixed_info_ptr->group_flag == QFNL_GROUP_STAT_STARTED) { // while this group is started
     for (j=1;
          (j <= my_fixed_start.numECBs)&&(found_fixed_info_ptr->group_flag == QFNL_GROUP_STAT_STARTED);
          j++) { // loop to generate batch of ECBs

        ecb_parameters.ecbNum = ecbNum;  // pass ecb number to child
        ecbNum++; // increment for next iteration of loop
        // printf("\n !!!!! ecb_parameters.ecbNum = %d !!!!!\n", ecb_parameters.ecbNum);

        // generate random seed to be passed to child ECB to seed its own srand function call

        ecb_parameters.child_seed = qfnl_random(found_fixed_info_ptr->begin_ordinal,
                                                found_fixed_info_ptr->end_ordinal);

                                 //       (int) found_fixed_info_ptr->begin_ordinal,
                                 //       (int) found_fixed_info_ptr->end_ordinal);

        if (debug) printf("\n !!!! Begin Ordinal=%x End Ordinal=%x Input Seed=%d !!!!\n",
                          found_fixed_info_ptr->begin_ordinal,
                          found_fixed_info_ptr->end_ordinal,
                          my_fixed_start.seed);
        if (verbose) printf("\n !!!! The SEED to pass to CHILD is %x !!!!\n", ecb_parameters.child_seed);

        // create child ECB

        swisc_create(&create_parameters);

     } // end loop to generate batch of ECBs

     // Wait 10 milliseconds and put the parent ECB back on the input list before generating the next batch of children ECBs

     event_blk.evnpstinf.evnbkc1 = 1;
     tpf_evntc_timeout(&event_blk, EVENT_CNT, 'N', 10, EVNTC_1052, TPF_EVENT_MILLI);
     evnwc(&event_blk, EVENT_CNT);

     dlayc();     // take a trip to the input list

   } // end while this group is STARTed

 } // end else group found

 // We only reach here if this group is no longer started.  So decrement the total number of active instances
 //   and exit this instance

 tpf_stpoc(TPF_STPOC_ON, stpoc_value);  // allow ourselves to modify key 9 storage gotten by tpf_gsysc

 tpf_serialized_update8(-1, (tsug_t *)&found_fixed_info_ptr->group_stats.instanceCount);

 printMsg(10,'I',WTOPC_NO_CHAIN,
          "Instance for Fixed Group %d is EXITING. Total instances left = %ld", //d26352
          my_fixed_start.group, found_fixed_info_ptr->group_stats.instanceCount);

 // It may not be necessary, but just as a safety mechanism, if there are no more instances left
 //   set the locking indicator for this group to be the default of locking.  Also inform the user
 //   that the group is now STOPPED.

 if (found_fixed_info_ptr->group_stats.instanceCount == 0) {
   found_fixed_info_ptr->group_stats.lockingActive = 1;
   printMsg(57,'I',WTOPC_NO_CHAIN,"FIXED GROUP %d STOPPED", my_fixed_start.group);
 }

 tpf_stpoc(TPF_STPOC_OFF, stpoc_value); // done modifying for now

 return; // return to exit

} // end qfnl_fixed_start

qfnl_fixed_info_Ptr qfnl_locate_fixed_group (int group,
                                             boolean beNoisy,
                                             boolean lookInFileSystem)
{ // BEGINNING of function qfnl_locate_fixed_group

/* ****************************************************************************************************** */
/* NOTE: The purpose of this function is to find the fixed file group either in the Fixed File Control    */
/*   Table, or in the file system.  If it is not found in the Fixed File Control Table, but exists on     */
/*   in the file system, it will be put into the Fixed File Control Table.                                */
/*  Input:                                                                                                */
/*        int group is the group number we are looking for                                                */
/*        boolean beNoisy = TRUE print informational or internal error messages                           */
/*                          FALSE work silently, only put out messages on first major error               */
/*        boolean lookInFileSystem                                                                        */
/*                          TRUE look in file system if group not found in table                          */
/*                          FILE do not look in file system; return error if entry not found in table     */
/*  Output is the address of the group information in the Fixed File Control Table.  If any error is      */
/*    encountered or the fixed file group is not found, NULL is returned.                                 */
/*                                                                                                        */
/*                                                                                                        */
/* ****************************************************************************************************** */

 qfnl_fixed_info_Ptr group_info_ptr = NULL; // hold address of group information in Fixed File Control Table;
                                            // NULL until found or put there.

 // Find the Fixed File Control Table or allocate it, if it is not already in memory

 qfnl_fixed_ctl_tbl_Ptr fixed_table_ptr = get_fixed_table();
 int i;

 // If we could not find or allocate the Fixed File Control Table in memory, we will not be able to
 //    put the group information in there, even if it could be found on file.  So return error.

 if (fixed_table_ptr == NULL) {

   if (beNoisy)
      printMsg(2,'E',WTOPC_NO_CHAIN,"CANNOT FIND OR ALLOCATE FIXED FILE CONTROL TABLE");
   return (NULL);

 } else {  // have table for fixed

   if (beNoisy)
      printMsg(3,'I',WTOPC_NO_CHAIN,"ADDRESS OF FIXED CONTROL TABLE IS %.16lX", (long) fixed_table_ptr);

    /*----------------------------------------------------------------------------------------------------*/
    // Now that the Fixed File Control Table exists in memory, first search it to see if there is
    //   record of this group.  If there is, then we are done and we ignore the file system.
    //   If there is no entry for the group in this table, we will next look in the file system.
    /*----------------------------------------------------------------------------------------------------*/

    boolean we_are_done = false;

    group_info_ptr = fixed_table_ptr->my_fixed;  // point at first fixed group slot in table

    if (debug) printf("\n !!!!! The group we are trying to LOCATE is %d !!!!\n", group);

    if (lookInFileSystem) corhc(fixed_table_ptr);  // hold table before potentially adding to it

    for(i=0; (i < QFNL_MAX_FIXED)&&(!we_are_done); i++) { // for all items in table until we find a match
      if (debug) printf("\n !!!!! ......... Looking at Fixed Group Control Table entry for Group = %d !!!!!\n", group_info_ptr->group_number);
      if ( (group == group_info_ptr->group_number) &&
           (group_info_ptr->group_flag != QFNL_GROUP_STAT_FREESLOT) )
      {
        we_are_done = true;
        if (lookInFileSystem) coruc(fixed_table_ptr);  // unhold table now that we know we do not need to add to it
        if (debug) printf("\n!!!! LOCATE of GROUP %d FLAG is %d ADDRESS OF GROUP IS %.16lX!!!!\n",
                            group_info_ptr->group_number, group_info_ptr->group_flag, group_info_ptr);
      }

      if (!we_are_done) group_info_ptr++;  // advance to next slot in my_fixed array

    } // end for all items in table

    if (!we_are_done) { // if we are not done

       if (debug) printf("\n !!!! We are not done !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n");
       // If the caller said not to look in the file system, then just quietly exit (this will be the case for child ECBs)

         if (!lookInFileSystem) {
            return(NULL);
         }

       // Before looking in the file system, make sure our table is big enough to keep track of this cache

         if (fixed_table_ptr->number_of_groups >= QFNL_MAX_FIXED) {
             printMsg(22,'E',WTOPC_NO_CHAIN,"WE CANNOT KEEP TRACK OF MORE THAN %d FIXED FILE GROUPS", fixed_table_ptr->number_of_groups);
             printMsg(23,'I',WTOPC_NO_CHAIN,"FOR MORE FIXED FILE GROUPS, CHANGE QFNL_MAX_FIXED IN qfnl.h, REBUILD,RELOAD QFNL and re-IPL");

             coruc(fixed_table_ptr);  // unhold table before return
             return(NULL);
         }

         // Look for an item in the table that is not in use

         boolean found = false;
         group_info_ptr = fixed_table_ptr->my_fixed;  // point at first fixed group slot in table
         for(i=0; (i < QFNL_MAX_FIXED)&&(!found); i++) { // for all items in table until we find a free item
             if (group_info_ptr->group_flag == QFNL_GROUP_STAT_FREESLOT) {
                found = true;
             }

             if (!found) group_info_ptr++;  // only if not found, advance to next slot in my_fixed array

         } // end for all items in table until we find a free item

         if (!found) {
              printMsg(24,'E',WTOPC_NO_CHAIN,"LOGIC ERROR: NO FREE ITEM IN FIXED FILE GROUP TABLE OF %d ENTRIES FOUND", QFNL_MAX_FIXED);
              coruc(fixed_table_ptr); // unhold the table
              return(NULL);
         }

         // A free item in the table was found.  If there is no file in the file system indicating that a group with the same number
         //   was already built, we will return NULL.
         // Otherwise (if there is a file in the file system indicating that a group with the same number was already built),
         //   copy information from the file system for that group number to the table so that the caller can perform
         //   activity on that group number on this processor (e.g. ZTEST FFACT FIXED START etc.)

         // Check if there is already a file in the file system indicating a group of the same number was already created.

         boolean failure = false;        // initially assume no failure
         boolean already_built = false;  // tells if we already had built a group with the same number

         already_built = qfnl_check_if_groupAlreadyBuilt(FIXED_file_name, &already_fixed_build, &failure);

         if (failure) {
            // the file is out there indicating a build was done for this group, but a problem was found with the file,
            //  so just bail out.

            printMsg(38, 'E', WTOPC_NO_CHAIN, "LOCATE OF FIXED GROUP %d FAILED", group);
            coruc(fixed_table_ptr); // unhold table
            return(NULL);

         }

         if (already_built) { // if already built, copy the build values into the free entry that was found
                              //   in the Fixed File Control Table in memory and use that entry

            if (debug) printf("\n !!!!! Returned with ALREADY BUILT from check if group already exists !!!!\n");

            tpf_stpoc(TPF_STPOC_ON, stpoc_value);  // allow ourselves to modify key 9 storage gotten by tpf_gsysc

            // copy the values from the build that was alread done ...

            group_info_ptr->group_number = already_fixed_build.group;
            group_info_ptr->begin_ordinal = already_fixed_build.begin_ordinal;
            group_info_ptr->end_ordinal = already_fixed_build.end_ordinal;
            memcpy(group_info_ptr->rec_ID,already_fixed_build.hex_rid,QFNL_RID_SIZE);
            memcpy(group_info_ptr->rec_type,already_fixed_build.rtype,QFNL_RTYPE_SIZE);
            group_info_ptr->record_size = already_fixed_build.record_size;

            // Mark this group as ready to do work ...

            group_info_ptr->group_flag = QFNL_GROUP_STAT_STOPPED;   // mark group stopped, as build now complete
                                                                     // (i.e. no longer in progress)
            fixed_table_ptr->number_of_groups++;  // increment count of groups in memory

            tpf_stpoc(TPF_STPOC_OFF, stpoc_value); // done modifying for now

            coruc(fixed_table_ptr);   // unhold the table

            return(group_info_ptr);   // return this entry whose contents were read from the file system

         } else { // not already built, there is no entry for this group to return

            coruc(fixed_table_ptr);   // unhold the table

            return(NULL);

         }

     if (debug) printf("\n !!!!! We hit end of IF WE ARE NOT DONE logic !!!!\n");
    } // end if we are not done

 }         // end else have table for fixed

 if (debug) printf("\n !!!!! WE reached BOTTOM of LOCATE function !!!! \n");
 return(group_info_ptr);

} // END of function qfnl_locate_fixed_group

int qfnl_parse_fixed_start(struct IPRSE_output *parse,
                             int num_parms)
{

// NOTE: This function parses parameters for the ZTEST FIXED START command, and stores there
//   values into a structure named my_fixed_start.
// This function also performs checks that the generic parser cannot, such as ensuring
//   that when locking is active, the number of reads is >=  the number of writes
// This function returns 0 if an error was detected (and START should not be allowed) or
//                       1 on success (and the caller may continue START processing).

 my_fixed_start.doLocking = TRUE;  // default for ZTEST FIXED START is to do locking
 boolean lockByDefault = TRUE;     //  assume this is the default

 // Set other attributes to FALSE until we parse parameters that specify otherwise

 my_fixed_start.lowPriority = FALSE;  // do not make this instance a low priority ECB unless told otherwise
 my_fixed_start.inheritLow = FALSE;   // do not make child ECBs of this instance low priority unless told otherwise
 my_fixed_start.switchable = FALSE;   // do not make this instance a switchable ECB unless told otherwise
 my_fixed_start.inheritSW = FALSE;    // do not make child ECBs of this instance switchable unless told otherwise
 my_fixed_start.useResClass = FALSE;  // do not make this instance and its children use a resource class unless told otherwise
 my_fixed_start.useUtilClass = FALSE; // do not make this instance and its children use a utilization class unless told otherwise

 memset (my_fixed_start.groupString, '\0', sizeof(my_fixed_start.groupString));  // initialize group string to nulls
 memset (my_fixed_start.utilClass, ' ', sizeof(my_fixed_start.utilClass));  // also initialize UTILCLASS to blanks until we are told to use it,
                                                                            //   so that if we are told to use it and set it later, it will
                                                                            //   automatically be padded with blanks as required
 memset (&my_fixed_start.utilClass[8], '\0', 1);  // also ensure that UTILCLASS is null-terminated

 while ((num_parms > 0) && parse)
  { // while there are more parameters

   if (debug) printMsg(9, 'I', WTOPC_NO_CHAIN, "!!! PARSE PARM IS %s", parse->IPRSE_parameter);

   if (parse && (strncmp(parse->IPRSE_parameter, "Group-",6) == 0))
   { // if Group specified

    // get the group number and save it both as an integer and as a string

    my_fixed_start.group = atoi(parse->IPRSE_value);
    memcpy(my_fixed_start.groupString, parse->IPRSE_value, strlen(parse->IPRSE_value));

    if (my_fixed_start.group <=0) {
      printMsg(44,'E',WTOPC_NO_CHAIN,"GROUP %d NOT STARTED:  GROUP NUMBER MUST BE GREATER THAN 0", my_fixed_start.group);
      return(0);
    }

    //  The parser should prevent an integer greater than 5 digits, but I coded a check anyway just in case

    if (my_fixed_start.group > 99999) {
      printMsg(45,'E',WTOPC_NO_CHAIN,"GROUP %d NOT STARTED:  MAXIMUM GROUP NUMBER IS 99999", my_fixed_start.group);
      return(0);
    }

    if (debug) printMsg(1,'I',WTOPC_NO_CHAIN,"!!! input Group = %d",my_fixed_start.group);

    // generate the file name that will be used to communicate the existence of this group to other processors in the complex

    sprintf(FIXED_file_name,"%s%s%s",QFNL_DIR,QFNL_PFX,my_fixed_start.groupString);

   } // end if Group specified

   if (parse && (strncmp(parse->IPRSE_parameter, "Read-",5) == 0))
   { // if Read count specified

    my_fixed_start.numberOfReads = atoi(parse->IPRSE_value);

    if (debug) printMsg(1,'I',WTOPC_NO_CHAIN,"numberOfReads = %d",my_fixed_start.numberOfReads);

    if ( (my_fixed_start.numberOfReads < 0) ||
         (my_fixed_start.numberOfReads > 999) ) {
      printMsg(46,'E',WTOPC_NO_CHAIN,"GROUP %d NOT STARTED:  READS MUST BE BETWEEN 0 AND 999", my_fixed_start.group);
      return(0);
    }

   } // end if Read count specified

   if (parse && (strncmp(parse->IPRSE_parameter, "Write-",6) == 0))
   { // if Write count specified

    my_fixed_start.numberOfWrites = atoi(parse->IPRSE_value);

    if (debug) printMsg(1,'I',WTOPC_NO_CHAIN,"numberOfWrites = %d",my_fixed_start.numberOfWrites);

    if ( (my_fixed_start.numberOfWrites < 0) ||
         (my_fixed_start.numberOfWrites > 999) ) {
      printMsg(47,'E',WTOPC_NO_CHAIN,"GROUP %d NOT STARTED:  WRITES MUST BE BETWEEN 0 AND 999", my_fixed_start.group);
      return(0);
    }

   } // end if Write count specified

   if (parse && (strncmp(parse->IPRSE_parameter, "SEED-",5) == 0))
   { // if SEED specified

    my_fixed_start.seed = atoi(parse->IPRSE_value);

    if (debug) printMsg(1,'I',WTOPC_NO_CHAIN,"SEED = %d",my_fixed_start.seed);

    if ( (my_fixed_start.seed < 0) ||
         (my_fixed_start.seed > 999) ) {
      printMsg(48,'E',WTOPC_NO_CHAIN,"GROUP %d NOT STARTED:  SEED MUST BE BETWEEN 0 AND 999", my_fixed_start.group);
      return(0);
    }

   } // end if SEED count specified

   if (parse && (strncmp(parse->IPRSE_parameter, "ECBs-",5) == 0))
   { // if ECBs specified

    my_fixed_start.numECBs = atoi(parse->IPRSE_value);

    if (debug) printMsg(1,'I',WTOPC_NO_CHAIN,"numECBs = %d",my_fixed_start.numECBs);

    if ( (my_fixed_start.numECBs < 1) ||
         (my_fixed_start.numECBs > 999) ) {
      printMsg(49,'E',WTOPC_NO_CHAIN,"GROUP %d NOT STARTED:  NUMBER ECBS MUST BE BETWEEN 1 AND 999", my_fixed_start.group);
      return(0);
    }

   } // end if ECBs specified

   if (parse && (strncmp(parse->IPRSE_parameter, "Lock-",5) == 0))
   { // if Lock specified

    char * lock_ptr = NULL;

    lockByDefault = FALSE;    //  the user is not taking the default, but specifying it

    lock_ptr = parse->IPRSE_value;
    if ((*lock_ptr == 'y') || (*lock_ptr == 'Y')) {
       my_fixed_start.doLocking = true;
    }
    else if ((*lock_ptr == 'n') || (*lock_ptr == 'N')) {
       my_fixed_start.doLocking = false;
    }
    else {
       printErr(19,"BAD VALUE SPECIFIED FOR Lock; MUST BE Y or N");
       exit(0);
    }

   } // end if Lock specified

   if (parse && (strncmp(parse->IPRSE_parameter, "LOw-",4) == 0))
   { // if LOw specified

    char * low_ptr = NULL;

    low_ptr = parse->IPRSE_value;
    if ((*low_ptr == 'y') || (*low_ptr == 'Y')) {
       my_fixed_start.lowPriority = true;
    }
    else if ((*low_ptr == 'i') || (*low_ptr == 'I')) {
       my_fixed_start.lowPriority = true;  // inherit means the main ECB for instance has this attribute...
       my_fixed_start.inheritLow = true;   //   and any of its children will inherit this attribute as well
    }
    else {
       printErr(19,"BAD VALUE SPECIFIED FOR LOw: MUST BE Y or I");
       exit(0);
    }

    if (debug) printMsg(1,'I',WTOPC_NO_CHAIN,"VALUE OF my_fixed_start.lowPriority is %s",my_fixed_start.lowPriority ? "TRUE" : "FALSE");
    if (debug) printMsg(1,'I',WTOPC_NO_CHAIN,"VALUE OF my_fixed_start.inheritLow is %s",my_fixed_start.inheritLow ? "TRUE" : "FALSE");
   } // end if LOw specified

   if (parse && (strncmp(parse->IPRSE_parameter, "SWitch-",7) == 0))
   { // if SWitch specified

    char * switch_ptr = NULL;

    switch_ptr = parse->IPRSE_value;
    if ((*switch_ptr == 'y') || (*switch_ptr == 'Y')) {
       my_fixed_start.switchable = true;
    }
    else if ((*switch_ptr == 'i') || (*switch_ptr == 'I')) {
       my_fixed_start.switchable = true;   // inherit means the main ECB for instance has this attribute...
       my_fixed_start.inheritSW = true;    //   and any of its children will inherit this attribute as well
    }
    else {
       printErr(19,"BAD VALUE SPECIFIED FOR SWitch: MUST BE Y or I");
       exit(0);
    }

    if (debug) printMsg(1,'I',WTOPC_NO_CHAIN,"VALUE OF my_fixed_start.switchable is %s",my_fixed_start.switchable ? "TRUE" : "FALSE");
    if (debug) printMsg(1,'I',WTOPC_NO_CHAIN,"VALUE OF my_fixed_start.inheritSW is %s",my_fixed_start.inheritSW ? "TRUE" : "FALSE");
   } // end if SWitch specified

   if (parse && (strncmp(parse->IPRSE_parameter, "RESourceclass-",14) == 0))
   { // if RESourceclass specified

    char * res_ptr = NULL;

    res_ptr = parse->IPRSE_value;
    if ((*res_ptr == 'y') || (*res_ptr == 'Y')) {
       my_fixed_start.useResClass = true;
    }
    else if ((*res_ptr == 'n') || (*res_ptr == 'N')) {
       my_fixed_start.useResClass = false;
    }
    else {
       printErr(19,"BAD VALUE SPECIFIED FOR RESourceclass; MUST BE Y or N");
       exit(0);
    }

    if (debug) printMsg(1,'I',WTOPC_NO_CHAIN,"VALUE OF my_fixed_start.useResClass is %s",my_fixed_start.useResClass ? "TRUE" : "FALSE");

   } // end if RESourceclass specified

   if (parse && (strncmp(parse->IPRSE_parameter, "UTILclass-",10) == 0))
   { // if UTILclass specified

       my_fixed_start.useUtilClass = TRUE; // indicate that this instance will use a utilization class
       memcpy(my_fixed_start.utilClass, parse->IPRSE_value, strlen(parse->IPRSE_value));  // copy the utilization class for later use
       if (debug) printMsg(1,'I',WTOPC_NO_CHAIN,"!!! my_fixed_start.utilClass initially set to %s!!!",my_fixed_start.utilClass);

   } // end if UTILclass specified


   parse = parse->IPRSE_next;
   num_parms--;

  } // end while there are more parameters

 // If the user has not specified a value for LOCK-Y|N, inform them of the default

 if (lockByDefault) {
  printMsg(50,'I',WTOPC_NO_CHAIN,"FOR ATTEMPTED START ON GROUP %d DEFAULT VALUE LOCK-Y ASSUMED",my_fixed_start.group);
 }

 // Do not allow both Read and Write to be zero

 if ( (my_fixed_start.numberOfReads == 0) && (my_fixed_start.numberOfWrites == 0) ) {
  printMsg(51,'E',WTOPC_NO_CHAIN,"START OF GROUP %d FAILED.  READ AND WRITE CANNOT BOTH BE ZERO",
                                   my_fixed_start.group);
  return (0);
 }

 // If Locking is in effect, ensure that Number of Reads is GREATER OR EQUAL to Number of Writes

 if ( my_fixed_start.doLocking &&
      (my_fixed_start.numberOfReads < my_fixed_start.numberOfWrites) ) {
  printMsg(52,'E',WTOPC_NO_CHAIN,"START OF GROUP %d FAILED.  FOR LOCKING THERE MUST BE AT LEAST ONE READ FOR EVERY WRITE",
                                   my_fixed_start.group);
  return (0);
 }

 if (debug) printf("\n!!!!! my_fixed_start.doLocking = %d !!!!!\n", my_fixed_start.doLocking);
 return(1);
}

int qfnl_parse_fixed_status(struct IPRSE_output *parse,
                             int num_parms)
{

// NOTE: This function parses parameters for the ZTEST FIXED STATUS command, and stores there
//   values into a structure named my_fixed_status.

// This function returns 0 if an error was detected (and STATUS should not be allowed) or
//                       1 on success (and the caller may continue STATUS processing).


 memset (my_fixed_status.groupString, '\0', sizeof(my_fixed_status.groupString));  // initialize string to nulls

                            //
 my_fixed_status.group = 0; // assign zero for group number until it is parsed...
                            //   if it remains that way, the caller may interpret the STATUS request
                            //   to be an abbreviated request for all groups in memory
                            //

 while ((num_parms > 0) && parse)
  { // while there are more parameters

   if (debug) printMsg(9, 'I', WTOPC_NO_CHAIN, "!!! PARSE PARM IS %s", parse->IPRSE_parameter);

   if (parse && (strncmp(parse->IPRSE_parameter, "Group-",6) == 0))
   { // if Group specified

    // get the group number and save it both as an integer and as a string

    my_fixed_status.group = atoi(parse->IPRSE_value);
    memcpy(my_fixed_status.groupString, parse->IPRSE_value, strlen(parse->IPRSE_value));

    if (my_fixed_status.group <=0) {
      printMsg(53,'E',WTOPC_NO_CHAIN,"NO STATUS FOR GROUP %d:  GROUP NUMBER MUST BE GREATER THAN 0", my_fixed_status.group);
      return(0);
    }

    //  The parser should prevent an integer greater than 5 digits, but I coded a check anyway just in case

    if (my_fixed_status.group > 99999) {
      printMsg(54,'E',WTOPC_NO_CHAIN,"NO STATUS FOR GROUP %d:  MAXIMUM GROUP NUMBER IS 99999", my_fixed_status.group);
      return(0);
    }

    // generate the file name that will be used to determine the existence of this group in other processors in the complex,
    //    or even from a previous IPL of this processor

    sprintf(FIXED_file_name,"%s%s%s",QFNL_DIR,QFNL_PFX,my_fixed_status.groupString);

   } // end if Group specified

   parse = parse->IPRSE_next;
   num_parms--;

  } // end while there are more parameters

 return(1);  // return successful parse
} // end of function qfnl_parse_fixed_status

static void qfnl_fixed_status(struct IPRSE_output *parse,
                             int num_parms)
{

 char statusString[15];                            // will be used for status display
 memset (statusString,'\0',sizeof(statusString)); // initialize to nulls

 char record_type[QFNL_RTYPE_SIZE+1];
 memset (record_type,'\0',sizeof(record_type));

 // parse the options specified for the ZTEST FFACT FIXED STATUS command...
 if ( qfnl_parse_fixed_status(parse,num_parms) == 0 )
   return;    // return if parser did not like something

 if (my_fixed_status.group == 0) { // if no group specified on STATUS command
    qfnl_fixed_general_status();   //   just give general status for all groups in memory
    return;
 }

 qfnl_fixed_info_Ptr found_fixed_info_ptr = qfnl_locate_fixed_group(my_fixed_status.group, TRUE, TRUE);
 if (found_fixed_info_ptr == NULL) {  // goup not found
   printMsg(55,'E',WTOPC_NO_CHAIN,"NO STATUS FOR GROUP %d: GROUP INFO FROM BUILD NOT FOUND",
            my_fixed_status.group);
   return;
 } else {  // group found

//   printMsg(2,'I',WTOPC_NO_CHAIN, "ADDRESS OF FIXED GROUP %d IS %.16lX",
//              my_fixed_status.group, (long) found_fixed_info_ptr);

   // determine status of the group from its entry in the table

   qfnl_check_fixed_status(statusString, found_fixed_info_ptr);

   // copy the record type from the group

   memcpy (record_type, found_fixed_info_ptr->rec_type, QFNL_RTYPE_SIZE);

   // print out status

   printMsg(56,'I',WTOPC_CHAIN,"STATUS OF FIXED FILE GROUP %d IS %s",
                   my_fixed_status.group, statusString);
   printMsg(56,0,WTOPC_CHAIN,"Additional details follow:");
   printMsg(56,0,WTOPC_CHAIN,"  ");
   printMsg(56,0,WTOPC_CHAIN, "ADDRESS OF FIXED GROUP %d IS %.16lX",
              my_fixed_status.group, (long) found_fixed_info_ptr);
   printMsg(56,0,WTOPC_CHAIN,"  ");
    printMsg(56,0,WTOPC_CHAIN,"RECORD TYPE = %s",
                record_type);
   printMsg(56,0,WTOPC_CHAIN, "RECORD ID (HEX) = %02x%02x",
                found_fixed_info_ptr->rec_ID[0], found_fixed_info_ptr->rec_ID[1]);
   printMsg(56,0,WTOPC_CHAIN,"BEGIN ORDINAL (HEX) = %x",
                found_fixed_info_ptr->begin_ordinal);
   printMsg(56,0,WTOPC_CHAIN,"END ORDINAL (HEX)  = %x",
                found_fixed_info_ptr->end_ordinal);
   printMsg(56,0,WTOPC_CHAIN,"RECORD SIZE = %d",
                found_fixed_info_ptr->record_size);
   printMsg(56,0,WTOPC_CHAIN,"NUMBER OF FINDS = %ld",                           //d26352
                found_fixed_info_ptr->group_stats.readCount);
   printMsg(56,0,WTOPC_CHAIN,"NUMBER OF HOLDS = %ld",                           //d26352
                found_fixed_info_ptr->group_stats.holdCount);
   printMsg(56,0,WTOPC_CHAIN,"NUMBER OF FILES = %ld",                           //d26352
                found_fixed_info_ptr->group_stats.writeCount);
   printMsg(56,0,WTOPC_CHAIN,"NUMBER OF INSTANCES = %ld",                       //d26352
                found_fixed_info_ptr->group_stats.instanceCount);
   printMsg(56,0,WTOPC_CHAIN,"NUMBER OF ECBs CREATED = %lu",                    //d26352
                found_fixed_info_ptr->group_stats.ecbCreateCount);
   printMsg(56,0,WTOPC_CHAIN,"NUMBER OF ECBs EXITED = %lu",                     //d26352
                found_fixed_info_ptr->group_stats.ecbExitCount);
   printMsg(56,0,WTOPC_CHAIN,"NUMBER OF ECBs CURRENTLY ACTIVE = %lu",           //d26352
                found_fixed_info_ptr->group_stats.ecbCount);
   printMsg(56,0,WTOPC_CHAIN,"NUMBER OF ERRORS = %ld",                          //d26352
                found_fixed_info_ptr->group_stats.errCount);
   if (found_fixed_info_ptr->group_stats.errCount > 0) {
     printMsg(56,0,WTOPC_CHAIN,"     BREAKDOWN OF ERRORS:");
     printMsg(56,0,WTOPC_CHAIN,"             MALLOC ERRORS = %ld",              //d26352
                found_fixed_info_ptr->group_stats.mallocErr);
     printMsg(56,0,WTOPC_CHAIN,"             ENVIRONMENTAL ERRORS = %ld",       //d26352
                found_fixed_info_ptr->group_stats.envErr);
     printMsg(56,0,WTOPC_CHAIN,"             I/O ERRORS = %ld",                 //d26352
                found_fixed_info_ptr->group_stats.ioErr);
     printMsg(56,0,WTOPC_CHAIN,"             RECORD CONTENT ERRORS = %ld",      //d26352
                found_fixed_info_ptr->group_stats.contentErr);
   }

   if (found_fixed_info_ptr->group_flag == QFNL_GROUP_STAT_STARTED)
     printMsg(56,0,WTOPC_CHAIN,"LOCKING IS CURRENTLY %s",
                found_fixed_info_ptr->group_stats.lockingActive ? "ON" : "OFF");


   printMsg(56,0,WTOPC_CHAIN_END,"END OF STATUS FOR FIXED GROUP %d", my_fixed_status.group);


 } // end else group found

 return;

} // end of function qfnl_fixed_status

void qfnl_fixed_stop_all_groups(void)
{

// This function will loop through the table of FIXED FILE Groups
//  and mark them all as STOPPED so that all activity on them stops.

  qfnl_fixed_ctl_tbl_Ptr fixed_table_ptr = get_fixed_table();

  if (fixed_table_ptr == NULL) {  // nothing in memory
     printMsg(58,'I',WTOPC_NO_CHAIN,"THERE ARE NO FIXED GROUPS IN MEMORY FOR THIS PROCESSOR");
     return;
  } else { // there could be fixed groups in memory

     int i;
     boolean at_least_one = false; // assume we do not have at least one group until found

     qfnl_fixed_info_Ptr group_info_ptr = fixed_table_ptr->my_fixed;  // point at first fixed group slot in table

     for(i=0; i < QFNL_MAX_FIXED; i++) { // for all items in table

      if ( (group_info_ptr->group_number != 0) &&
           (group_info_ptr->group_flag != QFNL_GROUP_STAT_FREESLOT) ) { // if found an in-use group in the table, set it to STOPPED

          at_least_one = true;  // we found at least one group in the table

          if (group_info_ptr->group_flag == QFNL_GROUP_STAT_BUILDING)
              printMsg(59,'W',WTOPC_NO_CHAIN,"FIXED GROUP %d BUILD IN PROGRESS: STOP ALLOWED BUT MIGHT TERMINATE BUILD",
                    group_info_ptr->group_number);

          tpf_stpoc(TPF_STPOC_ON, stpoc_value);  // allow ourselves to modify key 9 storage gotten by tpf_gsysc
          group_info_ptr->group_flag = QFNL_GROUP_STAT_STOPPED;   // set indicator that the group is now STOPPED
          tpf_stpoc(TPF_STPOC_OFF, stpoc_value); // done modifying for now

          if (group_info_ptr->group_stats.instanceCount == 0) {
              printMsg(57,'I',WTOPC_NO_CHAIN,"FIXED GROUP %d STOPPED", group_info_ptr->group_number);
          } else {
              printMsg(61,'I',WTOPC_NO_CHAIN,"FIXED GROUP %d STOPPING", group_info_ptr->group_number);
          }

      } // end if found an in-use group in the table

      group_info_ptr++; // advance to next slot in table

     } // end for all items in table

     // if there wasn't at least one group in memory that we could stop, tell the user

     if (!at_least_one) {
         printMsg(58,'I',WTOPC_NO_CHAIN,"THERE ARE NO FIXED GROUPS IN MEMORY FOR THIS PROCESSOR");
     }

  } // end else there could be fixed groups in memory

  return;  // return to caller

} // qfnl_fixed_stop_all_groups

void qfnl_fixed_general_status(void)
{

// This function will loop through the table of FIXED FILE Groups
//  and display general driver status information for each.

  qfnl_fixed_ctl_tbl_Ptr fixed_table_ptr = get_fixed_table();

  if (fixed_table_ptr == NULL) {  // nothing in memory
     printMsg(58,'I',WTOPC_NO_CHAIN,"THERE ARE NO FIXED GROUPS IN MEMORY FOR THIS PROCESSOR");
     return;
  } else { // there could be fixed groups in memory

     int i;
     boolean at_least_one = false; // assume we do not have at least one group until found

     char statusString[15];                            // will be used for status display
     char record_type[QFNL_RTYPE_SIZE+1];              // will be used to display record type

     printMsg(60,'I',WTOPC_CHAIN,"SUMMARY FOR ALL FIXED GROUPS IN MEMORY FOR THIS PROCESSOR"); // print out quick status header
     printMsg(60,0,WTOPC_CHAIN,
       "GROUP REC TYPE    STATUS   FINDS*           FILES*           ERRORS");

     qfnl_fixed_info_Ptr group_info_ptr = fixed_table_ptr->my_fixed;  // point at first fixed group slot in table

     for(i=0; i < QFNL_MAX_FIXED; i++) { // for all items in table

      memset (statusString,'\0',sizeof(statusString)); // initialize to nulls

      memset (record_type,' ',sizeof(record_type)-1);
      record_type[QFNL_RTYPE_SIZE] = '\0';

      if ( (group_info_ptr->group_number != 0) &&
           (group_info_ptr->group_flag != QFNL_GROUP_STAT_FREESLOT) ) { // if found an in-use group in the table, give its status

          at_least_one = true;  // we found at least one group to report status on

          qfnl_check_fixed_status(statusString, group_info_ptr);  // get its actual status to include in display

          memcpy(record_type, group_info_ptr->rec_type, strlen(group_info_ptr->rec_type));  // get record type for display

          printMsg(60,0,WTOPC_CHAIN,

              "%-5d %s   %s  %.16lX %.16lX %lu",                                //d26488
              group_info_ptr->group_number,
              record_type,
              statusString,
              group_info_ptr->group_stats.readCount,
              group_info_ptr->group_stats.writeCount,
              group_info_ptr->group_stats.errCount);

      } // end if found an in-use group in the table

      group_info_ptr++; // advance to next slot in table

     } // end for all items in table

     // if at least one item was found in the table that we could report status on, put this message to indicate that counts were in hex
     //    otherwise put out a line that says there are no groups to report status on

     if (at_least_one) {
         printMsg(60,0,WTOPC_CHAIN," * = ALL COUNTS WITH STAR (*) ARE IN HEX, ALL OTHER NUMBERS DECIMAL");
         printMsg(60,0,WTOPC_CHAIN,"For Additional Info enter ZTEST FFACT STATUS GROUP-n for each group number n");
     } else
         printMsg(60,0,WTOPC_CHAIN,"THERE ARE NO IN-USE GROUPS IN MEMORY FOR THIS PROCESSOR TO REPORT STATUS ON");

     printMsg(60,0,WTOPC_CHAIN_END,"END OF SUMMARY STATUS DISPLAY FOR FIXED GROUPS");

  } // end else there could be fixed groups in memory

  return;  // return to caller

} // end of qfnl_fixed_general_status

void qfnl_check_fixed_status(char * statusString, qfnl_fixed_info_Ptr group_info_ptr)
{

// This function will store the status of the group referenced by group_info_ptr
//   at the location pointed to by statusString

 switch (group_info_ptr->group_flag)
  {
     case QFNL_GROUP_STAT_STARTED:
       memcpy(statusString, "STARTED", strlen("STARTED"));
       break;

     case QFNL_GROUP_STAT_STOPPED:
       memcpy(statusString, "STOPPED", strlen("STOPPED"));
       break;

     case QFNL_GROUP_STAT_BUILDING:
       memcpy(statusString, "BUILDING", strlen("BUILDING"));
       break;

     case QFNL_GROUP_STAT_FREESLOT:
       memcpy(statusString, "NOT IN USE", strlen("NOT IN USE"));
       break;

     default:
       memcpy(statusString, "UNKNOWN", strlen("UNKNOWN"));
       break;

  } // end switch

 return;

} // end function qfnl_check_fixed_status

static void qfnl_fixed_stop(struct IPRSE_output *parse,
                             int num_parms)
{

 // parse the ZTEST FFACT FIXED STOP command...

 if ( qfnl_parse_fixed_stop(parse,num_parms) == 0 )
   return;    // return if parser did not like something

 // if group was set to zero it means the user did not
 //   specify a group so we will stop activity for all
 //   fixed file groups

 if (my_fixed_stop.group == 0) {
   qfnl_fixed_stop_all_groups();
   return;
 }

 // attempt to locate the group being stopped

 qfnl_fixed_info_Ptr found_fixed_info_ptr = qfnl_locate_fixed_group(my_fixed_stop.group, FALSE, FALSE);
 if (found_fixed_info_ptr == NULL) {  // goup not found

   printMsg(64,'E',WTOPC_NO_CHAIN,"CANNOT STOP FIXED GROUP %d: GROUP NOT FOUND IN MEMORY",
            my_fixed_stop.group);
   return;

 } else {  // group found

   // if the group has the indication on that a BUILD is in progress, we will allow you to set
   //    the group to being STOPPED state, but that does not stop any build that may still be
   //    ongoing.  (NOTE: When the build is in done it should set the indicator back to STOPPED
   //    again.)

   if (found_fixed_info_ptr->group_flag == QFNL_GROUP_STAT_BUILDING)
     printMsg(59,'W',WTOPC_NO_CHAIN,"FIXED GROUP %d BUILD IN PROGRESS: STOP ALLOWED BUT MIGHT TERMINATE BUILD",
              my_fixed_stop.group);

   tpf_stpoc(TPF_STPOC_ON, stpoc_value);  // allow ourselves to modify key 9 storage gotten by tpf_gsysc
   found_fixed_info_ptr->group_flag = QFNL_GROUP_STAT_STOPPED;   // set indicator that the group is now STOPPED
   tpf_stpoc(TPF_STPOC_OFF, stpoc_value); // done modifying for now

   if (found_fixed_info_ptr->group_stats.instanceCount == 0) {
     printMsg(57,'I',WTOPC_NO_CHAIN,"FIXED GROUP %d STOPPED", my_fixed_start.group);
   } else {
     printMsg(61,'I',WTOPC_NO_CHAIN,"FIXED GROUP %d STOPPING", my_fixed_stop.group);
   }

 } // end else group found

} // end function qfnl_fixed_stop


int qfnl_parse_fixed_stop(struct IPRSE_output *parse,
                             int num_parms)
{

// NOTE: This function parses parameters for the ZTEST FIXED STOP command, and stores there
//   values into a structure named my_fixed_stop.

// This function returns 0 if an error was detected (and STOP should not be allowed) or
//                       1 on success (and the caller may continue STOP processing).


 my_fixed_stop.group = 0; // assign zero for group number until it is parsed...
                          // NOTE: If it remains zero we will interpret this as a request
                          //  to stop all fixed activity on this processor (all fixed groups)

 while ((num_parms > 0) && parse)
  { // while there are more parameters

   if (debug) printMsg(9, 'I', WTOPC_NO_CHAIN, "!!! PARSE PARM IS %s", parse->IPRSE_parameter);

   if (parse && (strncmp(parse->IPRSE_parameter, "Group-",6) == 0))
   { // if Group specified

    // get the group number and save it both as an integer and as a string

    my_fixed_stop.group = atoi(parse->IPRSE_value);

    if (my_fixed_stop.group <=0) {
      printMsg(62,'E',WTOPC_NO_CHAIN,"STOP COMMAND FOR GROUP %d REJECTED:  GROUP NUMBER MUST BE GREATER THAN 0", my_fixed_stop.group);
      return(0);
    }

    //  The parser should prevent an integer greater than 5 digits, but I coded a check anyway just in case

    if (my_fixed_status.group > 99999) {
      printMsg(63,'E',WTOPC_NO_CHAIN,"STOP COMMAND FOR GROUP %d REJECTED:  MAXIMUM GROUP NUMBER IS 99999", my_fixed_stop.group);
      return(0);
    }

   } // end if Group specified

   parse = parse->IPRSE_next;
   num_parms--;

  } // end while there are more parameters

 return(1);  // return successful parse
} // end of function qfnl_parse_fixed_stop

int qfnl_parse_fixed_kill(struct IPRSE_output *parse,
                             int num_parms)
{

// NOTE: This function parses parameters for the ZTEST FIXED KILL command, and stores there
//   values into a structure named my_fixed_kill.

// This function returns 0 if an error was detected (and KILL should not be allowed) or
//                       1 on success (and the caller may continue KILL processing).


 memset (my_fixed_kill.groupString, '\0', sizeof(my_fixed_kill.groupString));  // initialize string to nulls

 my_fixed_kill.group = 0; // assign zero for group number until it is parsed...

 while ((num_parms > 0) && parse)
  { // while there are more parameters

   if (debug) printMsg(9, 'I', WTOPC_NO_CHAIN, "!!! PARSE PARM IS %s", parse->IPRSE_parameter);

   if (parse && (strncmp(parse->IPRSE_parameter, "Group-",6) == 0))
   { // if Group specified

    // get the group number and save it both as an integer and as a string

    my_fixed_kill.group = atoi(parse->IPRSE_value);
    memcpy(my_fixed_kill.groupString, parse->IPRSE_value, strlen(parse->IPRSE_value));

    if (my_fixed_kill.group <=0) {
      printMsg(67,'E',WTOPC_NO_CHAIN,"CANNOT KILL GROUP %d:  GROUP NUMBER MUST BE GREATER THAN 0", my_fixed_kill.group);
      return(0);
    }

    //  The parser should prevent an integer greater than 5 digits, but I coded a check anyway just in case

    if (my_fixed_kill.group > 99999) {
      printMsg(69,'E',WTOPC_NO_CHAIN,"CANNOT KILL GROUP %d:  MAXIMUM GROUP NUMBER IS 99999", my_fixed_kill.group);
      return(0);
    }

    // generate the file name that will be used to determine the existence of this group in other processors in the complex,
    //    or even from a previous IPL of this processor

    sprintf(FIXED_file_name,"%s%s%s",QFNL_DIR,QFNL_PFX,my_fixed_kill.groupString);

   } // end if Group specified

   parse = parse->IPRSE_next;
   num_parms--;

  } // end while there are more parameters

 return(1);  // return successful parse
} // end of function qfnl_parse_fixed_kill

static void qfnl_fixed_kill(struct IPRSE_output *parse,
                             int num_parms)
{

 boolean killed_something = false; // assume nothing was killed for this group until something is removed for it

 // parse the ZTEST FFACT FIXED KILL command...

 if ( qfnl_parse_fixed_kill(parse,num_parms) == 0 )
   return;    // return if parser did not like something

 // attempt to locate the group being killed in memory

 qfnl_fixed_info_Ptr found_fixed_info_ptr = qfnl_locate_fixed_group(my_fixed_kill.group, FALSE, FALSE);
 if (found_fixed_info_ptr == NULL) {  // goup not found

   printMsg(71,'I',WTOPC_NO_CHAIN,"FIXED GROUP %d NOT FOUND IN MEMORY",
            my_fixed_kill.group);

 } else {  // group found in memory, free its slot in the table as long as it is in STOPPED state

   if (found_fixed_info_ptr->group_flag != QFNL_GROUP_STAT_STOPPED) { // if group in memory not STOPPED, do not allow KILL

      printMsg(72,'E',WTOPC_NO_CHAIN,"YOU MUST STOP GROUP %d ON THIS PROCESSOR BEFORE YOU CAN KILL IT HERE",
               my_fixed_kill.group);
      return;
   } // end if group in memory not STOPPED

   // The fixed group is STOPPED, so we have now will set its slot to FREE and clean up the slot in the table to
   //    be all zeroes

   tpf_stpoc(TPF_STPOC_ON, stpoc_value);  // allow modify key 9
   found_fixed_info_ptr->group_flag = QFNL_GROUP_STAT_FREESLOT;
   memset(found_fixed_info_ptr, 0x00, sizeof(qfnl_fixed_info));
   tpf_stpoc(TPF_STPOC_OFF, stpoc_value);  // done with modification

   // Decrement count of fixed file groups that the driver has in memory for this processor

   qfnl_fixed_ctl_tbl_Ptr  fixed_table_ptr = get_fixed_table();

   if (fixed_table_ptr == NULL) {
      char errMsg[120] = "";                                                    //d26352
      sprintf(errMsg,"ON KILL OF FIXED GROUP %d, FAILURE TO LOCATE FIXED GROUP TABLE TO DECREMENT COUNT OF FIXED GROUPS IN MEMORY",my_fixed_kill.group);
      printMsg(92, 'E', WTOPC_NO_CHAIN, errMsg);
   } 
   else {                                                                       //d26441
      tpf_stpoc(TPF_STPOC_ON, stpoc_value);  // allow modify key 9              //d26441
      fixed_table_ptr->number_of_groups--;
      tpf_stpoc(TPF_STPOC_OFF, stpoc_value);  // done with modification         //d26441
   }                                                                            //d26441

   killed_something = true;  // we killed something for this group

   // Inform the user that memory has been cleared for this group

   printMsg(74,'I',WTOPC_NO_CHAIN,"MEMORY FOR GROUP %d HAS BEEN CLEARED", my_fixed_kill.group);

 } // end group found in memory

 // Remove any file from the file system that represents this group, so it cannot be started following an IPL

 // NOTE: I decided to make the driver silent if it could not delete the file because you have to manually kill a group
 //    on all processors in the complex and the file will only get deleted the first time so don't make not
 //    being able to delete the file an error.

 // if ( remove(FIXED_file_name) != 0 )
 //   printMsg(75,'E',WTOPC_NO_CHAIN,"ERROR DELETING FILE %s",FIXED_file_name);
 // else {

 if ( remove(FIXED_file_name) == 0 ) {
   printMsg(76,'I',WTOPC_NO_CHAIN,"FILE %s DELETED",FIXED_file_name);
   killed_something = true;  // we killed something for this group
 }

 if (killed_something)
   printMsg(66,'I',WTOPC_NO_CHAIN,"GROUP %d KILLED ON THIS PROCESSOR", my_fixed_kill.group);
 else
   printMsg(77,'I',WTOPC_NO_CHAIN,"NOTHING TO REMOVE FOR GROUP %d", my_fixed_kill.group);

 return;

} // end qfnl_fixed_kill

int qfnl_parse_fixed_clear(struct IPRSE_output *parse,
                             int num_parms)
{

// NOTE: This function parses parameters for the ZTEST FIXED CLEAR command, and stores there
//   values into a structure named my_fixed_clear.

// This function returns 0 if an error was detected (and CLEAR should not be allowed) or
//                       1 on success (and the caller may continue CLEAR processing).

 my_fixed_clear.group = 0; // assign zero for group number until it is parsed...

 while ((num_parms > 0) && parse)
  { // while there are more parameters

   if (debug) printMsg(9, 'I', WTOPC_NO_CHAIN, "!!! PARSE PARM IS %s", parse->IPRSE_parameter);

   if (parse && (strncmp(parse->IPRSE_parameter, "Group-",6) == 0))
   { // if Group specified

    // get the group number and save it both as an integer and as a string

    my_fixed_clear.group = atoi(parse->IPRSE_value);

    if (my_fixed_clear.group <=0) {
      printMsg(68,'E',WTOPC_NO_CHAIN,"CANNOT CLEAR GROUP %d:  GROUP NUMBER MUST BE GREATER THAN 0", my_fixed_clear.group);
      return(0);
    }

    //  The parser should prevent an integer greater than 5 digits, but I coded a check anyway just in case

    if (my_fixed_clear.group > 99999) {
      printMsg(70,'E',WTOPC_NO_CHAIN,"CANNOT CLEAR GROUP %d:  MAXIMUM GROUP NUMBER IS 99999", my_fixed_clear.group);
      return(0);
    }

   } // end if Group specified

   parse = parse->IPRSE_next;
   num_parms--;

  } // end while there are more parameters

 return(1);  // return successful parse
} // end of function qfnl_parse_fixed_clear

static void qfnl_fixed_clear(struct IPRSE_output *parse,
                             int num_parms)
{

 boolean cleared_something = false; // assume nothing was cleared for this group until something is removed for it

 // parse the ZTEST FFACT FIXED CLEAR command...

 if ( qfnl_parse_fixed_clear(parse,num_parms) == 0 )
   return;    // return if parser did not like something

 // attempt to locate the group whose counts are being cleared in memory

 qfnl_fixed_info_Ptr found_fixed_info_ptr = qfnl_locate_fixed_group(my_fixed_clear.group, FALSE, FALSE);
 if (found_fixed_info_ptr == NULL) {  // goup not found

   printMsg(71,'I',WTOPC_NO_CHAIN,"FIXED GROUP %d NOT FOUND IN MEMORY",
            my_fixed_clear.group);

 } else {  // group found in memory, free its slot in the table as long as it is in STOPPED state

   if (found_fixed_info_ptr->group_flag != QFNL_GROUP_STAT_STOPPED) { // if group in memory not STOPPED, do not allow CLEAR

      printMsg(73,'E',WTOPC_NO_CHAIN,"YOU MUST STOP GROUP %d ON THIS PROCESSOR BEFORE YOU CAN CLEAR ITS COUNTERS",
               my_fixed_clear.group);
      return;
   } // end if group in memory not STOPPED

   // The fixed group is STOPPED, so we have now will clear its counters

   tpf_stpoc(TPF_STPOC_ON, stpoc_value);  // allow modify key 9
   memset(&(found_fixed_info_ptr->group_stats), 0x00, sizeof(qfnl_group_stats));
   tpf_stpoc(TPF_STPOC_OFF, stpoc_value);  // done with modification

   cleared_something = true;  // we cleared counters for this group

 } // end group found in memory

 if (cleared_something)
   printMsg(65,'I',WTOPC_NO_CHAIN,"COUNTS FOR GROUP %d CLEARED ON THIS PROCESSOR", my_fixed_clear.group);
 else
   printMsg(78,'E',WTOPC_NO_CHAIN,"NO COUNTS TO CLEAR FOR GROUP %d ON THIS PROCESSOR", my_fixed_clear.group);

 return;

} // end qfnl_fixed_clear

int qfnl_parse_pool_build(struct IPRSE_output *parse,
                             int num_parms)
{

//
// This function parses the build command and if an error is found it returns 0, otherwise
//   it returns 1.
//
// NOTE: This function parses the input parameters and uses them to assign values to
//   the my_pool_build structure that represents the pool group build that is being
//   requested.

  my_pool_build.group = 0;
  my_pool_build.number_records = 0;
  memset (my_pool_build.hex_rid,'\0',sizeof(my_pool_build.hex_rid));
  my_pool_build.getfc_rid = 0;
  memset (my_pool_build.stpool,' ',sizeof(my_pool_build.stpool));
  char rid_zeroes[QFNL_RID_SIZE];  // used to check if all zeroes specified for record ID
  memset (rid_zeroes, 0x00, QFNL_RID_SIZE);

  unsigned int input_rid_length = 0;
  char input_rid[5];
  memset (input_rid,'\0',sizeof(input_rid));
  memset (my_pool_build.groupString, '\0', sizeof(my_pool_build.groupString));

  if (debug) printf("\n!! Hooray!! Recognized POOL BUILD !!\n");
  /******************************************************************/
  /* Loop through all CREATE parameters                             */
  /******************************************************************/

  parse = parse->IPRSE_next;
  num_parms--;

  while ((num_parms > 0) && parse)
  { // while there are more parameters

   if (debug) printMsg(9, 'I', WTOPC_NO_CHAIN, "!!! PARSE PARM IS %s", parse->IPRSE_parameter);

   if (parse && (strncmp(parse->IPRSE_parameter, "Group-",6) == 0))
   { // if Group specified

    // get the group number and save it both as an integer and as a string

    my_pool_build.group = atoi(parse->IPRSE_value);
    memcpy(my_pool_build.groupString, parse->IPRSE_value, strlen(parse->IPRSE_value));

    if (my_pool_build.group <=0) {
      printMsg(37,'E',WTOPC_NO_CHAIN,"GROUP %d NOT DEFINED:  GROUP NUMBER MUST BE GREATER THAN 0", my_pool_build.group);
      return(0);
    }

    //  The parser should prevent an integer greater than 5 digits, but I coded a check anyway just in case

    if (my_pool_build.group > 99999) {
      printMsg(37,'E',WTOPC_NO_CHAIN,"GROUP %d NOT DEFINED:  MAXIMUM GROUP NUMBER IS 99999", my_pool_build.group);
      return(0);
    }

    if (debug) printMsg(1,'I',WTOPC_NO_CHAIN,"!!! input Group = %d",my_pool_build.group);
    if (debug) printMsg(1,'I',WTOPC_NO_CHAIN,"!!! input Group String = %s",my_pool_build.groupString);

   } // end if Group specified


   if (parse && (strncmp(parse->IPRSE_parameter, "RID-",4) == 0))
   { // if RID specified

    input_rid_length = convert_input_data_to_hex(parse->IPRSE_value,
                                        input_rid);
    if (debug) printMsg(1,'I',WTOPC_NO_CHAIN,"!!! input RID length = %d", input_rid_length);
    memcpy(my_pool_build.hex_rid,input_rid,input_rid_length);
    if (debug) printMsg(1,'I',WTOPC_NO_CHAIN,"!!! input RID in character = %s", my_pool_build.hex_rid);
    if (debug) {
       unsigned short hex_rid;
       memcpy(&hex_rid, my_pool_build.hex_rid, sizeof(hex_rid));
       printMsg(1,'I',WTOPC_NO_CHAIN,"!!! input RID in hex = %x", hex_rid);
    }
    if ( memcmp(my_pool_build.hex_rid, rid_zeroes, QFNL_RID_SIZE) == 0 ) {  // if RID specified is all zeroes
      printMsg(81,'E',WTOPC_NO_CHAIN,"RECORD ID ALL ZEROES NOT ALLOWED WHEN BUILDING A GROUP");
      return(0);
    }

   } // end if RID specified

   if (parse && (strncmp(parse->IPRSE_parameter, "Num-",4) == 0))
   { // if Num specified

    // get input value number of records for this group

    my_pool_build.number_records = atoi(parse->IPRSE_value);

    if (debug) printMsg(1,'I',WTOPC_NO_CHAIN,"!!! NUMBER OF RECORDS IN HEX = %x IN DECIMAL = %d !!!",
                 my_pool_build.number_records, my_pool_build.number_records);

    if (my_pool_build.number_records == 0) { // if zero records requested
      printMsg(82,'E',WTOPC_NO_CHAIN,"YOU CANNOT HAVE A GROUP WITH ZERO RECORDS");
      return(0);
    }

    if (my_pool_build.number_records > 999999) { // if more than 999999 records requested
      printMsg(103,'E',WTOPC_NO_CHAIN,"YOU CANNOT HAVE A POOL GROUP WITH MORE THAN 999,999 RECORDS");
      return(0);
    }

   } // end if Num specified

   if (parse && (strncmp(parse->IPRSE_parameter, "Expire-",7) == 0))
   { // if Expire specified

    // get input value for number of minutes after which pools expire
    int expiry_in_minutes = atoi(parse->IPRSE_value);

    if (expiry_in_minutes > 24*60) { // if number of minutes for expiry exceeds maximum
      printMsg(83,'E',WTOPC_NO_CHAIN,"MAXIMUM VALUE FOR EXPIRE = 1440 (NUMBER OF MINUTES IN 24 HOURS)");
      return(0);
    }

    my_pool_build.expiry = expiry_in_minutes * 60;  // store expiry for group in seconds

    if (debug) printMsg(1,'I',WTOPC_NO_CHAIN,"!!! VALUE FOR EXPIRY IN DECIMAL SECONDS IS = %d !!!",
                 my_pool_build.expiry);

   } // end if Expire specified

   if (parse && (strncmp(parse->IPRSE_parameter, "STPool-",7) == 0))
   { // if STPool specified

    int len; // length of STPool value
    char * toUpperValue; // Point to value to convert to upper case

    len = strlen(parse->IPRSE_value);
    toUpperValue = parse->IPRSE_value;

    int i;
    for(i=0; i<len; i++, toUpperValue++) {
      my_pool_build.stpool[i] = toupper(*toUpperValue);
    }
    my_pool_build.stpool[i] = '\0';

    if (debug) printMsg(1,'I',WTOPC_NO_CHAIN,"!!! VALUE FOR STPOOL = %s !!!",
                 my_pool_build.stpool);

    // Validate that STPOOL specified is valid

    if (memcmp(my_pool_build.stpool, "SST", 3) == 0) { // if SST requested

      // set the correct record ID for this section name based on the device type
      switch (my_pool_build.stpool[3])
      {
        case 'A':
         my_pool_build.getfc_rid = ridSSTA;
         break;

        case 'B':
         my_pool_build.getfc_rid = ridSSTB;
         break;

        case 'C':
         my_pool_build.getfc_rid = ridSSTC;
         break;

        case 'D':
         my_pool_build.getfc_rid = ridSSTD;
         break;

        default:
         printMsg(84,'E',WTOPC_NO_CHAIN,"INVALID DEVICE %c TYPE SPECIFIED:  MUST BE A, B, C, or D)",
                     my_pool_build.stpool[3]);
         return(0);
      }

    // end if SST requested
    } else if (memcmp(my_pool_build.stpool, "LST", 3) == 0) { // if LST requested

      // set the correct record ID for this section name based on the device type
      switch (my_pool_build.stpool[3])
      {
        case 'A':
         my_pool_build.getfc_rid = ridLSTA;
         break;

        case 'B':
         my_pool_build.getfc_rid = ridLSTB;
         break;

        case 'C':
         my_pool_build.getfc_rid = ridLSTC;
         break;

        case 'D':
         my_pool_build.getfc_rid = ridLSTD;
         break;

        default:
         printMsg(84,'E',WTOPC_NO_CHAIN,"INVALID DEVICE %c TYPE SPECIFIED:  MUST BE A, B, C, or D)",
                     my_pool_build.stpool[3]);
         return(0);
      }

    // end if LST requested
    } else if (memcmp(my_pool_build.stpool, "4ST", 3) == 0) { // if 4ST requested

      // set the correct record ID for this section name based on the device type
      switch (my_pool_build.stpool[3])
      {
        case 'A':
         my_pool_build.getfc_rid = rid4STA;
         break;

        case 'B':
         my_pool_build.getfc_rid = rid4STB;
         break;

        case 'C':
         my_pool_build.getfc_rid = rid4STC;
         break;

        case 'D':
         my_pool_build.getfc_rid = rid4STD;
         break;

        default:
         printMsg(84,'E',WTOPC_NO_CHAIN,"INVALID DEVICE %c TYPE SPECIFIED:  MUST BE A, B, C, or D)",
                     my_pool_build.stpool[3]);
         return(0);
      }
    // end if 4ST requested

    } else if (memcmp(my_pool_build.stpool, "4S6", 3) == 0) { // if 4S6 requested

      // set the correct record ID for this section name based on the device type
      switch (my_pool_build.stpool[3])
      {
        case 'A':
         my_pool_build.getfc_rid = rid4S6A;
         break;

        case 'B':
         my_pool_build.getfc_rid = rid4S6B;
         break;

        case 'C':
         my_pool_build.getfc_rid = rid4S6C;
         break;

        case 'D':
         my_pool_build.getfc_rid = rid4S6D;
         break;

        default:
         printMsg(84,'E',WTOPC_NO_CHAIN,"INVALID DEVICE %c TYPE SPECIFIED:  MUST BE A, B, C, or D)",
                     my_pool_build.stpool[3]);
         return(0);
      }

    // end if 4S6 requested
    } else { // STPool requested is not valid

      printMsg(85,'E',WTOPC_NO_CHAIN,
               "%c IS NOT A VALID SHORT TERM POOL SECTION SUPPORTED BY THIS DRIVER",    //d26352
                  my_pool_build.stpool[3]);
      return(0);

    } // end else STPool requested is not valid

    if (debug) printMsg(1,'I',WTOPC_NO_CHAIN,"!!! VALUE FOR getfc_rid = %x !!!",
                 my_pool_build.getfc_rid);

   } // end if STPool specified

   parse = parse->IPRSE_next;
   num_parms--;

  } // end while there are more parameters


 // Even though the CPU ID that does the build is not an explicit parameter, determine the
 //   CPU ID on which the build command was entered and save it as one of the build
 //   parameters.  Currently we do not use this information for builds of pool groups,
 //   and require each group to be built on each processor for pools.  But save the CPU ID anyway
 //   for now just in case it is needed in the future..
 //  (NOTE: Right now, the CPU ID is only used for fixed file groups, not pool file groups.)

 my_pool_build.cpuID = thisCPUid();

 return(1);

}


/* ---------------------------------------------------------------------------- Beginning of qfnl_pool_build   -------------------------------------------------- */

static void qfnl_pool_build(struct IPRSE_output *parse,
                             int num_parms)

{ // beginning of qfnl_pool_build

 int i = 0;  // index

 qfnl_pool_ctl_tbl_Ptr  pool_table_ptr = NULL;

 if (debug) printf("\n!! Hooray!! Recognized POOL BUILD !!\n");

 // parse the options specified for the ZTEST FFACT POOL BUILD command...
 if ( qfnl_parse_pool_build(parse,num_parms) == 0 )
   exit(0);   // exit if parse function did not like something

 /*******************************************************************/
 /*   Make sure POOLS are available for GETFC                       */
 /*******************************************************************/

 char *state_ptr = (char*)cinfc_fast_ss(CINFC_CMMOPS, ecbptr()->ce1dbi);
 if (*state_ptr & GFS_INACTIVE) { // GFS_INACTIVE - BIT7 in CMMOPS
 // GFS is not active - error
   printMsg(99,'E',WTOPC_CHAIN,"POOLS MUST BE ACTIVE FOR ZTEST FFACT POOL BUILD.");
   printMsg(99,0,WTOPC_CHAIN,"       PLEASE CHECK SUBSYSTEM STATE AND IF ");
   printMsg(99,0,WTOPC_CHAIN_END,"       NECESSARY ENTER ZPOOL 1052 UP.");
   exit(0);
 }
 pool_table_ptr = get_pool_table();

 if (pool_table_ptr == NULL) {
   printMsg(17,'E',WTOPC_NO_CHAIN,"CANNOT FIND OR ALLOCATE POOL FILE CONTROL TABLE - BUILD FAILED");
   return;
 } else {  // have table for pool

   printMsg(3,'I',WTOPC_NO_CHAIN,"ADDRESS OF POOL CONTROL TABLE IS %.16lX", (long) pool_table_ptr);

/*----------------------------------------------------------------------------------------------------*/
    // Before we add a group number for the current group we are building to our table, make sure
    //  it is not already there.
    //  If it is already there and is active, do not perform another build but inform the user that if they
    //  wish to do a build, they must either specify another group number or kill the current group
    //  number so that it can be re-used.

    qfnl_pool_info * group_info_ptr = pool_table_ptr->my_pool;


    // Lock the table so no other ECB on the same processor adds or deletes an element from under us

    corhc(pool_table_ptr); // hold table before search

    for(i=0; (i < QFNL_MAX_POOL); i++) { // for all items in table until we find a match
      if ( (my_pool_build.group == group_info_ptr->group_number) &&
           (group_info_ptr->group_flag != QFNL_GROUP_STAT_FREESLOT) )
      {

        printMsg(21,'E',WTOPC_NO_CHAIN,"POOL GROUP %d HAS ALREADY BEEN BUILT.  USE A DIFFERENT GROUP NUMBER OR KILL POOL GROUP %d",
                   my_pool_build.group, my_pool_build.group);

        // Unhold the table of pool groups so other ECBs can get at it
        coruc(pool_table_ptr); // unhold table before exit

        exit(0);

      } else

        group_info_ptr++;  // advance to next slot in my_pool array

    } // end for all items in table

    // Now that we know the group is not already in our table, we can add it there...

    // Make sure our table is big enough to keep track of this group

    if (pool_table_ptr->number_of_groups >= QFNL_MAX_POOL) {
        printMsg(22,'E',WTOPC_NO_CHAIN,"WE CANNOT KEEP TRACK OF MORE THAN %d POOL FILE GROUPS", pool_table_ptr->number_of_groups);
        printMsg(23,'I',WTOPC_NO_CHAIN,"FOR MORE POOL FILE GROUPS, CHANGE QFNL_MAX_POOL IN qfnl.h, REBUILD,RELOAD QFNL and re-IPL");

        // Unhold the table of pool groups so other ECBs can get at it
        coruc(pool_table_ptr); // unhold table before exit

        exit(0);
    }

    // Look for an item in the table that is not in use

    boolean found = false;
    group_info_ptr = pool_table_ptr->my_pool;
    for(i=0; (i < QFNL_MAX_POOL)&&(!found); i++) { // for all items in table until we find a free item
        if (group_info_ptr->group_flag == QFNL_GROUP_STAT_FREESLOT) {
           found = true;
        }

        if (!found) group_info_ptr++;  // only if not found, advance to next slot in my_pool array

    } // end for all items in table until we find a free item

    if (!found) {
         printMsg(24,'E',WTOPC_NO_CHAIN,"LOGIC ERROR: NO FREE ITEM IN POOL FILE GROUP TABLE OF %d ENTRIES FOUND", QFNL_MAX_POOL);

         // Unhold the table of pool groups so other ECBs can get at it
         coruc(pool_table_ptr); // unhold table before exit

         exit(0);
    }

    // A free item in the table was found.

    tpf_stpoc(TPF_STPOC_ON, stpoc_value);  // allow ourselves to modify key 9 storage gotten by tpf_gsysc

    group_info_ptr->group_flag = QFNL_GROUP_STAT_BUILDING;  // build is now in progress for this group
    group_info_ptr->group_number = my_pool_build.group;     // ... copy in values for new build request ...
    group_info_ptr->number_records = my_pool_build.number_records;
    group_info_ptr->expiry = my_pool_build.expiry;       // REMEMBER: expiry is converted to seconds
    memcpy(group_info_ptr->hex_rid,my_pool_build.hex_rid,QFNL_RID_SIZE);
    group_info_ptr->handle.getfc_rid = my_pool_build.getfc_rid;
    memcpy(group_info_ptr->stpool,my_pool_build.stpool,QFNL_STPOOL_SIZE);
    group_info_ptr->pool_batch_ptr = NULL;  // initially null as nothing gotten yet
    memset(&group_info_ptr->tod_built, 0x00, sizeof(group_info_ptr->tod_built)); // zero tod for this group until build completes

    // Update count of pool file groups that the driver will remember

    pool_table_ptr->number_of_groups++;

    tpf_stpoc(TPF_STPOC_OFF, stpoc_value); // done modifying for now

    coruc(pool_table_ptr); // unhold the table now that we are done adding to it


    // Allocate a DECB to work with the pool records that will be in this group.
    //  If we cannot allocate the DECB, bail out as there is no point in continuing with the build.

    TPF_DECB *decb;
    DECBC_RC decbRC;
    char errMsg[100] = ""; // use to generate error messages

    decb = tpf_decb_create(NULL, &decbRC);

    if (decb == NULL) { // if could not allocate a DECB, build failed
       sprintf(errMsg,"ALLOCATE DECB FAILED FOR BUILD OF POOL GROUP %d", my_pool_build.group);
       printMsg(27, 'E', WTOPC_NO_CHAIN, errMsg);
       exit(0);
    } // end if could not allocate a DECB

    // Now that we have created a DECB, allocate enough system heap to hold the pools for this group.

    //   NOTE: To do this, first determine the total size of heap that would be needed to keep track of
    //         all the pools in this group.  This would be the size of the header for the "batch" of pools
    //         PLUS (the number of records)*(size of each item in the "batch")

    uint total_size = sizeof(qfnl_pool_batch) +
                       my_pool_build.number_records*sizeof(qfnl_pool_item);

    if (debug) printMsg(1,'I',WTOPC_NO_CHAIN,"!!! TOTAL SIZE OF BUFFER NEEDED TO HOLD ALL POOL INFO = %d !!!",
                 total_size);

    // Calculate the number of 4K frames that would be needed to hold all pools for this group

    uint frames = total_size / 4096;
    if ( (total_size % 4096) !=0 ) frames++;

    if (debug) printMsg(1,'I',WTOPC_NO_CHAIN,"!!! NUMBER OF 4K FRAMES NEEDED FOR GROUP %d = %d !!!",
                 my_pool_build.group, frames);

    // Allocate the storage needed to hold and manage the pool records for this group

    //  NOTE: currently the token will be non-unique, and we will allocate from 64-bit system heap...

    unsigned int flags = GSYSC_64BIT;

    // Create the token, which will be the 4-character subsystem name concatenated with the first 4 characters of the
    //    string version of the group number (padded to the right with blanks and null-terminated).

    char token[ITOKENLEN];

    memset(token,0x40,ITOKENLEN-1); // pad with blanks
    memset(&token[ITOKENLEN-1],'\0',1); // null terminate

    struct uatbc_area uatbcParms; /* Parameters for UATBC function   */

    uatbcParms.uat_call = UAT_ESSU;
    uatbc (&uatbcParms);
    memcpy(&token[0],&uatbcParms.uat_ss_name,sizeof(uatbcParms.uat_ss_name));

    int index;
    int k = 0;
    char nullString[] = "";

                         // Only move characters from groupString to token if they are non-null
    for (index = sizeof(uatbcParms.uat_ss_name); index < ITOKENLEN-1; index++, k++) {
       if ( memcmp(&my_pool_build.groupString[k], nullString, 1) != 0 )
          memcpy(&token[index], &my_pool_build.groupString[k],1);
    }

    if (debug) printMsg(1,'I',WTOPC_NO_CHAIN,"!!! THE TOKEN GENERATED = %s!!!",
                 token);

    // Set up the owner name for the tpf_gsysc request

    char owner [IOWNERLEN] = QFNL_POOL_OWNER;
    memcpy(&owner[10],&uatbcParms.uat_ss_name,sizeof(uatbcParms.uat_ss_name));

    if (debug) printMsg(1,'I',WTOPC_NO_CHAIN,"!!! WE WILL GET STORAGE WITH OWNER NAME = %s!!!",
                 owner);

    // Attempt to allocate storage to hold pools

    tpf_stpoc(TPF_STPOC_ON, stpoc_value);  // allow ourselves to modify key 9 storage gotten by tpf_gsysc

    group_info_ptr->pool_batch_ptr = (qfnl_pool_batch_Ptr) tpf_gsysc(frames, token, owner, flags);
    printMsg(3,'I',WTOPC_NO_CHAIN,"ADDRESS OF POOLS FOR THIS GROUP IS %.16lX",
          (long) group_info_ptr->pool_batch_ptr);

    if ( group_info_ptr->pool_batch_ptr == NULL ) { // if failed to allocate, get out

      sprintf(errMsg,"ALLOCATE STORAGE FOR POOLS FAILED FOR POOL GROUP %d; ERRNO = %d", my_pool_build.group, errno);
      printMsg(86, 'E', WTOPC_NO_CHAIN, errMsg);
      if (decb) {
        if (levtest(decb))
           relcc(decb);
        tpf_decb_release(decb);
      }
      exit(0);

    }

    // Now that we got the storage successfully and chained it to the table of pool groups, initialize the header

    memcpy (group_info_ptr->pool_batch_ptr->groupString,
            my_pool_build.groupString,
            sizeof(group_info_ptr->pool_batch_ptr->groupString));           // copy string version of group number

    group_info_ptr->pool_batch_ptr->group_number = my_pool_build.group;     // copy numerical group number

    group_info_ptr->pool_batch_ptr->num_frames = frames;                    // copy number of frames that were allocated

    memcpy(group_info_ptr->pool_batch_ptr->token, token, ITOKENLEN);        // copy the token value


    // Initialize prior to Main Loop                                            

    int success_count = 0;  // we will count how many pools successfully processed

    unsigned int pool_count = 0;     // count of pools as we attempt them

    boolean failure = false;  // so far no reason to stop

    qfnl_pool_item_Ptr pool_item = (qfnl_pool_item_Ptr) &(group_info_ptr->pool_batch_ptr->first_item);  // point at first slot in our table

    TPF_FA8 fileaddr;  // used to hold pool file address when gotten
    char FARFstring[QFNL_FARF_STRING_SIZE];  // used for debugging purposes

    tpf_TOD_ext_type tod;  // used to store TOD which will be copied to the pool item

    void * record_ptr = NULL;  // pointer for record read (and later filed)

    // Main Loop which will get and initialize all pools in this group:

    for (pool_count = 1;
         (pool_count <= my_pool_build.number_records) && (!failure);
         pool_count++, pool_item++) { // for every pool in the group unless there is a failure

      if (debug) printMsg(1,'I',WTOPC_NO_CHAIN,"!!! Entered loop, pool_count = %d !!", pool_count);
      tpf_stpoc(TPF_STPOC_OFF, stpoc_value); // turn off modify before giving up control
      fileaddr = getfc(decb,
                       GETFC_TYPE1,
                       group_info_ptr->handle.handle_id,
                       (int)(GETFC_NOBLOCK+GETFC_FILL),
                       GETFC_NOSERRC);
      tpf_stpoc(TPF_STPOC_ON, stpoc_value); // allow modify again

      if (!fileaddr) { // if GETFC failed

         failure = true;
         sprintf(errMsg,"GETFC FOR %dTH POOL IN POOL GROUP %d FAILED, BUILD TERMINATED", pool_count, my_pool_build.group);
         printMsg(87, 'E', WTOPC_NO_CHAIN, errMsg);

        // end if GETFC failed
      } else { // succesfully got pool record

        int convertOK = qfnl_FARF6_to_hex_string(FARFstring, sizeof(FARFstring), fileaddr);
        if ( convertOK == 1 ) {  // converted ok
           if (debug) printf("\n!!! Hooray... The %dTH pool in group %d file address is %s !!\n",
                      pool_count, my_pool_build.group, FARFstring);
        }

        pool_item->FA = fileaddr;  // copy pool file address to table

        // Flag the item as invalid until we can successfully initialize it
        pool_item->flags = QFNL_POOL_ITEM_NOT_VALID;

        if (success_count == 0) { // if this is the first record in the group, try to read it to get its size
                                  // NOTE:   we will use the core block gotten on the read to initialize all the
                                  //     other records in the group
           // read the first record

           tpf_stpoc(TPF_STPOC_OFF, stpoc_value); // turn off modify before giving up control
           record_ptr = find_record_ext(decb,                               // decb
                                        &fileaddr,                          // file address of record to read
                                        RECID_RESET,                        // bypass record ID check
                                        '\0',                               // bypass RCC
                                        NOHOLD_WAIT,                        // find without hold and wait
                                        FIND_DEFEXT);                       // use default non GDS
           tpf_stpoc(TPF_STPOC_ON, stpoc_value); // allow modify again

           if (record_ptr == NULL) { // if failure reading first pool

             failure = true;
             if (convertOK)
               sprintf(errMsg, "ERROR READING POOL FILE ADDRESS %s, SUD = %2.2X",
                                FARFstring,  decb->idecsud);
             else
               sprintf(errMsg, "ERROR READING %dTH POOL IN GROUP %d, SUD = %2.2X",
                                pool_count, my_pool_build.group,  decb->idecsud);
             printMsg(88, 'E', WTOPC_NO_CHAIN, errMsg);


           } // end if failure reading first pool

           if (!failure) {
             group_info_ptr->record_size = decb->cbrw.cb_size; // get the record size
             if (debug) printf("\n !!! Hooray .... Record SIZE first record in group is %d", group_info_ptr->record_size);
           }

        } // end if this is the first record ...

        // file out the record for the build
        if (!failure) {

          // Modify the contents of the record in memory

          tpf_stpoc(TPF_STPOC_OFF, stpoc_value); // turn off modify before giving up control
          if (qfnl_modify_pool_record(record_ptr,
                                       fileaddr,
                                       group_info_ptr,
                                       &tod) == 0) {  // if modify fails quit now
            failure = true;
          } else { // otherwise copy the time that is stored in the record as time the pool was added
            tpf_stpoc(TPF_STPOC_ON, stpoc_value); // allow modify again
            memcpy(&pool_item->tod_added, &tod, sizeof(tod));
          }
        }

        if (!failure) {  // if no failures yet, try to file the record

          tpf_stpoc(TPF_STPOC_OFF, stpoc_value); // turn off modify before giving up control
          file_record_ext(decb,
                          &fileaddr,
                          group_info_ptr->hex_rid,
                          '\0',
                          NOREL,
                          FILE_DEFEXT);

          if ( waitc() ) {  // if error on wait
             failure = true;
             if (convertOK) {
               sprintf(errMsg, "ERROR FILING POOL FILE ADDRESS %s, SUD = %2.2X",
                                FARFstring,  decb->idecsud);
             } else {
               sprintf(errMsg, "ERROR FILING %dTH POOL IN GROUP %d, SUD = %2.2X",
                                pool_count, my_pool_build.group,  decb->idecsud);
             }
             printMsg(89, 'E', WTOPC_NO_CHAIN, errMsg);
          }

          tpf_stpoc(TPF_STPOC_ON, stpoc_value); // allow modify again

        } // end if no failures yet, try to file the record

        if (!failure) {
          success_count++;   // increment records successfully processed

          // Only if the pool is successfully filed, store the TOD it was last updated in the item for this pool in the batch of pools for this group
          memcpy(&pool_item->tod_updated, &tod, sizeof(tod));

          // Flag the item as valid
          pool_item->flags = QFNL_POOL_ITEM_VALID;
        }

        if (group_info_ptr->group_flag == QFNL_GROUP_STAT_STOPPED) { // If group suddenly stopped, terminate the build
                                                                     //   as someone most likely entered a ZTEST FFACT POOL STOP
                                                                     //   command...

             failure = true;
             printMsg(31,'E',WTOPC_NO_CHAIN,"BUILD FOR GROUP %d TERMINATING: STOPPED STATUS DETECTED",
                                group_info_ptr->group_number);

        } // end If group suddenly stopped

      } // end else successfully got pool record

    } // end for every pool in the group unless there is a failure

    if (failure) {

       //  If there was a failure, determine if the failure was total
       //     or if the build just terminated prematurely, and update the pool group table
       //     accordingly...

       if (success_count == 0) { // if no records were successfully processed, consider it a total failure

          printMsg(25, 'E', WTOPC_NO_CHAIN, "BUILD FOR POOL GROUP %d FAILED", my_pool_build.group);

          // On a total failure, clean up by removing all record of the failed build from our pool group control table...

          group_info_ptr->group_flag = QFNL_GROUP_STAT_FREESLOT;  // nothing built so make table slot free
          group_info_ptr->group_number = 0;
          group_info_ptr->number_records = 0;
          group_info_ptr->handle.getfc_rid = 0;
          group_info_ptr->record_size = 0;
          group_info_ptr->expiry = 0;
          memset(group_info_ptr->stpool,0x00,QFNL_STPOOL_SIZE);
          memset(group_info_ptr->hex_rid,0x00,QFNL_RID_SIZE);
          memset(&group_info_ptr->tod_built, 0x00, sizeof(tod));

          // We must also free the storage that was allocated for the batch of pools and set its pointer to NULL

          if ( tpf_rsysc( (void *) group_info_ptr->pool_batch_ptr,
                          frames,
                          token ) != RSYSC_OK ) {
             sprintf(errMsg,"FAILURE TO FREE STORAGE FOR POOLS FOR POOL GROUP %d; ERRNO = %d", my_pool_build.group, errno);
             printMsg(90, 'E', WTOPC_NO_CHAIN, errMsg);
          } else
             group_info_ptr->pool_batch_ptr = NULL;

          // Decrement count of pool file groups that the driver will remember

          pool_table_ptr->number_of_groups--;

       } else {  // partially successful

          printMsg(35, 'E', WTOPC_NO_CHAIN, "BUILD FOR POOL GROUP %d TERMINATED AT %dTH POOL",
                   group_info_ptr->group_number, success_count);
          group_info_ptr->group_flag = QFNL_GROUP_STAT_STOPPED;   // build no longer in progress
          group_info_ptr->number_records = success_count;

       }         // end else partially successful

    // end if failure
    } else { // else successful build

       group_info_ptr->group_flag = QFNL_GROUP_STAT_STOPPED;   // mark group stopped, as build now complete
                                                                     // (i.e. no longer in progress)

       printMsg(36, 'I', WTOPC_NO_CHAIN, "BUILD FOR POOL GROUP %d COMPLETED", group_info_ptr->group_number);
    }        // end else successful build

    // Update info on this POOL group to indicate TOD that it was built

    memcpy(&group_info_ptr->tod_built, &tod, sizeof(tod));
    tpf_stpoc(TPF_STPOC_OFF, stpoc_value); // finally done modifying

    memcpy (&my_pool_build.tod_built, &tod, sizeof(tod)); // save for expiry cretc

    if (levtest(decb))
       relcc(decb);           // We are done with the block used to initialize records, so release it

    // If value of Expire is ZERO, then no pools will ever expire for this group so we are done.  Otherwise...

    if (my_pool_build.expiry !=0 ) { // if we want expiry for this group,
                                     //   we will cret an ecb to periodically flag the pools as not valid

       char * coreBlock = NULL;

       // Get a core block to pass to the expiry ecb...

       coreBlock = (char *) getcc (D3, (enum t_getfmt)(GETCC_TYPE +
                              GETCC_FILL), L4, 0x00);

       // We will pass the ecb a copy of the build parameters for this group

       memcpy(coreBlock, &my_pool_build, sizeof(my_pool_build));

       // Create the ecb X seconds for now based on expiry value

       cretc_level (CRETC_SECONDS|CRETC_1052,
                    QFNL( ),
                    my_pool_build.expiry,
                    "EXPR",
                    D3);

       if (levtest(D3))
          relcc(D3);           // We are done with the block used for cretc_level, so release it

    } // end if we want expiry for this group

    tpf_decb_release(decb);   // We are done with the DECB as well, so release it

    return;

 } // end have table for pool

 return;

} // end of qfnl_pool_build


static void qfnl_pool_start(struct IPRSE_output *parse,
                             int num_parms)
{ // beginning of qfnl_pool_start

 if (debug) printf("\n!! Hooray!! Recognized POOL START !!\n");

 qfnl_child_ecb_parm_list ecb_parameters;       // parameters to pass to children ECBs
 struct swisc_create_input create_parameters;   // used for swisc parameters

 struct ev0bk event_blk;    // for event processing

 // parse the options specified for the ZTEST FFACT POOL START command...
 if ( qfnl_parse_pool_start(parse,num_parms) == 0 )
   return;    // return if parser did not like something

 // At this point, we know that the ZTEST FFACt POOL START command passed preliminary syntax validation.
 //  The next thing to do is see if we can find information for that group in the
 //  POOL Group Control Table

 qfnl_pool_info_Ptr found_pool_info_ptr = qfnl_locate_pool_group(my_pool_start.group, TRUE);
 if (found_pool_info_ptr == NULL) {  // goup not found
   printMsg(39,'E',WTOPC_NO_CHAIN,"START OF GROUP %d FAILED: GROUP INFO FROM BUILD NOT FOUND",
            my_pool_start.group);
   return;
 } else {  // group found

   printMsg(3,'I',WTOPC_NO_CHAIN,"ADDRESS OF pool group found is %.16lX", (long) found_pool_info_ptr);
   if (debug) printf("\n !!!! Start Command Preliminary Syntax Checks Accepted !!!");

   // First check that a build for this group is not still in progress.  If that is the case, we cannot
   //  start an instance of activity on this group until the build completes.  So reject the request.

   if (found_pool_info_ptr->group_flag == QFNL_GROUP_STAT_BUILDING) {
     printMsg(40,'E',WTOPC_NO_CHAIN,"START COMMAND FOR POOL GROUP %d REJECTED: BUILD IS IN PROGRESS",
              my_pool_start.group);
     return;
   }

   // If this instance of the driver is to be started with a utilization class, ensure that the utilization class
   //   is valid (defined on file and loaded into memory) before proceeding...

   // if ( (my_pool_start.useUtilClass)&&(lodic_isUtilClass(my_pool_start.utilClass)==0) ) {
   if (my_pool_start.useUtilClass) { // if we are using a utilization class, validate it
     if (lodic_isUtilClass(my_pool_start.utilClass)==0) { // if utilizaiton class not valid
       printMsg(110,'E',WTOPC_NO_CHAIN,"START COMMAND FOR GROUP %d REJECTED: UTILCLASS %s NOT IN MEMORY",
                my_pool_start.group, my_pool_start.utilClass);
       return;
     }
   } // end if we are using a utilization class, validate it

   /* ---------------------------  Beginning of fix to allow ZTEST i FFACT POOL START to work when called via ZEXEC ------------------ */
   //
   //  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
   //  NOTE !!! It is important for this to code, which contains a fork to a child ECB who will create the grandchildren
   //         that do the actual work, to be done BEFORE setting of any ECB attributes via tpf_easetc
   //  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
   //
   // Instead of doing the work ourselves, we fork another ECB that will serve as the potential long-running ECB to do the
   //     work.  The reason for this is to allow ZEXEC to execute a ZTEST FFACT POOL START command which loops
   //     forever.  If the ECB for the command itself loops forever, ZEXEC will never get control back and will timeout.
   //     So we want the ZTEST FFACT POOL START parent ECB to exit, and have the child created by fork do the looping and possibly
   //     be a longrunning ECB that continuously generates work.

   pid_t pid = fork();
   if (pid == -1) {  // if fork failed...
     // Access profile definition in memory to flag error and stop profile

     tpf_stpoc(TPF_STPOC_ON, stpoc_value);  // allow ourselves to modify key 9 storage gotten by tpf_gsysc
     tpf_serialized_update8(1, (tsug_t *)&found_pool_info_ptr->group_stats.envErr);  // consider this an environmental error
     found_pool_info_ptr->group_flag = QFNL_GROUP_STAT_STOPPED; // stop all activity this group
     tpf_stpoc(TPF_STPOC_OFF, stpoc_value); // done modifying for now

     // take a dump and exit
     char errMsg[100] = "";
     sprintf(errMsg,"START COMMAND FOR GROUP %d ATTEMPT TO EXECUTE fork() FAILED",
   		     found_pool_info_ptr->group_number);
     serrc_op(SERRC_ALL_HEAP,0x5555,errMsg,NULL);
     return;  // return to exit
   }  // end if fork failed...

   if (pid) { // if parent of fork, do nothing but exit
     if (debug) printMsg(1,'I',WTOPC_NO_CHAIN,"!!!!! PARENT of START of GROUP %d exiting!!!!!", found_pool_info_ptr->group_number);
     return; // return to exit
   } // end if parent of fork

   // We only arrive here if we are the child ECB of the fork.  This child will loop the number of times (or indefinitely) on the profile,
   //   creating its own children to operate on that profile.

   /* ----------------------------     end   of fix to allow ZTEST i FFACT POOL START to work when called via ZEEXC ------------------ */
   // Now check if this group is already started.  If it is, then make sure that the LOCK-Y or N
   //  matches the locking status for this group.
   // Otherwise if the group is not already started, this is the first instance for the group
   //  so set the locking status for the group to match the locking status specified for this instance.

   boolean lockBasedOnInstance = false;  // initially assume this instance will not determine locking status for whole group

   if (found_pool_info_ptr->group_flag == QFNL_GROUP_STAT_STARTED ) { // group already started

     if ( (found_pool_info_ptr->group_stats.lockingActive) && (my_pool_start.doLocking == 0) ) {
       printMsg(41,'E',WTOPC_NO_CHAIN,"START COMMAND FOR POOL GROUP %d REJECTED: ACTIVITY ALREADY RUNNING WITH LOCKING",
                my_pool_start.group);
       return;
     }

     if ( (found_pool_info_ptr->group_stats.lockingActive == 0) && (my_pool_start.doLocking) ) {
       printMsg(42,'E',WTOPC_NO_CHAIN,"START COMMAND FOR POOL GROUP %d REJECTED: ACTIVITY ALREADY RUNNING WITHOUT LOCKING",
                my_pool_start.group);
       return;
     }

   } else { // group not started, so set its locking status to match this instance
     lockBasedOnInstance = true;   // we will set locking status for group based on this instance
   }

   // Before actually starting activity for this instance, check whether or not we need to use a LODIC utilization class
   //    and/or resource class...
   // NOTE: The driver will use LODIC_IBMBATCH for its resource class

   int rc = 1;  // assume success until failure

   if (my_pool_start.useUtilClass) {

     // if we want to use both a utilization class and resource class, then do so...

     if (my_pool_start.useResClass) {

       printMsg(1,'I',WTOPC_NO_CHAIN,"calling lodic_ext(LODIC_ECBCREATE+LODIC_IBMBATCH+LODIC_CPU_UTIL, NULL, my_pool_start.utilClass);");
       rc = lodic_ext(LODIC_ECBCREATE+LODIC_IBMBATCH+LODIC_CPU_UTIL, NULL, my_pool_start.utilClass);

     } else { // otherwise use just a utilization class...

       printMsg(1,'I',WTOPC_NO_CHAIN,"calling lodic_ext(LODIC_ECBCREATE+LODIC_NONE+LODIC_CPU_UTIL, NULL,  my_pool_start.utilClass);");
       rc = lodic_ext(LODIC_ECBCREATE+LODIC_NONE+LODIC_CPU_UTIL, NULL, my_pool_start.utilClass);

     } // end else otherwise use just a utilization class...

   } else {

      if (my_pool_start.useResClass) { // if use just resource class

       printMsg(1,'I',WTOPC_NO_CHAIN,"calling lodic_ext(LODIC_ECBCREATE+LODIC_IBMBATCH, NULL);");
         rc = lodic_ext(LODIC_ECBCREATE+LODIC_IBMBATCH, NULL);

      } // end if use just resource class

   }

   // if there was an error from lodic_ext, abort the start and inform the user

   char errMsg[100] = ""; // use to generate error messages

   if (rc == 0) {  // if error form lodic

      switch (errno) {
         case ETPFLODIC_LOWRES:
            sprintf(errMsg, "GROUP %d NOT STARTED: LODIC FAILED DUE TO LOW RESOURCES",
                    my_pool_start.group);
            break;
         case ETPFLODIC_MAXSUSP:
            sprintf(errMsg, "GROUP %d NOT STARTED: LODIC FAILED DUE TO MAX SUSPEND (NOT EXPECTED)",
                    my_pool_start.group);
            break;
         case ETPFLODIC_NOTFOUND:
            sprintf(errMsg, "GROUP %d NOT STARTED: UTIL CLASS %s DOES NOT EXIST",
                    my_pool_start.group, my_pool_start.utilClass);
            break;
         default:
            sprintf(errMsg, "GROUP %d NOT STARTED: LODIC FAILED WITH UNKNOWN ERROR",
                    my_pool_start.group);
            break;

      printMsg(112, 'E', WTOPC_NO_CHAIN, errMsg);
      return; // bail out

      } // end switch

   } // end if error form lodic

   // ALSO before actually starting activity for this instance, check and perform any tpf_easetc that was
   //    requested on the START command...

   if (my_pool_start.lowPriority) { // if low priority requested
      if (my_pool_start.inheritLow) {
        tpf_easetc(TPF_EASETC_LOWPRIORITY, TPF_EASETC_INHERIT_YES+TPF_EASETC_SET_ON);
      } else {
        tpf_easetc(TPF_EASETC_LOWPRIORITY, TPF_EASETC_INHERIT_NO+TPF_EASETC_SET_ON);
      }
   }

   if (my_pool_start.switchable) { // if switchable requested
      if (my_pool_start.inheritSW) {
        tpf_easetc(TPF_EASETC_SWITCHABLE, TPF_EASETC_INHERIT_YES+TPF_EASETC_SET_ON);
      } else {
        tpf_easetc(TPF_EASETC_SWITCHABLE, TPF_EASETC_INHERIT_NO+TPF_EASETC_SET_ON);
      }
   }

   //
   // Now that we know that the group is not currently being built, and there there is no conflict
   //   with this instance and previously started instances still running in terms of locking,
   //   and any lodic_ext and/or tpf_easetc that was requested has been performed,
   //   prepare to spawn child ECBs to perform activity against this group:
   //
   //     Set status indicator for this group to show that activity has been started for it
   //       and increment the total count of active instances for this group.
   //       ALSO, if this is the first instance to be started for this group,
   //        set status of whether or not locking should be done for the instances operating on
   //        this group.
   //

   //     Rather than blindly using the record ID for the group, read the first record in the group using record ID zero
   //       to determine the record ID.  Verify that the record ID read from the first record in this fashion matches the
   //       record ID that had been stored in the group information for this group during BUILD.  (Also verify that the
   //       first record is initalized properly.)
   //
   //     Then pass the record ID that was read to the child ECBs as a parameter.
   //
   //     NOTE: This is only done for purposes of testing find with record ID zero.

   if (!qfnl_first_record_ok(found_pool_info_ptr, ecb_parameters.rec_ID)) { // if problem getting record ID from first record
                                                                             //   or the record is not properly initialized
     printMsg(43,'E',WTOPC_NO_CHAIN,"START OF POOL GROUP %d ABORTED DUE TO PROBLEM WITH FIRST RECORD",
                                     my_pool_start.group);
     return; // !!! Important to bail out before setting any started indicator for this instance !!!
   }

   tpf_stpoc(TPF_STPOC_ON, stpoc_value);  // allow ourselves to modify key 9 storage gotten by tpf_gsysc
   found_pool_info_ptr->group_flag = QFNL_GROUP_STAT_STARTED;
   if (lockBasedOnInstance)
     found_pool_info_ptr->group_stats.lockingActive = my_pool_start.doLocking;
   tpf_serialized_update8(1, (tsug_t *)&found_pool_info_ptr->group_stats.instanceCount);
   tpf_stpoc(TPF_STPOC_OFF, stpoc_value); // done modifying for now

   //
   //     Set up child ECB parameters for the instance being started.
   //        NOTE: The only parameter we do not set is the child_seed.  That will be different for each
   //        ECB that is created.
   //

   ecb_parameters.group = my_pool_start.group;  // group number ECB is to operate on
   ecb_parameters.group_type = QFNL_POOL;       // dealing with ZTEST FFACt POOL START ...
   ecb_parameters.numberOfReads = my_pool_start.numberOfReads;  // number of reads to be done by each child ECB
   ecb_parameters.numberOfWrites = my_pool_start.numberOfWrites;  //number of writes to be done by each child ECB
   ecb_parameters.doLocking = my_pool_start.doLocking;  // whether ECB is to lock the records prior to writing to them



   //
   //     Set up SWISC parameters that will be used evey time a new ECB is created for this instance
   //

   create_parameters.program = "QFNL";
   create_parameters.istream = ecbptr()->ce1isn;         // create ECB on current i-stream for starters
   create_parameters.cpu_list = SWISC_LIST_READY;        // and put it on ready list
   create_parameters.bypass = SWISC_ACTNUM_CURRENT;
   create_parameters.ebw_data = &ecb_parameters;
   create_parameters.ebw_data_length = sizeof(ecb_parameters);
   create_parameters.data_level = NO_LVL;

   //
   // While activity for this group has been STARTed, continue to generate the specified number of ECBs on the
   //   input list, waiting a bit after generating each batch of ECBs
   //

   int j; // index of inner loop

   srand(my_pool_start.seed);  // seed the random generator

   // int K = 1;  // TEMP PATCH!!!!!

   // printf("\n!!!! my_pool_start.numECBs = %d !!!!\n", my_pool_start.numECBs);
   unsigned int ecbNum = 1;  // pass a sequential number to each child ECB useful just for debugging messages

   // for (K=1; K <= 10; K++ ) { // TEMP PATCH!!! temporary loop of 10 just to print random numbers for first 10 batches of ECBs

   while (found_pool_info_ptr->group_flag == QFNL_GROUP_STAT_STARTED) { // while this group is started
     for (j=1;
          (j <= my_pool_start.numECBs)&&(found_pool_info_ptr->group_flag == QFNL_GROUP_STAT_STARTED);
          j++) { // loop to generate batch of ECBs

        ecb_parameters.ecbNum = ecbNum;  // pass ecb number to child
        ecbNum++; // increment for next iteration of loop
        // printf("\n !!!!! ecb_parameters.ecbNum = %d !!!!!\n", ecb_parameters.ecbNum);

        // generate random seed to be passed to child ECB to seed its own srand function call

        ecb_parameters.child_seed = qfnl_random(1,
                                                found_pool_info_ptr->number_records);

        if (verbose) printf("\n !!!! The SEED to pass to CHILD is %d !!!!\n", ecb_parameters.child_seed);

        // create child ECB

        swisc_create(&create_parameters);

     } // end loop to generate batch of ECBs

     // Wait 10 milliseconds and put the parent ECB back on the input list before generating the next batch of children ECBs

     event_blk.evnpstinf.evnbkc1 = 1;
     tpf_evntc_timeout(&event_blk, EVENT_CNT, 'N', 10, EVNTC_1052, TPF_EVENT_MILLI);
     evnwc(&event_blk, EVENT_CNT);

     dlayc();     // take a trip to the input list

   } // end while this group is STARTed

 } // end else group found

 // We only reach here if this group is no longer started.  So decrement the total number of active instances
 //   and exit this instance

 tpf_stpoc(TPF_STPOC_ON, stpoc_value);  // allow ourselves to modify key 9 storage gotten by tpf_gsysc

 tpf_serialized_update8(-1, (tsug_t *)&found_pool_info_ptr->group_stats.instanceCount);

 printMsg(10,'I',WTOPC_NO_CHAIN,
          "Instance for Pool Group %d is EXITING.  Total instances left = %ld", //d26352
          my_pool_start.group, found_pool_info_ptr->group_stats.instanceCount);

 // It may not be necessary, but just as a safety mechanism, if there are no more instances left
 //   set the locking indicator for this group to be the default of locking.  Also inform the user
 //   that the group is now STOPPED.

 if (found_pool_info_ptr->group_stats.instanceCount == 0) {
   found_pool_info_ptr->group_stats.lockingActive = 1;
   printMsg(57,'I',WTOPC_NO_CHAIN,"POOL GROUP %d STOPPED", my_pool_start.group);
 }

 tpf_stpoc(TPF_STPOC_OFF, stpoc_value); // done modifying for now

 return; // return to exit

} // end of qfnl_pool_start

qfnl_pool_info_Ptr qfnl_locate_pool_group (int group,
                                           boolean beNoisy)
{ // BEGINNING of function qfnl_locate_pool_group

/* ****************************************************************************************************** */
/* NOTE: The purpose of this function is to find the POOL group in the Pool Group Control                 */
/*   Table.                                                                                               */
/*  Input:                                                                                                */
/*        int group is the group number we are looking for                                                */
/*        boolean beNoisy = TRUE print informational or internal error messages                           */
/*                          FALSE work silently, only put out messages on first major error               */
/*  Output is the address of the group information in the POOL File Control Table.  If any error is       */
/*    encountered or the pool file group is not found, NULL is returned.                                  */
/*                                                                                                        */
/*                                                                                                        */
/* ****************************************************************************************************** */

 qfnl_pool_info_Ptr group_info_ptr = NULL; // hold address of group information in POOL Group Control Table;
                                           // NULL until found.

 // Find the Pool Group Control Table or allocate it, if it is not already in memory

 qfnl_pool_ctl_tbl_Ptr pool_table_ptr = get_pool_table();
 int i;

 // If we could not find or allocate the Pool Group Control Table in memory, we will not be able to
 //    put the group information in there.  So return error.

 if (pool_table_ptr == NULL) {

   if (beNoisy)
      printMsg(2,'E',WTOPC_NO_CHAIN,"CANNOT FIND OR ALLOCATE POOL GROUP CONTROL TABLE");
   return (NULL);

 } else {  // have table for POOL

   if (beNoisy)
      printMsg(3,'I',WTOPC_NO_CHAIN,"ADDRESS OF POOL GROUP CONTROL TABLE IS %.16lX", (long) pool_table_ptr);

    /*----------------------------------------------------------------------------------------------------*/
    // Now that the Pool Group Control Table exists in memory, first search it to see if there is
    //   record of this group.  If there is, then we are done.
    /*----------------------------------------------------------------------------------------------------*/

    boolean we_are_done = false;

    group_info_ptr = pool_table_ptr->my_pool;  // point at first pool group slot in table

    if (debug) printf("\n !!!!! The group we are trying to LOCATE is %d !!!!\n", group);

    for(i=0; (i < QFNL_MAX_POOL)&&(!we_are_done); i++) { // for all items in table until we find a match
      if (debug) printf("\n !!!!! ......... Looking at POOL Group Control Table entry for Group = %d !!!!!\n", group_info_ptr->group_number);
      if ( (group == group_info_ptr->group_number) &&
           (group_info_ptr->group_flag != QFNL_GROUP_STAT_FREESLOT) )
      {
        we_are_done = true;
        if (debug) printf("\n!!!! LOCATE of GROUP %d FLAG is %d ADDRESS OF GROUP IS %.16lX!!!!\n",
                            group_info_ptr->group_number, group_info_ptr->group_flag, group_info_ptr);
      }

      if (!we_are_done) group_info_ptr++;  // advance to next slot in my_pool array

    } // end for all items in table

    if (!we_are_done) { // if we are not done, we did not find it, so reset group_info_ptr to NULL
      group_info_ptr = NULL;
      if (debug)
         printf("\n !!! We could not find group %d in memory for this processor !!!!\n", group);
    } // end if we are not done, we did not find it ...

 }         // end else have table for POOL

 // At this point, group_info_ptr is either the group that was found or is still NULL if not found.
 return(group_info_ptr);

} // END of function qfnl_locate_pool_group

int qfnl_parse_pool_start(struct IPRSE_output *parse,
                          int num_parms)
{

// NOTE: This function parses parameters for the ZTEST POOL START command, and stores there
//   values into a structure named my_pool_start.
// This function also performs checks that the generic parser cannot, such as ensuring
//   that when locking is active, the number of reads is >=  the number of writes
// This function returns 0 if an error was detected (and START should not be allowed) or
//                       1 on success (and the caller may continue START processing).

 my_pool_start.doLocking = TRUE;  // default for ZTEST POOL START is to do locking
 boolean lockByDefault = TRUE;     //  assume this is the default

 // Set other attributes to FALSE until we parse parameters that specify otherwise

 my_pool_start.lowPriority = FALSE;  // do not make this instance a low priority ECB unless told otherwise
 my_pool_start.inheritLow = FALSE;   // do not make child ECBs of this instance low priority unless told otherwise
 my_pool_start.switchable = FALSE;   // do not make this instance a switchable ECB unless told otherwise
 my_pool_start.inheritSW = FALSE;    // do not make child ECBs of this instance switchable unless told otherwise
 my_pool_start.useResClass = FALSE;  // do not make this instance and its children use a resource class unless told otherwise
 my_pool_start.useUtilClass = FALSE; // do not make this instance and its children use a utilization class unless told otherwise

 memset (my_pool_start.groupString, '\0', sizeof(my_pool_start.groupString));  // initialize string to nulls
 memset (my_pool_start.utilClass, ' ', sizeof(my_pool_start.utilClass));  // also initialize UTILCLASS to blanks until we are told to use it,
                                                                            //   so that if we are told to use it and set it later, it will
                                                                            //   automatically be padded with blanks as required
 memset (&my_pool_start.utilClass[8], '\0', 1);  // also ensure that UTILCLASS is null-terminated

 while ((num_parms > 0) && parse)
  { // while there are more parameters

   if (debug) printMsg(9, 'I', WTOPC_NO_CHAIN, "!!! PARSE PARM IS %s", parse->IPRSE_parameter);

   if (parse && (strncmp(parse->IPRSE_parameter, "Group-",6) == 0))
   { // if Group specified

    // get the group number and save it both as an integer and as a string

    my_pool_start.group = atoi(parse->IPRSE_value);
    memcpy(my_pool_start.groupString, parse->IPRSE_value, strlen(parse->IPRSE_value));

    if (my_pool_start.group <=0) {
      printMsg(44,'E',WTOPC_NO_CHAIN,"GROUP %d NOT STARTED:  GROUP NUMBER MUST BE GREATER THAN 0", my_pool_start.group);
      return(0);
    }

    //  The parser should prevent an integer greater than 5 digits, but I coded a check anyway just in case

    if (my_pool_start.group > 99999) {
      printMsg(45,'E',WTOPC_NO_CHAIN,"GROUP %d NOT STARTED:  MAXIMUM GROUP NUMBER IS 99999", my_pool_start.group);
      return(0);
    }

    if (debug) printMsg(1,'I',WTOPC_NO_CHAIN,"!!! input Group = %d",my_pool_start.group);

   } // end if Group specified

   if (parse && (strncmp(parse->IPRSE_parameter, "Read-",5) == 0))
   { // if Read count specified

    my_pool_start.numberOfReads = atoi(parse->IPRSE_value);

    if (debug) printMsg(1,'I',WTOPC_NO_CHAIN,"numberOfReads = %d",my_pool_start.numberOfReads);

    if ( (my_pool_start.numberOfReads < 0) ||
         (my_pool_start.numberOfReads > 999) ) {
      printMsg(46,'E',WTOPC_NO_CHAIN,"GROUP %d NOT STARTED:  READS MUST BE BETWEEN 0 AND 999", my_pool_start.group);
      return(0);
    }

   } // end if Read count specified

   if (parse && (strncmp(parse->IPRSE_parameter, "Write-",6) == 0))
   { // if Write count specified

    my_pool_start.numberOfWrites = atoi(parse->IPRSE_value);

    if (debug) printMsg(1,'I',WTOPC_NO_CHAIN,"numberOfWrites = %d",my_pool_start.numberOfWrites);

    if ( (my_pool_start.numberOfWrites < 0) ||
         (my_pool_start.numberOfWrites > 999) ) {
      printMsg(47,'E',WTOPC_NO_CHAIN,"GROUP %d NOT STARTED:  WRITES MUST BE BETWEEN 0 AND 999", my_pool_start.group);
      return(0);
    }

   } // end if Write count specified

   if (parse && (strncmp(parse->IPRSE_parameter, "SEED-",5) == 0))
   { // if SEED specified

    my_pool_start.seed = atoi(parse->IPRSE_value);

    if (debug) printMsg(1,'I',WTOPC_NO_CHAIN,"SEED = %d",my_pool_start.seed);

    if ( (my_pool_start.seed < 0) ||
         (my_pool_start.seed > 999) ) {
      printMsg(48,'E',WTOPC_NO_CHAIN,"GROUP %d NOT STARTED:  SEED MUST BE BETWEEN 0 AND 999", my_pool_start.group);
      return(0);
    }

   } // end if SEED count specified

   if (parse && (strncmp(parse->IPRSE_parameter, "ECBs-",5) == 0))
   { // if ECBs specified

    my_pool_start.numECBs = atoi(parse->IPRSE_value);

    if (debug) printMsg(1,'I',WTOPC_NO_CHAIN,"numECBs = %d",my_pool_start.numECBs);

    if ( (my_pool_start.numECBs < 1) ||
         (my_pool_start.numECBs > 999) ) {
      printMsg(49,'E',WTOPC_NO_CHAIN,"GROUP %d NOT STARTED:  NUMBER ECBS MUST BE BETWEEN 1 AND 999", my_pool_start.group);
      return(0);
    }

   } // end if ECBs specified

   if (parse && (strncmp(parse->IPRSE_parameter, "Lock-",5) == 0))
   { // if Lock specified

    char * lock_ptr = NULL;

    lockByDefault = FALSE;    //  the user is not taking the default, but specifying it

    lock_ptr = parse->IPRSE_value;
    if ((*lock_ptr == 'y') || (*lock_ptr == 'Y')) {
       my_pool_start.doLocking = true;
    }
    else if ((*lock_ptr == 'n') || (*lock_ptr == 'N')) {
       my_pool_start.doLocking = false;
    }
    else {
       printErr(19,"BAD VALUE SPECIFIED FOR Lock; MUST BE Y or N");
       exit(0);
    }

   } // end if Lock specified

   if (parse && (strncmp(parse->IPRSE_parameter, "LOw-",4) == 0))
   { // if LOw specified

    char * low_ptr = NULL;

    low_ptr = parse->IPRSE_value;
    if ((*low_ptr == 'y') || (*low_ptr == 'Y')) {
       my_pool_start.lowPriority = true;
    }
    else if ((*low_ptr == 'i') || (*low_ptr == 'I')) {
       my_pool_start.lowPriority = true;  // inherit means the main ECB for instance has this attribute...
       my_pool_start.inheritLow = true;   //   and any of its children will inherit this attribute as well
    }
    else {
       printErr(19,"BAD VALUE SPECIFIED FOR LOw: MUST BE Y or I");
       exit(0);
    }

    if (debug) printMsg(1,'I',WTOPC_NO_CHAIN,"VALUE OF my_pool_start.lowPriority is %s",my_pool_start.lowPriority ? "TRUE" : "FALSE");
    if (debug) printMsg(1,'I',WTOPC_NO_CHAIN,"VALUE OF my_pool_start.inheritLow is %s",my_pool_start.inheritLow ? "TRUE" : "FALSE");
   } // end if LOw specified

   if (parse && (strncmp(parse->IPRSE_parameter, "SWitch-",7) == 0))
   { // if SWitch specified

    char * switch_ptr = NULL;

    switch_ptr = parse->IPRSE_value;
    if ((*switch_ptr == 'y') || (*switch_ptr == 'Y')) {
       my_pool_start.switchable = true;
    }
    else if ((*switch_ptr == 'i') || (*switch_ptr == 'I')) {
       my_pool_start.switchable = true;   // inherit means the main ECB for instance has this attribute...
       my_pool_start.inheritSW = true;    //   and any of its children will inherit this attribute as well
    }
    else {
       printErr(19,"BAD VALUE SPECIFIED FOR SWitch: MUST BE Y or I");
       exit(0);
    }

    if (debug) printMsg(1,'I',WTOPC_NO_CHAIN,"VALUE OF my_pool_start.switchable is %s",my_pool_start.switchable ? "TRUE" : "FALSE");
    if (debug) printMsg(1,'I',WTOPC_NO_CHAIN,"VALUE OF my_pool_start.inheritSW is %s",my_pool_start.inheritSW ? "TRUE" : "FALSE");
   } // end if SWitch specified

   if (parse && (strncmp(parse->IPRSE_parameter, "RESourceclass-",14) == 0))
   { // if RESourceclass specified

    char * res_ptr = NULL;

    res_ptr = parse->IPRSE_value;
    if ((*res_ptr == 'y') || (*res_ptr == 'Y')) {
       my_pool_start.useResClass = true;
    }
    else if ((*res_ptr == 'n') || (*res_ptr == 'N')) {
       my_pool_start.useResClass = false;
    }
    else {
       printErr(19,"BAD VALUE SPECIFIED FOR RESourceclass; MUST BE Y or N");
       exit(0);
    }

    if (debug) printMsg(1,'I',WTOPC_NO_CHAIN,"VALUE OF my_pool_start.useResClass is %s",my_pool_start.useResClass ? "TRUE" : "FALSE");

   } // end if RESourceclass specified

   if (parse && (strncmp(parse->IPRSE_parameter, "UTILclass-",10) == 0))
   { // if UTILclass specified

       my_pool_start.useUtilClass = TRUE; // indicate that this instance will use a utilization class
       memcpy(my_pool_start.utilClass, parse->IPRSE_value, strlen(parse->IPRSE_value));  // copy the utilization class for later use
       if (debug) printMsg(1,'I',WTOPC_NO_CHAIN,"!!! my_pool_start.utilClass initially set to %s!!!",my_pool_start.utilClass);

   } // end if UTILclass specified
   parse = parse->IPRSE_next;
   num_parms--;

  } // end while there are more parameters

 // If the user has not specified a value for LOCK-Y|N, inform them of the default

 if (lockByDefault) {
  printMsg(50,'I',WTOPC_NO_CHAIN,"FOR ATTEMPTED START ON GROUP %d DEFAULT VALUE LOCK-Y ASSUMED",my_pool_start.group);
 }

 // Do not allow both Read and Write to be zero

 if ( (my_pool_start.numberOfReads == 0) && (my_pool_start.numberOfWrites == 0) ) {
  printMsg(51,'E',WTOPC_NO_CHAIN,"START OF GROUP %d FAILED.  READ AND WRITE CANNOT BOTH BE ZERO",
                                   my_pool_start.group);
  return (0);
 }

 // If Locking is in effect, ensure that Number of Reads is GREATER OR EQUAL to Number of Writes

 if ( my_pool_start.doLocking &&
      (my_pool_start.numberOfReads < my_pool_start.numberOfWrites) ) {
  printMsg(52,'E',WTOPC_NO_CHAIN,"START OF GROUP %d FAILED.  FOR LOCKING THERE MUST BE AT LEAST ONE READ FOR EVERY WRITE",
                                   my_pool_start.group);
  return (0);
 }

 if (debug) printf("\n!!!!! my_pool_start.doLocking = %d !!!!!\n", my_pool_start.doLocking);
 return(1);
}

void qfnl_check_pool_status(char * statusString, qfnl_pool_info_Ptr group_info_ptr)
{

// This function will store the status of the group referenced by group_info_ptr
//   at the location pointed to by statusString

 switch (group_info_ptr->group_flag)
  {
     case QFNL_GROUP_STAT_STARTED:
       memcpy(statusString, "STARTED", strlen("STARTED"));
       break;

     case QFNL_GROUP_STAT_STOPPED:
       memcpy(statusString, "STOPPED", strlen("STOPPED"));
       break;

     case QFNL_GROUP_STAT_BUILDING:
       memcpy(statusString, "BUILDING", strlen("BUILDING"));
       break;

     case QFNL_GROUP_STAT_FREESLOT:
       memcpy(statusString, "NOT IN USE", strlen("NOT IN USE"));
       break;

     default:
       memcpy(statusString, "UNKNOWN", strlen("UNKNOWN"));
       break;

  } // end switch

 return;

} // end function qfnl_check_pool_status

int qfnl_parse_pool_status(struct IPRSE_output *parse,
                            int num_parms)
{

// NOTE: This function parses parameters for the ZTEST POOL STATUS command, and stores there
//   values into a structure named my_pool_status.

// This function returns 0 if an error was detected (and STATUS should not be allowed) or
//                       1 on success (and the caller may continue STATUS processing).


 memset (my_pool_status.groupString, '\0', sizeof(my_pool_status.groupString));  // initialize string to nulls

                            //
 my_pool_status.group = 0;  // assign zero for group number until it is parsed...
                            //   if it remains that way, the caller may interpret the STATUS request
                            //   to be an abbreviated request for all groups in memory
                            //

 while ((num_parms > 0) && parse)
  { // while there are more parameters

   if (debug) printMsg(9, 'I', WTOPC_NO_CHAIN, "!!! PARSE PARM IS %s", parse->IPRSE_parameter);

   if (parse && (strncmp(parse->IPRSE_parameter, "Group-",6) == 0))
   { // if Group specified

    // get the group number and save it both as an integer and as a string

    my_pool_status.group = atoi(parse->IPRSE_value);
    memcpy(my_pool_status.groupString, parse->IPRSE_value, strlen(parse->IPRSE_value));

    if (my_pool_status.group <=0) {
      printMsg(53,'E',WTOPC_NO_CHAIN,"NO STATUS FOR GROUP %d:  GROUP NUMBER MUST BE GREATER THAN 0", my_pool_status.group);
      return(0);
    }

    //  The parser should prevent an integer greater than 5 digits, but I coded a check anyway just in case

    if (my_pool_status.group > 99999) {
      printMsg(54,'E',WTOPC_NO_CHAIN,"NO STATUS FOR GROUP %d:  MAXIMUM GROUP NUMBER IS 99999", my_pool_status.group);
      return(0);
    }

   } // end if Group specified

   parse = parse->IPRSE_next;
   num_parms--;

  } // end while there are more parameters

 return(1);  // return successful parse
} // end of function qfnl_parse_pool_status

void qfnl_pool_general_status(void)
{

// This function will loop through the table of POOL FILE Groups
//  and display general driver status information for each.

  qfnl_pool_ctl_tbl_Ptr pool_table_ptr = get_pool_table();

  if (pool_table_ptr == NULL) {  // nothing in memory
     printMsg(93,'I',WTOPC_NO_CHAIN,"THERE ARE NO POOL GROUPS DEFINED FOR THIS PROCESSOR");
     return;
  } else { // there could be pool groups in memory

     int i;
     boolean at_least_one = false; // assume we do not have at least one group until found

     char statusString[15];                            // will be used for status display
     char pool_section[QFNL_STPOOL_SIZE+1];               // will be used to display pool section

     printMsg(60,'I',WTOPC_CHAIN,"SUMMARY FOR ALL POOL GROUPS IN MEMORY FOR THIS PROCESSOR"); // print out quick status header
     printMsg(60,0,WTOPC_CHAIN,
       "GROUP POOL SECTION    STATUS   FINDS*           FILES*           ERRORS");

     qfnl_pool_info_Ptr group_info_ptr = pool_table_ptr->my_pool;  // point at first pool group slot in table

     for(i=0; i < QFNL_MAX_POOL; i++) { // for all items in table

      memset (statusString,'\0',sizeof(statusString)); // initialize to nulls

      memset (pool_section,' ',sizeof(pool_section)-1);
      pool_section[QFNL_STPOOL_SIZE] = '\0';

      if ( (group_info_ptr->group_number != 0) &&
           (group_info_ptr->group_flag != QFNL_GROUP_STAT_FREESLOT) ) { // if found an in-use group in the table, give its status

          at_least_one = true;  // we found at least one group to report status on

          qfnl_check_pool_status(statusString, group_info_ptr);  // get its actual status to include in display

          memcpy(pool_section, group_info_ptr->stpool, strlen(group_info_ptr->stpool));  // get pool section for display

          printMsg(60,0,WTOPC_CHAIN,

              "%-5d %s           %s  %.16lX %.16lX %ld",                        //d26488
              group_info_ptr->group_number,
              pool_section,
              statusString,
              group_info_ptr->group_stats.readCount,
              group_info_ptr->group_stats.writeCount,
              group_info_ptr->group_stats.errCount);

      } // end if found an in-use group in the table

      group_info_ptr++; // advance to next slot in table

     } // end for all items in table

     // if at least one item was found in the table that we could report status on, put this message to indicate that counts were in hex
     //    otherwise put out a line that says there are no groups to report status on

     if (at_least_one) {
         printMsg(60,0,WTOPC_CHAIN," * = ALL COUNTS WITH STAR (*) ARE IN HEX, ALL OTHER NUMBERS DECIMAL");
         printMsg(60,0,WTOPC_CHAIN,"For Additional Info enter ZTEST POOL STATUS GROUP-n for each group number n");
     } else
         printMsg(60,0,WTOPC_CHAIN,"THERE ARE NO IN-USE GROUPS IN MEMORY FOR THIS PROCESSOR TO REPORT STATUS ON");

     printMsg(60,0,WTOPC_CHAIN_END,"END OF SUMMARY STATUS DISPLAY FOR POOL GROUPS");

  } // end else there could be pool groups in memory

  return;  // return to caller

} // end of qfnl_pool_general_status

static void qfnl_pool_status(struct IPRSE_output *parse,
                             int num_parms)
{

 char statusString[15];                            // will be used for status display
 memset (statusString,'\0',sizeof(statusString)); // initialize to nulls

 char pool_section[QFNL_STPOOL_SIZE+1];
 memset (pool_section,'\0',sizeof(pool_section));

 // parse the options specified for the ZTEST FFACT POOL STATUS command...
 if ( qfnl_parse_pool_status(parse,num_parms) == 0 )
   return;    // return if parser did not like something

 if (my_pool_status.group == 0) { // if no group specified on STATUS command
    qfnl_pool_general_status();   //   just give general status for all groups in memory
    return;
 }

 qfnl_pool_info_Ptr found_pool_info_ptr = qfnl_locate_pool_group(my_pool_status.group, TRUE);
 if (found_pool_info_ptr == NULL) {  // goup not found
   printMsg(55,'E',WTOPC_NO_CHAIN,"NO STATUS FOR POOL GROUP %d: GROUP INFO FROM BUILD NOT FOUND",
            my_pool_status.group);
   return;
 } else {  // group found

//   printMsg(2,'I',WTOPC_NO_CHAIN, "ADDRESS OF POOL GROUP %d IS %.16lX",
//              my_pool_status.group, (long) found_pool_info_ptr);

   // determine status of the group from its entry in the table

   qfnl_check_pool_status(statusString, found_pool_info_ptr);

   // copy the pool section from the group

   memcpy (pool_section, found_pool_info_ptr->stpool, QFNL_STPOOL_SIZE);

   // print out status

   printMsg(56,'I',WTOPC_CHAIN,"STATUS OF POOL GROUP %d IS %s",
                   my_pool_status.group, statusString);
   printMsg(56,0,WTOPC_CHAIN,"Additional details follow:");
   printMsg(56,0,WTOPC_CHAIN,"  ");
   printMsg(56,0,WTOPC_CHAIN, "ADDRESS OF POOL GROUP %d IS %.16lX",
              my_pool_status.group, (long) found_pool_info_ptr);
   printMsg(56,0,WTOPC_CHAIN,"  ");
   printMsg(56,0,WTOPC_CHAIN, "ADDRESS OF TABLE OF POOLS IN THIS GROUP IS %.16lX",
              (long) found_pool_info_ptr->pool_batch_ptr);
   printMsg(56,0,WTOPC_CHAIN,"  ");
   printMsg(56,0,WTOPC_CHAIN,"POOL SECTION = %s",
                pool_section);

   printMsg(56,0,WTOPC_CHAIN, "RECORD ID (HEX) = %02x%02x",
                found_pool_info_ptr->hex_rid[0], found_pool_info_ptr->hex_rid[1]);
   printMsg(56,0,WTOPC_CHAIN,"NUMBER OF POOLS IN THIS GROUP = %d",
                found_pool_info_ptr->number_records);
   printMsg(56,0,WTOPC_CHAIN,"RECORD SIZE = %d",
                found_pool_info_ptr->record_size);
   printMsg(56,0,WTOPC_CHAIN,"EXPIRY IN MINUTES = %d",
                found_pool_info_ptr->expiry/60);
   printMsg(56,0,WTOPC_CHAIN,"NUMBER OF FINDS = %ld",                           //d26352
                found_pool_info_ptr->group_stats.readCount);
   printMsg(56,0,WTOPC_CHAIN,"NUMBER OF HOLDS = %ld",                           //d26352
                found_pool_info_ptr->group_stats.holdCount);
   printMsg(56,0,WTOPC_CHAIN,"NUMBER OF FILES = %ld",                           //d26352
                found_pool_info_ptr->group_stats.writeCount);
   printMsg(56,0,WTOPC_CHAIN,"NUMBER OF INSTANCES = %ld",                       //d26352
                found_pool_info_ptr->group_stats.instanceCount);
   printMsg(56,0,WTOPC_CHAIN,"NUMBER OF ECBs CREATED = %lu",                    //d26352
                found_pool_info_ptr->group_stats.ecbCreateCount);
   printMsg(56,0,WTOPC_CHAIN,"NUMBER OF ECBs EXITED = %lu",                     //d26352
                found_pool_info_ptr->group_stats.ecbExitCount);
   printMsg(56,0,WTOPC_CHAIN,"NUMBER OF ECBs CURRENTLY ACTIVE = %lu",           //d26352
                found_pool_info_ptr->group_stats.ecbCount);
   printMsg(56,0,WTOPC_CHAIN,"NUMBER OF ERRORS = %ld",                          //d26352
                found_pool_info_ptr->group_stats.errCount);
   if (found_pool_info_ptr->group_stats.errCount > 0) {
     printMsg(56,0,WTOPC_CHAIN,"     BREAKDOWN OF ERRORS:");
     printMsg(56,0,WTOPC_CHAIN,"             MALLOC ERRORS = %ld",              //d26352
                found_pool_info_ptr->group_stats.mallocErr);
     printMsg(56,0,WTOPC_CHAIN,"             ENVIRONMENTAL ERRORS = %ld",       //d26352
                found_pool_info_ptr->group_stats.envErr);
     printMsg(56,0,WTOPC_CHAIN,"             I/O ERRORS = %ld",                 //d26352
                found_pool_info_ptr->group_stats.ioErr);
     printMsg(56,0,WTOPC_CHAIN,"             RECORD CONTENT ERRORS = %ld",      //d26352
                found_pool_info_ptr->group_stats.contentErr);
   }

   if (found_pool_info_ptr->group_flag == QFNL_GROUP_STAT_STARTED)
     printMsg(56,0,WTOPC_CHAIN,"LOCKING IS CURRENTLY %s",
                found_pool_info_ptr->group_stats.lockingActive ? "ON" : "OFF");


   printMsg(56,0,WTOPC_CHAIN_END,"END OF STATUS FOR POOL GROUP %d", my_pool_status.group);


 } // end else group found

 return;

} // end of function qfnl_pool_status

int qfnl_parse_pool_list(struct IPRSE_output *parse,
                          int num_parms)
{

// NOTE: This function parses parameters for the ZTEST POOL LIST command, and stores there
//   values into a structure named my_pool_list.

// This function returns 0 if an error was detected (and LIST should not be allowed) or
//                       1 on success (and the caller may continue LIST processing).


 memset (my_pool_list.groupString, '\0', sizeof(my_pool_list.groupString));  // initialize string to nulls
 my_pool_list.number_records = 0;  // initialize to zero records to list

                          //
 my_pool_list.group = 0;  // assign zero for group number until it is parsed...
                          //   if it remains that way, the caller may interpret the LIST request
                          //   to be an abbreviated request for all groups in memory
                          //

 while ((num_parms > 0) && parse)
  { // while there are more parameters

   if (debug) printMsg(9, 'I', WTOPC_NO_CHAIN, "!!! PARSE PARM IS %s", parse->IPRSE_parameter);

   if (parse && (strncmp(parse->IPRSE_parameter, "Group-",6) == 0))
   { // if Group specified

    // get the group number and save it both as an integer and as a string

    my_pool_list.group = atoi(parse->IPRSE_value);
    memcpy(my_pool_list.groupString, parse->IPRSE_value, strlen(parse->IPRSE_value));

    if (my_pool_list.group <=0) {
      printMsg(53,'E',WTOPC_NO_CHAIN,"NO LIST FOR POOL GROUP %d:  GROUP NUMBER MUST BE GREATER THAN 0", my_pool_list.group);
      return(0);
    }

    //  The parser should prevent an integer greater than 5 digits, but I coded a check anyway just in case

    if (my_pool_list.group > 99999) {
      printMsg(54,'E',WTOPC_NO_CHAIN,"NO LIST FOR POOL GROUP %d:  MAXIMUM GROUP NUMBER IS 99999", my_pool_list.group);
      return(0);
    }

   } // end if Group specified

   if (parse && (strncmp(parse->IPRSE_parameter, "Num-",4) == 0))
   { // if Num specified

    // get input value number of records for this group

    my_pool_list.number_records = atoi(parse->IPRSE_value);

    if (debug) printMsg(1,'I',WTOPC_NO_CHAIN,"!!! NUMBER OF RECORDS IN DECIMAL = %d !!!",
                 my_pool_list.number_records);

    if (my_pool_list.number_records == 0) { // if zero records requested
      printMsg(95,'E',WTOPC_NO_CHAIN,"IT DOES NOT MAKE SENSE TO LIST ZERO RECORDS");
      return(0);
    }

   } // end if Num specified

   parse = parse->IPRSE_next;
   num_parms--;

  } // end while there are more parameters

 return(1);  // return successful parse
} // end of function qfnl_parse_pool_list

static void qfnl_pool_list(struct IPRSE_output *parse,
                           int num_parms)
{

 char errMsg[100] = ""; // use to generate error messages

 // parse the options specified for the ZTEST FFACT POOL LIST command...
 if ( qfnl_parse_pool_list(parse,num_parms) == 0 )
   return;    // return if parser did not like something

 qfnl_pool_info_Ptr found_pool_info_ptr = qfnl_locate_pool_group(my_pool_list.group, TRUE);
 if (found_pool_info_ptr == NULL) {  // goup not found
   printMsg(96,'E',WTOPC_NO_CHAIN,"CANNOT LIST RECORDS IN POOL GROUP %d: GROUP INFO FROM BUILD NOT FOUND",
            my_pool_list.group);
   return;
 } else {  // group found

   // If the user requested to list more records for the group than currently exist, warn him or her that
   //   only the number of records currently in the group will be listed...

   unsigned int number_to_list = my_pool_list.number_records;  // get requested number of records to list

   if (number_to_list > found_pool_info_ptr->number_records) { // if requested more records than exist

      printMsg(97,'W',WTOPC_NO_CHAIN,"FOR POOL GROUP %d REQUESTED DISPLAY OF MORE RECORDS THAN CURRNTLY EXIST, WILL ONLY DISPLAY %d RECORDS",
             my_pool_list.group, found_pool_info_ptr->number_records);

      number_to_list = found_pool_info_ptr->number_records; // adjust number records to display = number records currently exist for group

   } // end if requested more records than exist

   unsigned int zero_based_index;  // used to index thru table of pool records for this group...

 /* --------------------------- Beginning of Logic to List File Addresses ---------------------------------------------------------------- */

   qfnl_pool_batch_Ptr pool_table = found_pool_info_ptr->pool_batch_ptr;  // point at table of pools for this group

   if (pool_table == NULL ) { // This is unexpected, but if the pool table isn't available then bail out!!!
       sprintf(errMsg,"LIST REQUEST FAILED: TABLE OF POOLS NOT AVAILABLE FOR POOL GROUP %d",
                       found_pool_info_ptr->group_number);
       printMsg(91, 'E', WTOPC_NO_CHAIN, errMsg);
       return;
   } else {  // pool table not NULL

      printMsg(98,'I',WTOPC_CHAIN,"LIST OF %d RECORDS IN POOL GROUP %d:",
                     my_pool_list.number_records, my_pool_list.group);

      char * index_item = &(pool_table->first_item);  // point at first item in table
      qfnl_pool_item_Ptr pool_item = (qfnl_pool_item_Ptr) index_item;  // access first item as item in pool group

      char output_line[80]; // line of file addresses to list
      memset(output_line,'\0',sizeof(output_line));
      output_line[0] = ' '; // set first character to blank

      for (zero_based_index = 0; zero_based_index < number_to_list; zero_based_index++) { // for each record to list

        sprintf(output_line,"%s %.16lX ",output_line,pool_item->FA);  // add a file address to the output line

        // If we have 4 file addresses, print an output line and re-initialize the next output line

        if ( ((zero_based_index+1)%4 == 0) ||
             (zero_based_index == number_to_list-1) ) {
          // If number of file addresses accessed is divisible by 4, or this is the last file address to list,
          //    output a line...
                  printMsg(98,0,WTOPC_CHAIN,"%s",output_line);
          //    re-initialize next line for output
                  memset(output_line,'\0',sizeof(output_line));
                  output_line[0] = ' '; // set first character to blank
        }

        if ( zero_based_index < (number_to_list-1) ) { // only advance if we will iterate thru loop again
             index_item = index_item + sizeof(qfnl_pool_item); // point at next item to list
             pool_item = (qfnl_pool_item_Ptr) index_item;  // access it as item in pool group
        }

      } // END for each record to list

      printMsg(98,0,WTOPC_CHAIN_END,"END OF LIST OF FILE ADDRESSES TO DISPLAY FOR GROUP %d", my_pool_list.group);

   } // end else pool table not NULL

 /* ---------------------------       End of Logic to List File Addresses ---------------------------------------------------------------- */

 } // end else group found

 return;

} // end of function qfnl_pool_list

int qfnl_parse_pool_stop(struct IPRSE_output *parse,
                            int num_parms)
{

// NOTE: This function parses parameters for the ZTEST POOL STOP command, and stores there
//   values into a structure named my_pool_stop.

// This function returns 0 if an error was detected (and STOP should not be allowed) or
//                       1 on success (and the caller may continue STOP processing).


 my_pool_stop.group = 0; // assign zero for group number until it is parsed...
                         // NOTE: If it remains zero we will interpret this as a request
                         //  to stop all pool activity on this processor (all pool groups)

 while ((num_parms > 0) && parse)
  { // while there are more parameters

   if (debug) printMsg(9, 'I', WTOPC_NO_CHAIN, "!!! PARSE PARM IS %s", parse->IPRSE_parameter);

   if (parse && (strncmp(parse->IPRSE_parameter, "Group-",6) == 0))
   { // if Group specified

    // get the group number and save it both as an integer and as a string

    my_pool_stop.group = atoi(parse->IPRSE_value);

    if (my_pool_stop.group <=0) {
      printMsg(62,'E',WTOPC_NO_CHAIN,"STOP COMMAND FOR GROUP %d REJECTED:  GROUP NUMBER MUST BE GREATER THAN 0", my_pool_stop.group);
      return(0);
    }

    //  The parser should prevent an integer greater than 5 digits, but I coded a check anyway just in case

    if (my_pool_status.group > 99999) {
      printMsg(63,'E',WTOPC_NO_CHAIN,"STOP COMMAND FOR GROUP %d REJECTED:  MAXIMUM GROUP NUMBER IS 99999", my_pool_stop.group);
      return(0);
    }

   } // end if Group specified

   parse = parse->IPRSE_next;
   num_parms--;

  } // end while there are more parameters

 return(1);  // return successful parse
} // end of function qfnl_parse_pool_stop

void qfnl_pool_stop_all_groups(void)
{

// This function will loop through the table of POOL Groups
//  and mark them all as STOPPED so that all activity on them stops.

  qfnl_pool_ctl_tbl_Ptr pool_table_ptr = get_pool_table();

  if (pool_table_ptr == NULL) {  // nothing in memory
     printMsg(93,'I',WTOPC_NO_CHAIN,"THERE ARE NO POOL GROUPS DEFINED FOR THIS PROCESSOR");
     return;
  } else { // there could be pool groups in memory

     int i;
     boolean at_least_one = false; // assume we do not have at least one group until found

     qfnl_pool_info_Ptr group_info_ptr = pool_table_ptr->my_pool;  // point at first pool group slot in table

     for(i=0; i < QFNL_MAX_POOL; i++) { // for all items in table

      if ( (group_info_ptr->group_number != 0) &&
           (group_info_ptr->group_flag != QFNL_GROUP_STAT_FREESLOT) ) { // if found an in-use group in the table, set it to STOPPED

          at_least_one = true;  // we found at least one group in the table

          if (group_info_ptr->group_flag == QFNL_GROUP_STAT_BUILDING)
              printMsg(59,'W',WTOPC_NO_CHAIN,"POOL GROUP %d BUILD IN PROGRESS: STOP ALLOWED BUT MIGHT TERMINATE BUILD",
                    group_info_ptr->group_number);

          tpf_stpoc(TPF_STPOC_ON, stpoc_value);  // allow ourselves to modify key 9 storage gotten by tpf_gsysc
          group_info_ptr->group_flag = QFNL_GROUP_STAT_STOPPED;   // set indicator that the group is now STOPPED
          tpf_stpoc(TPF_STPOC_OFF, stpoc_value); // done modifying for now

          if (group_info_ptr->group_stats.instanceCount == 0) {
              printMsg(57,'I',WTOPC_NO_CHAIN,"POOL GROUP %d STOPPED", group_info_ptr->group_number);
          } else {
              printMsg(61,'I',WTOPC_NO_CHAIN,"POOL GROUP %d STOPPING", group_info_ptr->group_number);
          }

      } // end if found an in-use group in the table

      group_info_ptr++; // advance to next slot in table

     } // end for all items in table

     // if there wasn't at least one group in memory that we could stop, tell the user

     if (!at_least_one) {
         printMsg(93,'I',WTOPC_NO_CHAIN,"THERE ARE NO POOL GROUPS DEFINED FOR THIS PROCESSOR");
     }

  } // end else there could be pool groups in memory

  return;  // return to caller

} // qfnl_pool_stop_all_groups

static void qfnl_pool_stop(struct IPRSE_output *parse,
                             int num_parms)
{

 // parse the ZTEST FFACT POOL STOP command...

 if ( qfnl_parse_pool_stop(parse,num_parms) == 0 )
   return;    // return if parser did not like something

 // if group was set to zero it means the user did not
 //   specify a group so we will stop activity for all
 //   POOL groups

 if (my_pool_stop.group == 0) {
   qfnl_pool_stop_all_groups();
   return;
 }

 // attempt to locate the group being stopped

 qfnl_pool_info_Ptr found_pool_info_ptr = qfnl_locate_pool_group(my_pool_stop.group, FALSE);
 if (found_pool_info_ptr == NULL) {  // goup not found

   printMsg(64,'E',WTOPC_NO_CHAIN,"CANNOT STOP POOL GROUP %d: GROUP NOT FOUND IN MEMORY",
            my_pool_stop.group);
   return;

 } else {  // group found

   // if the group has the indication on that a BUILD is in progress, we will allow you to set
   //    the group to being STOPPED state, but that does not stop any build that may still be
   //    ongoing.  (NOTE: When the build is in done it should set the indicator back to STOPPED
   //    again.)

   if (found_pool_info_ptr->group_flag == QFNL_GROUP_STAT_BUILDING)
     printMsg(59,'W',WTOPC_NO_CHAIN,"POOL GROUP %d BUILD IN PROGRESS: STOP ALLOWED BUT MIGHT TERMINATE BUILD",
              my_pool_stop.group);

   tpf_stpoc(TPF_STPOC_ON, stpoc_value);  // allow ourselves to modify key 9 storage gotten by tpf_gsysc
   found_pool_info_ptr->group_flag = QFNL_GROUP_STAT_STOPPED;   // set indicator that the group is now STOPPED
   tpf_stpoc(TPF_STPOC_OFF, stpoc_value); // done modifying for now

   if (found_pool_info_ptr->group_stats.instanceCount == 0) {
     printMsg(57,'I',WTOPC_NO_CHAIN,"POOL GROUP %d STOPPED", my_pool_start.group);
   } else {
     printMsg(61,'I',WTOPC_NO_CHAIN,"POOL GROUP %d STOPPING", my_pool_stop.group);
   }

 } // end else group found

} // end function qfnl_pool_stop

int qfnl_parse_pool_kill(struct IPRSE_output *parse,
                            int num_parms)
{

// NOTE: This function parses parameters for the ZTEST POOL KILL command, and stores there
//   values into a structure named my_pool_kill.

// This function returns 0 if an error was detected (and KILL should not be allowed) or
//                       1 on success (and the caller may continue KILL processing).


 memset (my_pool_kill.groupString, '\0', sizeof(my_pool_kill.groupString));  // initialize string to nulls

 my_pool_kill.group = 0; // assign zero for group number until it is parsed...

 while ((num_parms > 0) && parse)
  { // while there are more parameters

   if (debug) printMsg(9, 'I', WTOPC_NO_CHAIN, "!!! PARSE PARM IS %s", parse->IPRSE_parameter);

   if (parse && (strncmp(parse->IPRSE_parameter, "Group-",6) == 0))
   { // if Group specified

    // get the group number and save it both as an integer and as a string

    my_pool_kill.group = atoi(parse->IPRSE_value);
    memcpy(my_pool_kill.groupString, parse->IPRSE_value, strlen(parse->IPRSE_value));

    if (my_pool_kill.group <=0) {
      printMsg(67,'E',WTOPC_NO_CHAIN,"CANNOT KILL POOL GROUP %d:  GROUP NUMBER MUST BE GREATER THAN 0", my_pool_kill.group);
      return(0);
    }

    //  The parser should prevent an integer greater than 5 digits, but I coded a check anyway just in case

    if (my_pool_kill.group > 99999) {
      printMsg(69,'E',WTOPC_NO_CHAIN,"CANNOT KILL POOL GROUP %d:  MAXIMUM GROUP NUMBER IS 99999", my_pool_kill.group);
      return(0);
    }

   } // end if Group specified

   parse = parse->IPRSE_next;
   num_parms--;

  } // end while there are more parameters

 return(1);  // return successful parse
} // end of function qfnl_parse_pool_kill

static void qfnl_pool_kill(struct IPRSE_output *parse,
                            int num_parms)
{

 char errMsg[120] = "";                                                         //d26352

 boolean killed_something = false; // assume nothing was killed for this group until something is removed for it

 // parse the ZTEST FFACT POOL KILL command...

 if ( qfnl_parse_pool_kill(parse,num_parms) == 0 )
   return;    // return if parser did not like something

 // attempt to locate the group being killed in memory

 qfnl_pool_info_Ptr found_pool_info_ptr = qfnl_locate_pool_group(my_pool_kill.group, FALSE);
 if (found_pool_info_ptr == NULL) {  // goup not found

   printMsg(71,'I',WTOPC_NO_CHAIN,"POOL GROUP %d NOT FOUND IN MEMORY",
            my_pool_kill.group);

 } else {  // group found in memory, free its slot in the table as long as it is in STOPPED state

   if (found_pool_info_ptr->group_flag != QFNL_GROUP_STAT_STOPPED) { // if group in memory not STOPPED, do not allow KILL

      printMsg(72,'E',WTOPC_NO_CHAIN,"YOU MUST STOP POOL GROUP %d ON THIS PROCESSOR BEFORE YOU CAN KILL IT HERE",
               my_pool_kill.group);
      return;
   } // end if group in memory not STOPPED

   // The POOL group is STOPPED, so we will now do the following:
   //   -set its slot in the Pool Group table to FREE
   //   -free the table of pool record file addresses (pool "batch")
   //   -clean up the slot in the Pool Group table by setting it to all zeroes
   //   -decrement the total count of POOL groups allocated

   // ... but first lock the group so that we don't pull the rug from under expiry (which periodically updates entries
   //         in the group to flag them as invalid) ...

   corhc(found_pool_info_ptr);  // lock group

   tpf_stpoc(TPF_STPOC_ON, stpoc_value);  // allow modify key 9

   found_pool_info_ptr->group_flag = QFNL_GROUP_STAT_FREESLOT;

   // NOTE: BEFORE WE DO THIS WE ZERO OUT THE SLOT IN THE POOL GROUP TABLE,
   //  WE MUST RELEASE THE SYSTEM HEAP THAT IS USED TO STORE THE POOL
   //  ADDRESSES FOR THIS GROUP (THIS HEAP IS aka THE POOL "BATCH"

   if (found_pool_info_ptr->pool_batch_ptr != NULL) { // if pointer to system heap not NULL

          // free the storage that was allocated for the batch of pools and set its pointer to NULL

          uint frames = found_pool_info_ptr->pool_batch_ptr->num_frames;
          char token[ITOKENLEN];
          memcpy(token, found_pool_info_ptr->pool_batch_ptr->token, sizeof(token));

          if ( tpf_rsysc( (void *) found_pool_info_ptr->pool_batch_ptr,
                          frames,
                          token ) != RSYSC_OK ) {
             sprintf(errMsg,"FAILURE TO FREE STORAGE FOR POOLS FOR POOL GROUP %d; ERRNO = %d", my_pool_kill.group, errno);
             printMsg(90, 'E', WTOPC_NO_CHAIN, errMsg);
          } else {
             found_pool_info_ptr->pool_batch_ptr = NULL;
             printMsg(94, 'I', WTOPC_NO_CHAIN, "FREED SYSTEM HEAP CONTAINING POOLS FOR POOL GROUP %d", my_pool_kill.group);
          }

   } // end if pointer to system heap not NULL

   memset(found_pool_info_ptr, 0x00, sizeof(qfnl_pool_info));  // zero out the entire slot in Pool Group Table

   coruc(found_pool_info_ptr);  // unlock the group now that we are done freeing it

   // Decrement count of pool file groups that the driver will remember

   qfnl_pool_ctl_tbl_Ptr  pool_table_ptr = get_pool_table();

   if (pool_table_ptr == NULL) {
      sprintf(errMsg,"ON KILL OF POOL GROUP %d, FAILURE TO LOCATE POOL GROUP TABLE TO DECREMENT COUNT OF POOL GROUPS IN MEMORY",my_pool_kill.group);
      printMsg(92, 'E', WTOPC_NO_CHAIN, errMsg);
   } else
      pool_table_ptr->number_of_groups--;

   tpf_stpoc(TPF_STPOC_OFF, stpoc_value);  // done with modification

   killed_something = true;  // we killed something for this group

   // Inform the user that memory has been cleared for this group

   printMsg(74,'I',WTOPC_NO_CHAIN,"MEMORY FOR POOL GROUP %d HAS BEEN CLEARED", my_pool_kill.group);

 } // end group found in memory

 if (killed_something)
   printMsg(66,'I',WTOPC_NO_CHAIN,"POOL GROUP %d KILLED ON THIS PROCESSOR", my_pool_kill.group);
 else
   printMsg(77,'I',WTOPC_NO_CHAIN,"NOTHING TO REMOVE FOR POOL GROUP %d", my_pool_kill.group);

 return;

} // end qfnl_pool_kill

int qfnl_parse_pool_clear(struct IPRSE_output *parse,
                            int num_parms)
{

// NOTE: This function parses parameters for the ZTEST POOL CLEAR command, and stores there
//   values into a structure named my_pool_clear.

// This function returns 0 if an error was detected (and CLEAR should not be allowed) or
//                       1 on success (and the caller may continue CLEAR processing).

 my_pool_clear.group = 0; // assign zero for group number until it is parsed...

 while ((num_parms > 0) && parse)
  { // while there are more parameters

   if (debug) printMsg(9, 'I', WTOPC_NO_CHAIN, "!!! PARSE PARM IS %s", parse->IPRSE_parameter);

   if (parse && (strncmp(parse->IPRSE_parameter, "Group-",6) == 0))
   { // if Group specified

    // get the group number and save it both as an integer and as a string

    my_pool_clear.group = atoi(parse->IPRSE_value);

    if (my_pool_clear.group <=0) {
      printMsg(68,'E',WTOPC_NO_CHAIN,"CANNOT CLEAR POOL GROUP %d:  GROUP NUMBER MUST BE GREATER THAN 0", my_pool_clear.group);
      return(0);
    }

    //  The parser should prevent an integer greater than 5 digits, but I coded a check anyway just in case

    if (my_pool_clear.group > 99999) {
      printMsg(70,'E',WTOPC_NO_CHAIN,"CANNOT CLEAR POOL GROUP %d:  MAXIMUM GROUP NUMBER IS 99999", my_pool_clear.group);
      return(0);
    }

   } // end if Group specified

   parse = parse->IPRSE_next;
   num_parms--;

  } // end while there are more parameters

 return(1);  // return successful parse
} // end of function qfnl_parse_pool_clear

static void qfnl_pool_clear(struct IPRSE_output *parse,
                            int num_parms)
{

 boolean cleared_something = false; // assume nothing was cleared for this group until something is removed for it

 // parse the ZTEST FFACT POOL CLEAR command...

 if ( qfnl_parse_pool_clear(parse,num_parms) == 0 )
   return;    // return if parser did not like something

 // attempt to locate the group whose counts are being cleared in memory

 qfnl_pool_info_Ptr found_pool_info_ptr = qfnl_locate_pool_group(my_pool_clear.group, FALSE);
 if (found_pool_info_ptr == NULL) {  // goup not found

   printMsg(71,'I',WTOPC_NO_CHAIN,"POOL GROUP %d NOT FOUND IN MEMORY",
            my_pool_clear.group);

 } else {  // group found in memory, free its slot in the table as long as it is in STOPPED state

   if (found_pool_info_ptr->group_flag != QFNL_GROUP_STAT_STOPPED) { // if group in memory not STOPPED, do not allow CLEAR

      printMsg(73,'E',WTOPC_NO_CHAIN,"YOU MUST STOP POOL GROUP %d ON THIS PROCESSOR BEFORE YOU CAN CLEAR ITS COUNTERS",
               my_pool_clear.group);
      return;
   } // end if group in memory not STOPPED

   // The pool group is STOPPED, so we can now clear its counters

   tpf_stpoc(TPF_STPOC_ON, stpoc_value);  // allow modify key 9
   memset(&(found_pool_info_ptr->group_stats), 0x00, sizeof(qfnl_group_stats));
   tpf_stpoc(TPF_STPOC_OFF, stpoc_value);  // done with modification

   cleared_something = true;  // we cleared counters for this group

 } // end group found in memory

 if (cleared_something)
   printMsg(65,'I',WTOPC_NO_CHAIN,"COUNTS FOR POOL GROUP %d CLEARED ON THIS PROCESSOR", my_pool_clear.group);
 else
   printMsg(78,'E',WTOPC_NO_CHAIN,"NO COUNTS TO CLEAR FOR POOL GROUP %d ON THIS PROCESSOR", my_pool_clear.group);

 return;

} // end qfnl_pool_clear

static void qfnl_help(struct IPRSE_output *parse,
                             int num_parms)
{

    boolean is_for_fixed = false; // assume it is not for FIXED until you parse it
    boolean is_for_pool = false;  // assume it is not for POOL until you parse it

    /******************************************************************/
    /* Process all HELP parameters one by one                         */
    /******************************************************************/

    parse = parse->IPRSE_next;
    num_parms--;

    if (num_parms == 0) {

     printMsg(0,'I',WTOPC_CHAIN,"FIND FILE DRIVER HELP");
     printMsg(0,'I',WTOPC_CHAIN,"FOR HELP ON THE VARIOUS OPTIONS, ENTER ONE OF THE FOLLOWING:");
     printMsg(0,'I',WTOPC_CHAIN,"  ZTEST FFACt HELP Fixed");
     printMsg(0,'I',WTOPC_CHAIN,"       For info on how to work with FIXED FILE records");
     printMsg(0,'I',WTOPC_CHAIN,"  ZTEST FFACt HELP Pool");
     printMsg(0,'I',WTOPC_CHAIN,"       For info on how to work with POOL records");
     printMsg(0,'I',WTOPC_CHAIN_END,"END OF ZTEST FFACt HELP DISPLAY");

     exit(0);

    }

    if (parse && (strncmp(parse->IPRSE_parameter, "Fixed", 5) == 0)) {

     parse = parse->IPRSE_next;
     num_parms--;

     is_for_fixed = true;

     if (num_parms == 0) { // if last parm after "HELP" is "Fixed"

       printMsg(0,'I',WTOPC_CHAIN,"FIND FILE DRIVER HELP FOR FIXED");
       printMsg(0,'I',WTOPC_CHAIN,"FOR HELP ON THE VARIOUS OPTIONS, ENTER ONE OF THE FOLLOWING:");
       printMsg(0,'I',WTOPC_CHAIN,"  ZTEST FFACt HELP Fixed BUild");
       printMsg(0,'I',WTOPC_CHAIN,"       For info on how to BUILD a group of FIXED records");
       printMsg(0,'I',WTOPC_CHAIN,"  ZTEST FFACt HELP Fixed STARt");
       printMsg(0,'I',WTOPC_CHAIN,"       For info on how to START activity on a group of FIXED");
       printMsg(0,'I',WTOPC_CHAIN,"           records");
       printMsg(0,'I',WTOPC_CHAIN,"  ZTEST FFACt HELP Fixed STATus");
       printMsg(0,'I',WTOPC_CHAIN,"       For info on how to query STATUS on groups of FIXED");
       printMsg(0,'I',WTOPC_CHAIN,"           records");
       printMsg(0,'I',WTOPC_CHAIN,"  ZTEST FFACt HELP Fixed STOP");
       printMsg(0,'I',WTOPC_CHAIN,"       For info on how to STOP activity on a group of FIXED");
       printMsg(0,'I',WTOPC_CHAIN,"           records");
       printMsg(0,'I',WTOPC_CHAIN,"  ZTEST FFACt HELP Fixed CLEAR");
       printMsg(0,'I',WTOPC_CHAIN,"       For info on how to CLEAR counters for a group of");
       printMsg(0,'I',WTOPC_CHAIN,"           FIXED records after it is STOPped");
       printMsg(0,'I',WTOPC_CHAIN,"  ZTEST FFACt HELP Fixed KILL");
       printMsg(0,'I',WTOPC_CHAIN,"       For info on how to KILL (get the driver to forget)");
       printMsg(0,'I',WTOPC_CHAIN,"           a group of FIXED records");
       printMsg(0,'I',WTOPC_CHAIN_END,"END OF ZTEST FFACt HELP DISPLAY FOR FIXED");

       exit(0);

     } // end if last parm after "HELP" is "Fixed"

    } // end if "Fixed"

    if (parse && (strncmp(parse->IPRSE_parameter, "Pool", 4) == 0)) {

     parse = parse->IPRSE_next;
     num_parms--;

     is_for_pool = true;

     if (num_parms == 0) { // if last parm after "HELP" is "Pool"

       printMsg(0,'I',WTOPC_CHAIN,"FIND FILE DRIVER HELP FOR POOL");
       printMsg(0,'I',WTOPC_CHAIN,"FOR HELP ON THE VARIOUS OPTIONS, ENTER ONE OF THE FOLLOWING:");
       printMsg(0,'I',WTOPC_CHAIN,"  ZTEST FFACt HELP Pool BUild");
       printMsg(0,'I',WTOPC_CHAIN,"       For info on how to BUILD a group of POOL");
       printMsg(0,'I',WTOPC_CHAIN,"           records");
       printMsg(0,'I',WTOPC_CHAIN,"  ZTEST FFACt HELP Pool STARt");
       printMsg(0,'I',WTOPC_CHAIN,"       For info on how to START activity on a group of POOL");
       printMsg(0,'I',WTOPC_CHAIN,"           records");
       printMsg(0,'I',WTOPC_CHAIN,"  ZTEST FFACt HELP Pool STATus");
       printMsg(0,'I',WTOPC_CHAIN,"       For info on how to query STATUS on groups of POOL");
       printMsg(0,'I',WTOPC_CHAIN,"           records");
       printMsg(0,'I',WTOPC_CHAIN,"  ZTEST FFACt HELP Pool LIST");
       printMsg(0,'I',WTOPC_CHAIN,"       For info on how to LIST the pool file addresses");
       printMsg(0,'I',WTOPC_CHAIN,"           belonging to a specific POOL group");
       printMsg(0,'I',WTOPC_CHAIN,"  ZTEST FFACt HELP Pool STOP");
       printMsg(0,'I',WTOPC_CHAIN,"       For info on how to STOP activity on a group of POOL");
       printMsg(0,'I',WTOPC_CHAIN,"           records");
       printMsg(0,'I',WTOPC_CHAIN,"  ZTEST FFACt HELP Pool CLEAR");
       printMsg(0,'I',WTOPC_CHAIN,"       For info on how to CLEAR counters for a group of");
       printMsg(0,'I',WTOPC_CHAIN,"           POOL records after it is STOPped");
       printMsg(0,'I',WTOPC_CHAIN,"  ZTEST FFACt HELP Pool KILL");
       printMsg(0,'I',WTOPC_CHAIN,"       For info on how to KILL (get the driver to forget)");
       printMsg(0,'I',WTOPC_CHAIN,"           a group of POOL records");
       printMsg(0,'I',WTOPC_CHAIN_END,"END OF ZTEST FFACt HELP DISPLAY FOR POOL");

       exit(0);

     } // end if last parm after "HELP" is "Pool"

    } // end if "Pool"

    num_parms--;  // decrement count as next item after FIXED/POOL has already been parsed

    if (is_for_fixed) {  // if additional parameter on HELP for "FIXED"

       if (parse && (strncmp(parse->IPRSE_parameter, "BUild", 5) == 0)) { // If next parm is "BUILD"
         if (num_parms == 0) { // if last parm after "FIXED" is "BUILD"

           printMsg(0,'I',WTOPC_CHAIN,"FIND FILE DRIVER SYNTAX FOR FIXED BUILD");
           printMsg(0,'I',WTOPC_CHAIN," ");
           printMsg(0,'I',WTOPC_CHAIN,"  ZTEST FFACt Fixed BUild Group-n Type-type RID-rid");
           printMsg(0,'I',WTOPC_CHAIN,"           Begin-ord1 End-ord2 {Decb-count}");             //d26633
           printMsg(0,'I',WTOPC_CHAIN," ");
           printMsg(0,'I',WTOPC_CHAIN,"        where:");
           printMsg(0,'I',WTOPC_CHAIN," ");
           printMsg(0,'I',WTOPC_CHAIN,"         n = an integer between 1 and 99999 which is");
           printMsg(0,'I',WTOPC_CHAIN,"              the number to assign to this FIXED group");
           printMsg(0,'I',WTOPC_CHAIN,"         type = a string with the record type for the group");
           printMsg(0,'I',WTOPC_CHAIN,"              being built");
           printMsg(0,'I',WTOPC_CHAIN,"         rid = 4 hex digits = the 2-byte record ID to be");
           printMsg(0,'I',WTOPC_CHAIN,"              used to initialize the records");
           printMsg(0,'I',WTOPC_CHAIN,"         ord1 = 1 to 16 hex digits for the beginning ordinal");
           printMsg(0,'I',WTOPC_CHAIN,"              of the group");
           printMsg(0,'I',WTOPC_CHAIN,"         ord2 = 1 to 16 hex digits for the ending ordinal");
           printMsg(0,'I',WTOPC_CHAIN,"              of the group");
           printMsg(0,'I',WTOPC_CHAIN,"         count = number of DECBs used to initiliaze the"); //d26633
           printMsg(0,'I',WTOPC_CHAIN,"              ordinals.  The default is 10.");             //d26633
           printMsg(0,'I',WTOPC_CHAIN," ");
           printMsg(0,'I',WTOPC_CHAIN_END,"END OF ZTEST FFACt SYNTAX DISPLAY FOR FIXED BUILD");

           exit(0);
         } // if last parm after "FIXED" is "BUILD"

       } // end If next parm is "BUILD"

       if (parse && (strncmp(parse->IPRSE_parameter, "STARt", 5) == 0)) { // If next parm is "START"
         if (num_parms == 0) { // if last parm after "FIXED" is "START"

           printMsg(0,'I',WTOPC_CHAIN,"FIND FILE DRIVER SYNTAX FOR FIXED START");
           printMsg(0,'I',WTOPC_CHAIN," ");
           printMsg(0,'I',WTOPC_CHAIN,"  ZTEST FFACt Fixed STARt Group-n SEED-s Read-r Write-w");
           printMsg(0,'I',WTOPC_CHAIN,"            ECBs-x {Lock-Y|N}");
           printMsg(0,'I',WTOPC_CHAIN,"            {LOw-Y|I} {SWitch-Y|I}");
           printMsg(0,'I',WTOPC_CHAIN,"            {UTILclass-uClass_name} ");
           printMsg(0,'I',WTOPC_CHAIN,"            {RESourceclass-Y|N} ");
           printMsg(0,'I',WTOPC_CHAIN," ");
           printMsg(0,'I',WTOPC_CHAIN,"        where:");
           printMsg(0,'I',WTOPC_CHAIN," ");
           printMsg(0,'I',WTOPC_CHAIN,"         n = an integer between 1 and 99999 which is the");
           printMsg(0,'I',WTOPC_CHAIN,"              FIXED group number to start");
           printMsg(0,'I',WTOPC_CHAIN,"         s = decimal numeric seed from 0 to 999 used to");
           printMsg(0,'I',WTOPC_CHAIN,"              randomly access ordinals");
           printMsg(0,'I',WTOPC_CHAIN,"         r = decimal number of reads per ECB, from 0 to 999");
           printMsg(0,'I',WTOPC_CHAIN,"         w = decimal number of writes per ECB, from 0 to 999");
           printMsg(0,'I',WTOPC_CHAIN,"         x = decimal number of ECBs to spawn for each");
           printMsg(0,'I',WTOPC_CHAIN,"              iteration, from 1 to 999");
           printMsg(0,'I',WTOPC_CHAIN,"         Lock-Y (default) lock each record prior to writing");
           printMsg(0,'I',WTOPC_CHAIN,"              to it");
           printMsg(0,'I',WTOPC_CHAIN,"         Lock-N to bypass locking");
           printMsg(0,'I',WTOPC_CHAIN,"         LOw-Y sets the LOWPRIORITY attribute for the parent");   //d26696
           printMsg(0,'I',WTOPC_CHAIN,"              ECB ");
           printMsg(0,'I',WTOPC_CHAIN,"         LOw-I sets the LOWPRIORITY attribute for the parent");   //d26696
           printMsg(0,'I',WTOPC_CHAIN,"              ECB and its child ECBs will INHERIT it");
           printMsg(0,'I',WTOPC_CHAIN,"         SWitch-Y sets the SWITCHABLE attribute for the");
           printMsg(0,'I',WTOPC_CHAIN,"              parent ECB ");                                      //d26696
           printMsg(0,'I',WTOPC_CHAIN,"         SWitch-I sets the SWITCHABLE attribute for the");
           printMsg(0,'I',WTOPC_CHAIN,"              parent ECB and its child ECBs will");               //d26696
           printMsg(0,'I',WTOPC_CHAIN,"              INHERIT it");
           printMsg(0,'I',WTOPC_CHAIN,"         UTILclass-uClass_name causes this instance of the");
           printMsg(0,'I',WTOPC_CHAIN,"              drvier to use the specified utilization class");
           printMsg(0,'I',WTOPC_CHAIN,"         RESourceclass-Y this instance will use a LODIC ");
           printMsg(0,'I',WTOPC_CHAIN,"              resource class");
           printMsg(0,'I',WTOPC_CHAIN,"         RESourceclass-N this instance will NOT use a LODIC ");
           printMsg(0,'I',WTOPC_CHAIN,"              resource class -- THIS IS THE DEFAULT");
           printMsg(0,'I',WTOPC_CHAIN," ");
           printMsg(0,'I',WTOPC_CHAIN,"  NOTE: If Lock-Y the number of Reads must be greater than");
           printMsg(0,'I',WTOPC_CHAIN,"         or equal to the number of Writes");
           printMsg(0,'I',WTOPC_CHAIN," ");
           printMsg(0,'I',WTOPC_CHAIN_END,"END OF ZTEST FFACt SYNTAX DISPLAY FOR FIXED START");

           exit(0);
         } // if last parm after "FIXED" is "START"

       } // end If next parm is "START"

       if (parse && (strncmp(parse->IPRSE_parameter, "STATus", 6) == 0)) { // If next parm is "STATUS"
         if (num_parms == 0) { // if last parm after "FIXED" is "STATUS"

           printMsg(0,'I',WTOPC_CHAIN,"FIND FILE DRIVER SYNTAX FOR FIXED STATUS");
           printMsg(0,'I',WTOPC_CHAIN," ");
           printMsg(0,'I',WTOPC_CHAIN,"  ZTEST FFACt Fixed STATus {Group-n}");
           printMsg(0,'I',WTOPC_CHAIN," ");
           printMsg(0,'I',WTOPC_CHAIN,"        where:");
           printMsg(0,'I',WTOPC_CHAIN," ");
           printMsg(0,'I',WTOPC_CHAIN,"         (OPTIONAL) n = an integer between 1 and 99999 which");
           printMsg(0,'I',WTOPC_CHAIN,"                         is the FIXED group number to");
           printMsg(0,'I',WTOPC_CHAIN,"                         display status for");
           printMsg(0,'I',WTOPC_CHAIN," ");
           printMsg(0,'I',WTOPC_CHAIN,"  NOTE: If the Group is omitted, a status summary is");
           printMsg(0,'I',WTOPC_CHAIN,"         displayed for all groups in memory for");
           printMsg(0,'I',WTOPC_CHAIN,"         the current processor");
           printMsg(0,'I',WTOPC_CHAIN," ");
           printMsg(0,'I',WTOPC_CHAIN_END,"END OF ZTEST FFACt SYNTAX DISPLAY FOR FIXED STATUS");

           exit(0);
         } // if last parm after "FIXED" is "STATUS"

       } // end If next parm is "STATUS"

       if (parse && (strncmp(parse->IPRSE_parameter, "STOP", 4) == 0)) { // If next parm is "STOP"
         if (num_parms == 0) { // if last parm after "FIXED" is "STOP"

           printMsg(0,'I',WTOPC_CHAIN,"FIND FILE DRIVER SYNTAX FOR FIXED STOP");
           printMsg(0,'I',WTOPC_CHAIN," ");
           printMsg(0,'I',WTOPC_CHAIN,"  ZTEST FFACt Fixed STOP {Group-n}");
           printMsg(0,'I',WTOPC_CHAIN," ");
           printMsg(0,'I',WTOPC_CHAIN,"        where:");
           printMsg(0,'I',WTOPC_CHAIN," ");
           printMsg(0,'I',WTOPC_CHAIN,"         n = an integer between 1 and 99999 which is the");
           printMsg(0,'I',WTOPC_CHAIN,"              FIXED group number to stop");
           printMsg(0,'I',WTOPC_CHAIN," ");
           printMsg(0,'I',WTOPC_CHAIN,"         (If you omit the Group parameter, activity ");
           printMsg(0,'I',WTOPC_CHAIN,"           for all groups on this processor will be  ");
           printMsg(0,'I',WTOPC_CHAIN,"           stopped) ");
           printMsg(0,'I',WTOPC_CHAIN," ");
           printMsg(0,'I',WTOPC_CHAIN," ");
           printMsg(0,'I',WTOPC_CHAIN,"  NOTE: ZTEST FIXED FILE STOP only stops activity");
           printMsg(0,'I',WTOPC_CHAIN,"         for either all groups or the specified  ");
           printMsg(0,'I',WTOPC_CHAIN,"         group on the processor where it was entered");
           printMsg(0,'I',WTOPC_CHAIN," ");
           printMsg(0,'I',WTOPC_CHAIN_END,"END OF ZTEST FFACt SYNTAX DISPLAY FOR FIXED STOP");

           exit(0);
         } // if last parm after "FIXED" is "STOP"

       } // end If next parm is "STOP"

       if (parse && (strncmp(parse->IPRSE_parameter, "CLEAR", 5) == 0)) { // If next parm is "CLEAR"
         if (num_parms == 0) { // if last parm after "FIXED" is "CLEAR"

           printMsg(0,'I',WTOPC_CHAIN,"FIND FILE DRIVER SYNTAX FOR FIXED CLEAR");
           printMsg(0,'I',WTOPC_CHAIN," ");
           printMsg(0,'I',WTOPC_CHAIN,"  ZTEST FFACt Fixed CLEAR Group-n");
           printMsg(0,'I',WTOPC_CHAIN," ");
           printMsg(0,'I',WTOPC_CHAIN,"        where:");
           printMsg(0,'I',WTOPC_CHAIN," ");
           printMsg(0,'I',WTOPC_CHAIN,"         n = an integer between 1 and 99999 which is the");
           printMsg(0,'I',WTOPC_CHAIN,"              FIXED group number whose counts should");
           printMsg(0,'I',WTOPC_CHAIN,"              be CLEARed");
           printMsg(0,'I',WTOPC_CHAIN," ");
           printMsg(0,'I',WTOPC_CHAIN,"  NOTE: (1) To CLEAR counts for a group on a given");
           printMsg(0,'I',WTOPC_CHAIN,"             processor, it must be STOPPED first");
           printMsg(0,'I',WTOPC_CHAIN,"        (2) If you want to CLEAR counts for this group");
           printMsg(0,'I',WTOPC_CHAIN,"             on other processors, you must enter a ");
           printMsg(0,'I',WTOPC_CHAIN,"             ZTEST FFACt Fixed CLEAR command on each");
           printMsg(0,'I',WTOPC_CHAIN,"             processor where you want to CLEAR counts");
           printMsg(0,'I',WTOPC_CHAIN,"             for that group");
           printMsg(0,'I',WTOPC_CHAIN," ");
           printMsg(0,'I',WTOPC_CHAIN_END,"END OF ZTEST FFACt SYNTAX DISPLAY FOR FIXED CLEAR");

           exit(0);
         } // if last parm after "FIXED" is "CLEAR"

       } // end If next parm is "CLEAR"

       if (parse && (strncmp(parse->IPRSE_parameter, "KILL", 4) == 0)) { // If next parm is "KILL"
         if (num_parms == 0) { // if last parm after "FIXED" is "KILL"

           printMsg(0,'I',WTOPC_CHAIN,"FIND FILE DRIVER SYNTAX FOR FIXED KILL");
           printMsg(0,'I',WTOPC_CHAIN," ");
           printMsg(0,'I',WTOPC_CHAIN,"  ZTEST FFACt Fixed KILL Group-n");
           printMsg(0,'I',WTOPC_CHAIN," ");
           printMsg(0,'I',WTOPC_CHAIN,"        where:");
           printMsg(0,'I',WTOPC_CHAIN," ");
           printMsg(0,'I',WTOPC_CHAIN,"         n = an integer between 1 and 99999 which is the");
           printMsg(0,'I',WTOPC_CHAIN,"              FIXED group number to KILL");
           printMsg(0,'I',WTOPC_CHAIN," ");
           printMsg(0,'I',WTOPC_CHAIN,"  NOTE: (1) To KILL a group on a given processor, it must be");
           printMsg(0,'I',WTOPC_CHAIN,"             STOPPED first");
           printMsg(0,'I',WTOPC_CHAIN,"        (2) You must enter a KILL command on each active");
           printMsg(0,'I',WTOPC_CHAIN,"             processor in the complex to ensure that");
           printMsg(0,'I',WTOPC_CHAIN,"             the group is truly eliminated");
           printMsg(0,'I',WTOPC_CHAIN,"        (3) For FIXED, any processor added to the complex");
           printMsg(0,'I',WTOPC_CHAIN,"             will not see any group that is KILLED");
           printMsg(0,'I',WTOPC_CHAIN,"             because its associated file is removed");
           printMsg(0,'I',WTOPC_CHAIN,"             from the file system");
           printMsg(0,'I',WTOPC_CHAIN," ");
           printMsg(0,'I',WTOPC_CHAIN_END,"END OF ZTEST FFACt SYNTAX DISPLAY FOR FIXED KILL");

           exit(0);
         } // if last parm after "FIXED" is "KILL"

       } // end If next parm is "KILL"

    } // end if additional parameter on HELP for "FIXED"

    if (is_for_pool) { // if additional parameter on HELP for "POOL"

       if (parse && (strncmp(parse->IPRSE_parameter, "BUild", 5) == 0)) { // If next parm is "BUILD"
         if (num_parms == 0) { // if last parm after "POOL" is "BUILD"

           printMsg(0,'I',WTOPC_CHAIN,"FIND FILE DRIVER SYNTAX FOR POOL BUILD");
           printMsg(0,'I',WTOPC_CHAIN," ");
           printMsg(0,'I',WTOPC_CHAIN,"  ZTEST FFACt Pool BUild Group-n Num-number");
           printMsg(0,'I',WTOPC_CHAIN,"           STPool-section RID-rid Expire-time");
           printMsg(0,'I',WTOPC_CHAIN," ");
           printMsg(0,'I',WTOPC_CHAIN,"        where:");
           printMsg(0,'I',WTOPC_CHAIN," ");
           printMsg(0,'I',WTOPC_CHAIN,"         n = an integer between 1 and 99999 which is");
           printMsg(0,'I',WTOPC_CHAIN,"              the number to assign to this POOL group");
           printMsg(0,'I',WTOPC_CHAIN," ");
	   printMsg(0,'I',WTOPC_CHAIN,"         number = an integer greater than 0 but");
	   printMsg(0,'I',WTOPC_CHAIN,"          less than or equal to 999,999 to indicate");
	   printMsg(0,'I',WTOPC_CHAIN,"          the number to pool records to get for this");
	   printMsg(0,'I',WTOPC_CHAIN,"          POOL group");
           printMsg(0,'I',WTOPC_CHAIN," ");
	   printMsg(0,'I',WTOPC_CHAIN,"         section = specifies the short-term pool section");
	   printMsg(0,'I',WTOPC_CHAIN,"              from which records for this POOL group");
	   printMsg(0,'I',WTOPC_CHAIN,"              will be gotten, including the device type.");
	   printMsg(0,'I',WTOPC_CHAIN,"              Examples: SSTA, SSTB, SSTC, SSTD,");
	   printMsg(0,'I',WTOPC_CHAIN,"                        LSTA, LSTB, LSTC, LSTD,");
	   printMsg(0,'I',WTOPC_CHAIN,"                        4STA, 4STB, 4STC, 4STD,");
	   printMsg(0,'I',WTOPC_CHAIN,"                        4S6A, 4S6B, 4S6C, 4S6D");
           printMsg(0,'I',WTOPC_CHAIN," ");
	   printMsg(0,'I',WTOPC_CHAIN,"         rid = 4 hex digits = the 2-byte record ID to be");
	   printMsg(0,'I',WTOPC_CHAIN,"              used to initialize the records");
           printMsg(0,'I',WTOPC_CHAIN," ");
	   printMsg(0,'I',WTOPC_CHAIN,"         time = number of minutes greater than or equal");
	   printMsg(0,'I',WTOPC_CHAIN,"              to zero and less than or equal to 1440");
	   printMsg(0,'I',WTOPC_CHAIN,"              after which time, pools that were added");
	   printMsg(0,'I',WTOPC_CHAIN,"              to the group will age out, and new pools");
	   printMsg(0,'I',WTOPC_CHAIN,"              will be gotten in their place.");
	   printMsg(0,'I',WTOPC_CHAIN,"         NOTE: If you specify a time of 0, the pool ");
	   printMsg(0,'I',WTOPC_CHAIN,"               records belonging to this group will ");
	   printMsg(0,'I',WTOPC_CHAIN,"               NEVER age out.");
	   printMsg(0,'I',WTOPC_CHAIN," ");
	   printMsg(0,'I',WTOPC_CHAIN_END,"END OF ZTEST FFACt SYNTAX DISPLAY FOR POOL BUILD");

	   exit(0);
	 } // if last parm after "POOL" is "BUILD"

       } // end If next parm is "BUILD"

	       if (parse && (strncmp(parse->IPRSE_parameter, "STARt", 5) == 0)) { // If next parm is "START"
		 if (num_parms == 0) { // if last parm after "POOL" is "START"

		   printMsg(0,'I',WTOPC_CHAIN,"FIND FILE DRIVER SYNTAX FOR POOL START");
		   printMsg(0,'I',WTOPC_CHAIN," ");
		   printMsg(0,'I',WTOPC_CHAIN,"  ZTEST FFACt Pool STARt Group-n SEED-s Read-r Write-w");
		   printMsg(0,'I',WTOPC_CHAIN,"            ECBs-x {Lock-Y|N}");
           printMsg(0,'I',WTOPC_CHAIN,"            {LOw-Y|I} {SWitch-Y|I}");
           printMsg(0,'I',WTOPC_CHAIN,"            {UTILclass-uClass_name} ");
           printMsg(0,'I',WTOPC_CHAIN,"            {RESourceclass-Y|N} ");
		   printMsg(0,'I',WTOPC_CHAIN," ");
		   printMsg(0,'I',WTOPC_CHAIN,"        where:");
		   printMsg(0,'I',WTOPC_CHAIN," ");
		   printMsg(0,'I',WTOPC_CHAIN,"         n = an integer between 1 and 99999 which is the");
		   printMsg(0,'I',WTOPC_CHAIN,"              POOL group number to start");
		   printMsg(0,'I',WTOPC_CHAIN,"         s = decimal numeric seed from 0 to 999 used to");
		   printMsg(0,'I',WTOPC_CHAIN,"              randomly access ordinals");
		   printMsg(0,'I',WTOPC_CHAIN,"         r = decimal number of reads per ECB, from 0 to 999");
		   printMsg(0,'I',WTOPC_CHAIN,"         w = decimal number of writes per ECB, from 0 to 999");
		   printMsg(0,'I',WTOPC_CHAIN,"         x = decimal number of ECBs to spawn for each");
		   printMsg(0,'I',WTOPC_CHAIN,"              iteration, from 1 to 999");
		   printMsg(0,'I',WTOPC_CHAIN,"         Lock-Y (default) lock each record prior to writing");
		   printMsg(0,'I',WTOPC_CHAIN,"              to it");
		   printMsg(0,'I',WTOPC_CHAIN,"         Lock-N to bypass locking");
           printMsg(0,'I',WTOPC_CHAIN,"         LOw-Y sets the LOWPRIORITY attribute for the parent");    //d26696
           printMsg(0,'I',WTOPC_CHAIN,"              ECB ");
           printMsg(0,'I',WTOPC_CHAIN,"         LOw-I sets the LOWPRIORITY attribute for the parent");    //d26696
           printMsg(0,'I',WTOPC_CHAIN,"              ECB and its child ECBs will INHERIT it");
           printMsg(0,'I',WTOPC_CHAIN,"         SWitch-Y sets the SWITCHABLE attribute for the");     
           printMsg(0,'I',WTOPC_CHAIN,"              parent ECB ");                                       //d26696
           printMsg(0,'I',WTOPC_CHAIN,"         SWitch-I sets the SWITCHABLE attribute for the");
           printMsg(0,'I',WTOPC_CHAIN,"              parent ECB and its child ECBs will");                //d26696
           printMsg(0,'I',WTOPC_CHAIN,"              INHERIT it");
           printMsg(0,'I',WTOPC_CHAIN,"         UTILclass-uClass_name causes this instance of the");
           printMsg(0,'I',WTOPC_CHAIN,"              drvier to use the specified utilization class");
           printMsg(0,'I',WTOPC_CHAIN,"         RESourceclass-Y this instance will use a LODIC ");
           printMsg(0,'I',WTOPC_CHAIN,"              resource class");
           printMsg(0,'I',WTOPC_CHAIN,"         RESourceclass-N this instance will NOT use a LODIC ");
           printMsg(0,'I',WTOPC_CHAIN,"              resource class -- THIS IS THE DEFAULT");
		   printMsg(0,'I',WTOPC_CHAIN," ");
		   printMsg(0,'I',WTOPC_CHAIN,"  NOTE: If Lock-Y the number of Reads must be greater than");
		   printMsg(0,'I',WTOPC_CHAIN,"         or equal to the number of Writes");
		   printMsg(0,'I',WTOPC_CHAIN," ");
		   printMsg(0,'I',WTOPC_CHAIN_END,"END OF ZTEST FFACt SYNTAX DISPLAY FOR POOL START");

		   exit(0);
		 } // if last parm after "POOL" is "START"

	       } // end If next parm is "START"

	       if (parse && (strncmp(parse->IPRSE_parameter, "STATus", 6) == 0)) { // If next parm is "STATUS"
		 if (num_parms == 0) { // if last parm after "POOL" is "STATUS"

		   printMsg(0,'I',WTOPC_CHAIN,"FIND FILE DRIVER SYNTAX FOR POOL STATUS");
		   printMsg(0,'I',WTOPC_CHAIN," ");
		   printMsg(0,'I',WTOPC_CHAIN,"  ZTEST FFACt Pool STATus {Group-n}");
		   printMsg(0,'I',WTOPC_CHAIN," ");
		   printMsg(0,'I',WTOPC_CHAIN,"        where:");
		   printMsg(0,'I',WTOPC_CHAIN," ");
		   printMsg(0,'I',WTOPC_CHAIN,"         (OPTIONAL) n = an integer between 1 and 99999 which");
		   printMsg(0,'I',WTOPC_CHAIN,"                         is the POOL group number to");
		   printMsg(0,'I',WTOPC_CHAIN,"                         display status for");
		   printMsg(0,'I',WTOPC_CHAIN," ");
		   printMsg(0,'I',WTOPC_CHAIN,"  NOTE: If the Group is omitted, a status summary is");
		   printMsg(0,'I',WTOPC_CHAIN,"         displayed for all pool groups in memory for");
		   printMsg(0,'I',WTOPC_CHAIN,"         the current processor");
		   printMsg(0,'I',WTOPC_CHAIN," ");
		   printMsg(0,'I',WTOPC_CHAIN_END,"END OF ZTEST FFACt SYNTAX DISPLAY FOR POOL STATUS");

		   exit(0);
		 } // if last parm after "POOL" is "STATUS"

	       } // end If next parm is "STATUS"

	       if (parse && (strncmp(parse->IPRSE_parameter, "LIST", 4) == 0)) { // If next parm is "LIST"
		 if (num_parms == 0) { // if last parm after "POOL" is "LIST"

		   printMsg(0,'I',WTOPC_CHAIN,"FIND FILE DRIVER SYNTAX FOR POOL LIST");
		   printMsg(0,'I',WTOPC_CHAIN," ");
		   printMsg(0,'I',WTOPC_CHAIN,"  ZTEST FFACt Pool LIST Group-n NUM-number");
		   printMsg(0,'I',WTOPC_CHAIN," ");
		   printMsg(0,'I',WTOPC_CHAIN,"        where:");
		   printMsg(0,'I',WTOPC_CHAIN," ");
		   printMsg(0,'I',WTOPC_CHAIN,"         n = an integer between 1 and 99999 which");
		   printMsg(0,'I',WTOPC_CHAIN,"                is the POOL group number to");
		   printMsg(0,'I',WTOPC_CHAIN,"                list file addresses for");
		   printMsg(0,'I',WTOPC_CHAIN," ");
	           printMsg(0,'I',WTOPC_CHAIN,"         number = an integer greater than 0 but");
	           printMsg(0,'I',WTOPC_CHAIN,"          less than or equal to 999,999 to indicate");
	           printMsg(0,'I',WTOPC_CHAIN,"          the number of file addresses for this POOL");
	           printMsg(0,'I',WTOPC_CHAIN,"          group that you would like to list");
		   printMsg(0,'I',WTOPC_CHAIN," ");
		   printMsg(0,'I',WTOPC_CHAIN_END,"END OF ZTEST FFACt SYNTAX DISPLAY FOR POOL LIST");

		   exit(0);
		 } // if last parm after "POOL" is "LIST"

	       } // end If next parm is "LIST"

	       if (parse && (strncmp(parse->IPRSE_parameter, "STOP", 4) == 0)) { // If next parm is "STOP"
		 if (num_parms == 0) { // if last parm after "POOL" is "STOP"

		   printMsg(0,'I',WTOPC_CHAIN,"FIND FILE DRIVER SYNTAX FOR POOL STOP");
		   printMsg(0,'I',WTOPC_CHAIN," ");
		   printMsg(0,'I',WTOPC_CHAIN,"  ZTEST FFACt Pool STOP {Group-n}");
		   printMsg(0,'I',WTOPC_CHAIN," ");
		   printMsg(0,'I',WTOPC_CHAIN,"        where:");
		   printMsg(0,'I',WTOPC_CHAIN," ");
		   printMsg(0,'I',WTOPC_CHAIN,"         n = an integer between 1 and 99999 which is the");
		   printMsg(0,'I',WTOPC_CHAIN,"              POOL group number to stop");
		   printMsg(0,'I',WTOPC_CHAIN," ");
		   printMsg(0,'I',WTOPC_CHAIN,"         (If you omit the Group parameter, activity ");
		   printMsg(0,'I',WTOPC_CHAIN,"           for all pool groups on this processor will be");
		   printMsg(0,'I',WTOPC_CHAIN,"           stopped) ");
		   printMsg(0,'I',WTOPC_CHAIN," ");
		   printMsg(0,'I',WTOPC_CHAIN," ");
		   printMsg(0,'I',WTOPC_CHAIN,"  NOTE: ZTEST POOL POOL STOP only stops activity");
		   printMsg(0,'I',WTOPC_CHAIN,"         for either all pool groups or the specified  ");
		   printMsg(0,'I',WTOPC_CHAIN,"         group on the processor where it was entered");
		   printMsg(0,'I',WTOPC_CHAIN," ");
		   printMsg(0,'I',WTOPC_CHAIN_END,"END OF ZTEST FFACt SYNTAX DISPLAY FOR POOL STOP");

		   exit(0);
		 } // if last parm after "POOL" is "STOP"

	       } // end If next parm is "STOP"

	       if (parse && (strncmp(parse->IPRSE_parameter, "CLEAR", 5) == 0)) { // If next parm is "CLEAR"
		 if (num_parms == 0) { // if last parm after "POOL" is "CLEAR"

		   printMsg(0,'I',WTOPC_CHAIN,"FIND FILE DRIVER SYNTAX FOR POOL CLEAR");
		   printMsg(0,'I',WTOPC_CHAIN," ");
		   printMsg(0,'I',WTOPC_CHAIN,"  ZTEST FFACt Pool CLEAR Group-n");
		   printMsg(0,'I',WTOPC_CHAIN," ");
		   printMsg(0,'I',WTOPC_CHAIN,"        where:");
		   printMsg(0,'I',WTOPC_CHAIN," ");
		   printMsg(0,'I',WTOPC_CHAIN,"         n = an integer between 1 and 99999 which is the");
		   printMsg(0,'I',WTOPC_CHAIN,"              POOL group number whose counts should");
		   printMsg(0,'I',WTOPC_CHAIN,"              be CLEARed");
		   printMsg(0,'I',WTOPC_CHAIN," ");
		   printMsg(0,'I',WTOPC_CHAIN,"  NOTE: (1) To CLEAR counts for a group on a given");
		   printMsg(0,'I',WTOPC_CHAIN,"             processor, it must be STOPPED first");
		   printMsg(0,'I',WTOPC_CHAIN,"        (2) If you want to CLEAR counts for this group");
		   printMsg(0,'I',WTOPC_CHAIN,"             on other processors, you must enter a ");
		   printMsg(0,'I',WTOPC_CHAIN,"             ZTEST FFACt Pool CLEAR command on each");
		   printMsg(0,'I',WTOPC_CHAIN,"             processor where you want to CLEAR counts");
		   printMsg(0,'I',WTOPC_CHAIN,"             for that group");
		   printMsg(0,'I',WTOPC_CHAIN," ");
		   printMsg(0,'I',WTOPC_CHAIN_END,"END OF ZTEST FFACt SYNTAX DISPLAY FOR POOL CLEAR");

		   exit(0);
		 } // if last parm after "POOL" is "CLEAR"

	       } // end If next parm is "CLEAR"

	       if (parse && (strncmp(parse->IPRSE_parameter, "KILL", 4) == 0)) { // If next parm is "KILL"
		 if (num_parms == 0) { // if last parm after "POOL" is "KILL"

		   printMsg(0,'I',WTOPC_CHAIN,"FIND FILE DRIVER SYNTAX FOR POOL KILL");
		   printMsg(0,'I',WTOPC_CHAIN," ");
		   printMsg(0,'I',WTOPC_CHAIN,"  ZTEST FFACt Pool KILL Group-n");
		   printMsg(0,'I',WTOPC_CHAIN," ");
		   printMsg(0,'I',WTOPC_CHAIN,"        where:");
		   printMsg(0,'I',WTOPC_CHAIN," ");
		   printMsg(0,'I',WTOPC_CHAIN,"         n = an integer between 1 and 99999 which is the");
		   printMsg(0,'I',WTOPC_CHAIN,"              POOL group number to KILL");
		   printMsg(0,'I',WTOPC_CHAIN," ");
		   printMsg(0,'I',WTOPC_CHAIN,"  NOTE: (1) To KILL a group on a given processor, it must be");
		   printMsg(0,'I',WTOPC_CHAIN,"             STOPPED first");
		   printMsg(0,'I',WTOPC_CHAIN,"        (2) You must enter a KILL command on each active");
		   printMsg(0,'I',WTOPC_CHAIN,"             processor in the complex to ensure that");
		   printMsg(0,'I',WTOPC_CHAIN,"             the group is truly eliminated");
		   printMsg(0,'I',WTOPC_CHAIN," ");
		   printMsg(0,'I',WTOPC_CHAIN_END,"END OF ZTEST FFACt SYNTAX DISPLAY FOR POOL KILL");

		   exit(0);
		 } // if last parm after "POOL" is "KILL"

	       } // end If next parm is "KILL"

	    } // end if additional parameter on HELP for "POOL"

	    printMsg(79,'E',WTOPC_NO_CHAIN,"ZTEST FFACt HELP command rejected: un-recognized parameter");

	} // end of function qfnl_help


	boolean qfnl_check_if_groupAlreadyBuilt(char * file_name,
						qfnl_fixed_build_parm_list_Ptr fixed_build_ptr,
						boolean * is_error_ptr)
	{ // begin qfnl_check_if_groupAlreadyBuilt

	//  The purpose of this function is to determine if there is a file name for
	//    a given fixed file group already in the file system.  If this is the case
	//    it will return a boolean TRUE and set the build parameter list structure
	//    pointed at by fixed_build_ptr to the values taken from the file.
	//  If a file for that fixed file group does not already exist, then boolean FALSE
	//    is returned.
	//  If any error is encountered with the file itself or which prevents us from reading
	//    the file, in addition to returning FALSE, the boolean pointed at by is_error_ptr
	//    will be set to TRUE.

	 FILE * group_file;                  // used to read the file
	 int group_file_length = 0;          // used to determine the length of the file
	 size_t read_length = 0;             // length returned from read
	 *is_error_ptr = FALSE;              // initially assume no error

	 group_file = fopen(file_name, "rb");

	 if (!group_file) { // if the file does not exist
	   return(FALSE);
	 }

	// if the file exists, we will do some minimal checking before we use it.
	//  For starters, make sure its length mathes the length of a qfnl_fixed_build_parm_list structure

	 fseek (group_file, 0, SEEK_END);
	 group_file_length = ftell (group_file);
	 fseek (group_file, 0, SEEK_SET);

	 if ( group_file_length != sizeof(qfnl_fixed_build_parm_list) ) {
	   printMsg(32,'E',WTOPC_NO_CHAIN,"FILE NAMED %s EXISTS BUT LENGTH IS INCORRECT", file_name);
	   *is_error_ptr = TRUE;
	   return(FALSE);
	 }

	// read and then close the file...

	 read_length = fread (fixed_build_ptr, 1, group_file_length, group_file);
	 fclose (group_file);

	// Check for any read error by checking the length read...If any error, bail out...

	 if ( read_length != (size_t) group_file_length ) {
	   printMsg(33,'E',WTOPC_NO_CHAIN,"FILE NAMED %s EXISTS BUT ERROR READING IT", file_name);
	   *is_error_ptr = TRUE;
	   return(FALSE);
	 }

	// Perform just one more bit of validation.  Check that the groupString matches the last part of the file name.
	//  For example, if the groupString is 99999, the file name is /usr/QFNL_FIXED_GROUP99999 (where "/usr" is QFNL_DIR
	//   and "QFNL_FIXED_GROUP" is QVFN_PFX), we want to compare the last 5 characters of the file name
	//  (99999) to verify that it matches the groupString...

	  if (memcmp(&file_name[strlen(QFNL_DIR)+strlen(QFNL_PFX)],
		     fixed_build_ptr->groupString,
		     strlen(fixed_build_ptr->groupString)) != 0 ){
	   printMsg(34,'E',WTOPC_NO_CHAIN,"FILE NAMED %s EXISTS BUT CONTENTS QUESTIONABLE", file_name);
	   *is_error_ptr = TRUE;
	   return(FALSE);
	  }

	// No error was encountered, so return TRUE that evidence that the group was built was found in the file system

	  return(TRUE);

	} // end qfnl_check_if_groupAlreadyBuilt

	int qfnl_FARF6_to_hex_string(char * FARFstring,
				     int input_size,
				     TPF_FA8 file_address)
	{

	  int myrc = 0;
	  int size = QFNL_FARF_STRING_SIZE;

	  // The String must be 17 bytes long to properly hold
	  //   the character hex representation of the FARF6 address
	  if (input_size != QFNL_FARF_STRING_SIZE)
	     return(myrc);
	  memset(FARFstring, 0xF0, QFNL_FARF_STRING_SIZE-1);
  FARFstring[QFNL_FARF_STRING_SIZE-1] = '\0';
  myrc = convertBinToHexStr(FARFstring,
                           &size,
                           (char *) &file_address,
                           sizeof(file_address));
  if (myrc)
     FARFstring[size] = '\0';
  return(myrc);
}

/*===================================================================*/
/* [Binary] To [Hexadecimal String] Conversion                       */
/*===================================================================*/
int convertBinToHexStr(char *dst, int *dstlen,
                        const char *src, int srclen)
{
  char lower;
  char tempValueUp, tempValueLow, asciiValUp, asciiValLow;
  int i;

  if (*dstlen <= 0)
  {
    return(0);
  }

  i = 0;
  lower = 15;  /* 00001111 */

  /* Automatically determine length of src*/
  if (srclen == -1)
  {
    srclen = strlen(src);
  }

  /* Set length of dst string */
  if (*dstlen > srclen * 2 + 1)
  {
    *dstlen = srclen * 2;
  }
  else /*becasue str is null terminated */
  {
    *dstlen = *dstlen - 1;
  }

  /* Initialize hex destination with zeroes                          */
  memset(dst, 0x00, *dstlen);

  while ( i < srclen && (2*i)< *dstlen )
  {
    /* Value of upper byte */
    tempValueUp = src[i] >> 4;
    /* Value of lower byte */
    tempValueLow = src[i] & lower;

    /* Convert to ascii */
    if (tempValueUp < 10)
    {
      asciiValUp = tempValueUp + '0';
    }
    else
    {
      asciiValUp = tempValueUp - 10 + 'A';
    }

    if (tempValueLow < 10)
    {
      asciiValLow = tempValueLow + '0';
    }
    else
    {
      asciiValLow = tempValueLow - 10 + 'A';
    }
    dst[2*i] = asciiValUp;
    if (2*i < *dstlen)
    {
      dst[2*i + 1] = asciiValLow;
    }

    i++;
  }

  /* Null terminate the string */
  dst[2*i] = 0;

  return(1);
}

void qfnl_pool_child_ecb (qfnl_child_ecb_parm_list_Ptr incoming_ecb_parm_ptr)
{

 // This function is called by the child ECB to perform FIND and FILE activity on the
 //   group as requested by the parent ECB

 // The child ECB will do the number of reads and writes requested, and then exit.
 //  It is up to the parent to wait and then kick off another batch of children ECBs.
 //
 // NOTE: Upon entry, each child ECB will update a running counter of the number of ECBs that have
 //         been created for the grup.  Upon exit, it will update a running counter of the number of ECBs that have
 //         exited for the group.

 struct swisc_immediate_input swisc_is;
 swisc_is.istream = SWISC_IS_BALANCE;
 swisc_immediate(&swisc_is);

 if (debug) printf("\n !!!!! The child ECB number %d will operate on group %d !!!!\n",
                     incoming_ecb_parm_ptr->ecbNum, incoming_ecb_parm_ptr->group);

 unsigned int random_index;  // will be used to hold random indexes that will be accessed for FIND/FILE activity


 // Feed the random number generator

 srand(incoming_ecb_parm_ptr->child_seed);

    // srand(time(NULL));

 // Locate the POOL file group in memory.  The parent (ZTEST FFFACt POOL START ecb) should have
 //   either found it there already or put it there from file.  But if the child does not find the
 //   group in memory any more, it will exit.

 qfnl_pool_info_Ptr my_group = qfnl_locate_pool_group (incoming_ecb_parm_ptr->group, FALSE);


 if (my_group != NULL ) { // if group found

  tpf_stpoc(TPF_STPOC_ON, stpoc_value);  // allow ourselves to modify key 9 storage gotten by tpf_gsysc
  tpf_serialized_update8(1, (tsug_t *)&my_group->group_stats.ecbCreateCount); // increment total count ECBs created
  tpf_serialized_update8(1, (tsug_t *)&my_group->group_stats.ecbCount);       // increment count ECBs active
  tpf_stpoc(TPF_STPOC_OFF, stpoc_value);  // done for now

  // get number of reads and writes for this child ECB for use in loop

  int numberOfReads = incoming_ecb_parm_ptr->numberOfReads;
  int numberOfWrites = incoming_ecb_parm_ptr->numberOfWrites;
  #define QFNL_LOCKING incoming_ecb_parm_ptr->doLocking

  // the following are used in case there are any error messages or dumps

  char errMsg[100] = "";
  boolean anyError = false;  // assume no errors for now


  // We need the following in order to do i/o:

  TPF_DECB *decb = NULL;
  DECBC_RC decbRC;
  TPF_FA8 fileaddr;
  void * record_ptr = NULL;  // will eventually point to record read or initialized
  enum t_find_decb findtype; // used for find
  enum t_act filetype;       // used for file

  char FARFstring[QFNL_FARF_STRING_SIZE];


  // allocate a DECB

  decb = tpf_decb_create(NULL, &decbRC);

  if (decb == NULL) { // if could not allocate a DECB
    anyError = true;
    sprintf(errMsg,"ALLOCATE DECB FAILED FOR CHILD ECB FOR POOL GROUP %d", my_group->group_number);
    serrc_op(SERRC_ALL_HEAP_RETURN,0x5555,errMsg, NULL);

    // increment error counter(s)
    tpf_stpoc(TPF_STPOC_ON, stpoc_value);   // allow ourselves to modify key 9 storage gotten by tpf_gsysc
    tpf_serialized_update8(1, (tsug_t *)&my_group->group_stats.errCount);
    tpf_serialized_update8(1, (tsug_t *)&my_group->group_stats.envErr);
    tpf_stpoc(TPF_STPOC_OFF, stpoc_value);  // done with modify for now

  }

  //
  // Main loop:
  //  The child ECB will loop and perform just the number of reads and writes requested.
  //  But if any error occurs, it will output an error message and exit.
  //

  boolean toggleNOREL = false; // used to toggle between FILUC and FILNC/UNFRC
  boolean held = false;  // will set to true if the record ends up getting held
  boolean readDone = false; // will set to true when the record has been read
                            //  (If not read then we got a core block and there is no minimal verify to do)

  while ( ((numberOfReads + numberOfWrites) > 0) && (!anyError) && (my_group->group_flag == QFNL_GROUP_STAT_STARTED) )

  { // While there are still reads or writes to do

     // generate a random index to work on

     random_index = qfnl_random(1, my_group->number_records);

     if (debug) printMsg(9,'I',WTOPC_NO_CHAIN,"!!!!! child ECB %d decimal random index is %d !!!!", incoming_ecb_parm_ptr->ecbNum, random_index);

     // find its file address

     if ( (anyError=qfnl_get_pool_fileaddr(my_group,
                                            random_index,
                                            decb,
                                            &fileaddr,
                                            FARFstring,
                                            QFNL_FARF_STRING_SIZE,
                                            true)) ) {

        // increment error counter(s)
        tpf_stpoc(TPF_STPOC_ON, stpoc_value);   // allow ourselves to modify key 9 storage gotten by tpf_gsysc
        tpf_serialized_update8(1, (tsug_t *)&my_group->group_stats.errCount);
        tpf_serialized_update8(1, (tsug_t *)&my_group->group_stats.envErr);
        tpf_stpoc(TPF_STPOC_OFF, stpoc_value);  // done with modify for now

        // bail out
        break;

     }

     //
     // If any reads, do a read, but decide whether or not you will do a lock:
     //    If we are not doing locking (LOCK-N was specified on the START)
     //      then don't hold.
     //    If we are doing locking, but there are no more writes to do
     //      then don't hold
     //    Otherwise, (we are locking, and there is still a write to do)
     //       do a hold
     //

     if (numberOfReads >= 1) { // if there is at least one read

       // If we are locking and there are still writes to do, then we must
       //   do a find with hold;
       // else just do a find without hold

       if ( (QFNL_LOCKING) && (numberOfWrites) ) {

           held = TRUE;
           findtype = HOLD_WAIT;

       } else {

           held = FALSE;
           findtype = NOHOLD_WAIT;

       }

       // read the record

       record_ptr = find_record_ext(decb,                               // decb
                                    &fileaddr,                          // file address of record to read
                                    incoming_ecb_parm_ptr->rec_ID,      // use record ID that was read from first record
                                    '\0',                               // bypass RCC
                                    findtype,                           // type of find controls hold or not
                                    FIND_DEFEXT);                       // use default non GDS

       if (record_ptr == NULL) { // if failure finding random ordinal

          anyError = true;
          sprintf(errMsg, "ERROR READING POOL GROUP %d INDEX %d, FA=%s SUD = %2.2X",
                  my_group->group_number, random_index, FARFstring, decb->idecsud);
          serrc_op(SERRC_ALL_HEAP_RETURN,0x5555,errMsg, NULL);

          // increment error counter(s)
          tpf_stpoc(TPF_STPOC_ON, stpoc_value);   // allow ourselves to modify key 9 storage gotten by tpf_gsysc
          tpf_serialized_update8(1, (tsug_t *)&my_group->group_stats.errCount);
          tpf_serialized_update8(1, (tsug_t *)&my_group->group_stats.ioErr);
          tpf_stpoc(TPF_STPOC_OFF, stpoc_value);  // done with modify for now

          // For certain errors, the CP will automatically unhold the lock so the application should not attempt to
          //   unhold as well, to avoid extraneous CTL-24 dumps...

          if ( (decb->idecsud & CXSGHE) ||
               (decb->idecsud & CXSGAE) )
              held = false;   // no longer being held by my application, as the CP did the unhold for me

          break;  // get out of loop so we can bail out

       // end if failure finding random ordinal
       } else { // read worked


          readDone = TRUE;

          numberOfReads--;      // decrement the number of reads left

          // Also increment the running count of reads done

          tpf_stpoc(TPF_STPOC_ON, stpoc_value);   // allow ourselves to modify key 9 storage gotten by tpf_gsysc
          tpf_serialized_update8(1, (tsug_t *)&my_group->group_stats.readCount);
          tpf_stpoc(TPF_STPOC_OFF, stpoc_value);  // done with modify for now

          // Increment running count of holds when appropriate

          if (held) {
            tpf_stpoc(TPF_STPOC_ON, stpoc_value);   // allow ourselves to modify key 9 storage gotten by tpf_gsysc
            tpf_serialized_update8(1, (tsug_t *)&my_group->group_stats.holdCount);
            tpf_stpoc(TPF_STPOC_OFF, stpoc_value);  // done with modify for now
          }

       } // end else read worked


     // end if there is at least one read

     } else if (numberOfWrites >= 1) { // if there are no reads left but at least one write,
                                       //    we must get and initialize a core block to write
                                       //    to the record

       record_ptr = getcc(decb,
                          (enum t_getfmt) (GETCC_SIZE+GETCC_FILL),
                          my_group->record_size,
                          0x00);

       if (record_ptr == NULL) {  // if error getting storage

          sprintf(errMsg, "ERROR GETTING STORAGE TO FILE POOL GROUP %d INDEX %d, FA=%s",
                  my_group->group_number, random_index, FARFstring);
          serrc_op(SERRC_ALL_HEAP_RETURN,0x5555,errMsg, NULL);
          anyError = true;

          // increment error counter(s)
          tpf_stpoc(TPF_STPOC_ON, stpoc_value);   // allow ourselves to modify key 9 storage gotten by tpf_gsysc
          tpf_serialized_update8(1, (tsug_t *)&my_group->group_stats.errCount);
          tpf_serialized_update8(1, (tsug_t *)&my_group->group_stats.mallocErr);
          tpf_stpoc(TPF_STPOC_OFF, stpoc_value);  // done with modify for now

          break;

       }

       readDone = FALSE;

     } // end if there are no reads left but at least one write


     if (readDone) { // if the record was read, perform minimial verification of its contents

       if ( !qfnl_is_recordOK(record_ptr, fileaddr, my_group) ) { // if record read is NOT OK

          sprintf(errMsg, "QUESTIONABLE RECORD CONTENTS POOL GROUP %d INDEX %d, FA=%s",
                  my_group->group_number, random_index, FARFstring);
          serrc_op(SERRC_ALL_HEAP_RETURN,0x5555,errMsg, NULL);
          anyError = true;

          // increment error counter(s)
          tpf_stpoc(TPF_STPOC_ON, stpoc_value);   // allow ourselves to modify key 9 storage gotten by tpf_gsysc
          tpf_serialized_update8(1, (tsug_t *)&my_group->group_stats.errCount);
          tpf_serialized_update8(1, (tsug_t *)&my_group->group_stats.contentErr);
          tpf_stpoc(TPF_STPOC_OFF, stpoc_value);  // done with modify for now

          break;

       } // if record read is NOT OK

     } // end if the record was read


     if (numberOfWrites >= 1) { // if there is at least one write left

        tpf_TOD_ext_type tod;  // used to store TOD which will be copied to the pool item to remember when its record was last filed

        // Before we can write the record out to the database, we must modify its contents

        if (qfnl_modify_pool_record(record_ptr,
                                    fileaddr,
                                    my_group,
                                    &tod) == 0) {  // if modify fails quit now
          sprintf(errMsg, "ERROR MODIFYING POOL GROUP %d INDEX %d, FA=%s",
                  my_group->group_number, random_index, FARFstring);
          serrc_op(SERRC_ALL_HEAP_RETURN,0x5555,errMsg, NULL);
          anyError = true;

          // increment error counter(s)
          tpf_stpoc(TPF_STPOC_ON, stpoc_value);   // allow ourselves to modify key 9 storage gotten by tpf_gsysc
          tpf_serialized_update8(1, (tsug_t *)&my_group->group_stats.errCount);
          tpf_serialized_update8(1, (tsug_t *)&my_group->group_stats.envErr);
          tpf_stpoc(TPF_STPOC_OFF, stpoc_value);  // done with modify for now

          break;
        }

        if (held) {

        // if the record was held, we have to unhold it, but toggle how this is done

           if (toggleNOREL) {

             filetype = NOREL;     // We will file with NOREL for this iteration
             toggleNOREL = FALSE;  // Next time we will use UNHOLD

           } else {

             filetype = UNHOLD;    // We will file with UNNHOLD for this iteration
             toggleNOREL = TRUE;   // Next time we will use NOREL

           }

        } else {  // else the record was not held, so we do not unhold it

           filetype = NOREL;       // we will file with NOREL
        }

        // file the record with or without unhold as was decided

        file_record_ext(decb,                               // decb
                        &fileaddr,                          // file address of record to read
                        incoming_ecb_parm_ptr->rec_ID,      // use record ID that was read from first record
                        '\0',                               // bypass RCC
                        filetype,                           // type of file controls unhold or not
                        FILE_DEFEXT);                       // use default non GDS

        // For file with NOREL, we must wait for completion and decide whether or not to unhold
        //   the record

        if (filetype == NOREL) {

          if ( waitc() ) {
            anyError = true;
            sprintf(errMsg, "ERROR FILING POOL GROUP %d INDEX %d, FA=%s SUD = %2.2X",
                    my_group->group_number, random_index, FARFstring, decb->idecsud);
            serrc_op(SERRC_ALL_HEAP_RETURN,0x5555,errMsg, NULL);

          // increment error counter(s)
            tpf_stpoc(TPF_STPOC_ON, stpoc_value);   // allow ourselves to modify key 9 storage gotten by tpf_gsysc
            tpf_serialized_update8(1, (tsug_t *)&my_group->group_stats.errCount);
            tpf_serialized_update8(1, (tsug_t *)&my_group->group_stats.ioErr);
            tpf_stpoc(TPF_STPOC_OFF, stpoc_value);  // done with modify for now

            break;

          }

          if ( held )   // if it was read with hold need to undhold it
            unfrc_ext(decb,FIND_DEFEXT);

        }

        // Now that the record was written to, save the timestamp that the record was last updated in
        //     the slot for this record in the table of pools

        qfnl_pool_batch_Ptr pool_table = my_group->pool_batch_ptr;  // point at table of pools for this group
        int zero_based = random_index--;  // make index zero-based
        char * index_item = &(pool_table->first_item);  // point at first item in table
        index_item = index_item + zero_based * sizeof(qfnl_pool_item); // point at item that was filed
        qfnl_pool_item_Ptr pool_item = (qfnl_pool_item_Ptr) index_item;  // point to it as a structure

        tpf_stpoc(TPF_STPOC_ON, stpoc_value);   // allow ourselves to modify key 9 storage gotten by tpf_gsysc
        memcpy(&pool_item->tod_updated, &tod, sizeof(tod));
        tpf_stpoc(TPF_STPOC_OFF, stpoc_value);  // done with modify for now

        numberOfWrites--;    // decrement the number of writes left

        // update running count of writes done
        tpf_stpoc(TPF_STPOC_ON, stpoc_value);   // allow ourselves to modify key 9 storage gotten by tpf_gsysc
        tpf_serialized_update8(1, (tsug_t *)&my_group->group_stats.writeCount);
        tpf_stpoc(TPF_STPOC_OFF, stpoc_value);  // done with modify for now

     } // end if there is at least one write left

     // release the block that was used to read or write to the record

     if ( levtest(decb) ) {
       relcc(decb);
       record_ptr = NULL;
     }

  } // end while there are still reads or writes to do


  // If there was any error, stop all activity against this group

  if (anyError) {  // if any error, stop the group
                   //  and release the core block if it exists
                   //  and unhold if hold was done

    tpf_stpoc(TPF_STPOC_ON, stpoc_value);
    my_group->group_flag = QFNL_GROUP_STAT_STOPPED;
    tpf_stpoc(TPF_STPOC_OFF, stpoc_value);

    if ( levtest(decb) ) {
      relcc(decb);
      record_ptr = NULL;
    }

    if ( held )   // if it was read with hold and
                  //   and not already unheld by the CP,
                  //   we need to unhold it
      unfrc_ext(decb,FIND_DEFEXT);

  } // end if any error

  // release the decb if there is one

  if (decb)
    tpf_decb_release(decb);


  // increment the nuber of ECBs that have exited and decrement the number of ECBs that are currently active, as
  //   this child ECB is about to exit

  tpf_stpoc(TPF_STPOC_ON, stpoc_value);  // allow ourselves to modify key 9 storage gotten by tpf_gsysc
  tpf_serialized_update8(1, (tsug_t *)&my_group->group_stats.ecbExitCount); // increment total count ECBs exited
  tpf_serialized_update8(-1, (tsug_t *)&my_group->group_stats.ecbCount);    // decrement count ECBs active
  tpf_stpoc(TPF_STPOC_OFF, stpoc_value);  // done for now

 } // end if group found

 // exit the child ECB as its work is ended

 if (verbose) printMsg(15,'I',WTOPC_NO_CHAIN,"SWISC ECB Exiting !!!");
 exit(0);

} // end qfnl_pool_child_ecb


void qfnl_fixed_child_ecb (qfnl_child_ecb_parm_list_Ptr incoming_ecb_parm_ptr)
{

 // This function is called by the child ECB to perform FIND and FILE activity on the
 //   group as requested by the parent ECB

 // The child ECB will do the number of reads and writes requested, and then exit.
 //  It is up to the parent to wait and then kick off another batch of children ECBs.
 //
 // NOTE: Upon entry, each child ECB will update a running counter of the number of ECBs that have
 //         been created for the grup.  Upon exit, it will update a running counter of the number of ECBs that have
 //         exited for the group.

 struct swisc_immediate_input swisc_is;
 swisc_is.istream = SWISC_IS_BALANCE;
 swisc_immediate(&swisc_is);

 if (debug) printf("\n !!!!! The child ECB number %d will operate on group %d !!!!\n",
                     incoming_ecb_parm_ptr->ecbNum, incoming_ecb_parm_ptr->group);

 unsigned int random_ordinal;  // will be used to hold random oridinals that will be accessed for FIND/FILE activity


 // Feed the random number generator

 srand(incoming_ecb_parm_ptr->child_seed);

    // srand(time(NULL));

 // Locate the fixed file group in memory.  The parent (ZTEST FFFACt FIXED START ecb) should have
 //   either found it there already or put it there from file.  But if the child does not find the
 //   group in memory any more, it will exit.

 qfnl_fixed_info_Ptr my_group = qfnl_locate_fixed_group (incoming_ecb_parm_ptr->group, FALSE, FALSE);


 if (my_group != NULL ) { // if group found

  tpf_stpoc(TPF_STPOC_ON, stpoc_value);  // allow ourselves to modify key 9 storage gotten by tpf_gsysc
  tpf_serialized_update8(1, (tsug_t *)&my_group->group_stats.ecbCreateCount); // increment total count ECBs created
  tpf_serialized_update8(1, (tsug_t *)&my_group->group_stats.ecbCount);       // increment count ECBs active
  tpf_stpoc(TPF_STPOC_OFF, stpoc_value);  // done for now

  // get number of reads and writes for this child ECB for use in loop

  int numberOfReads = incoming_ecb_parm_ptr->numberOfReads;
  int numberOfWrites = incoming_ecb_parm_ptr->numberOfWrites;
  #define QFNL_LOCKING incoming_ecb_parm_ptr->doLocking

  // the following are used in case there are any error messages or dumps

  char errMsg[100] = "";
  boolean anyError = false;  // assume no errors for now


  // We need the following in order to do i/o:

  TPF_DECB *decb = NULL;
  DECBC_RC decbRC;
  TPF_FA8 fileaddr;
  void * record_ptr = NULL;  // will eventually point to record read or initialized
  enum t_find_decb findtype; // used for find
  enum t_act filetype;       // used for file

  char FARFstring[QFNL_FARF_STRING_SIZE];


  // allocate a DECB

  decb = tpf_decb_create(NULL, &decbRC);

  if (decb == NULL) { // if could not allocate a DECB
    anyError = true;
    sprintf(errMsg,"ALLOCATE DECB FAILED FOR CHILD ECB FOR FIXED GROUP %d", my_group->group_number);
    serrc_op(SERRC_ALL_HEAP_RETURN,0x5555,errMsg, NULL);

    // increment error counter(s)
    tpf_stpoc(TPF_STPOC_ON, stpoc_value);   // allow ourselves to modify key 9 storage gotten by tpf_gsysc
    tpf_serialized_update8(1, (tsug_t *)&my_group->group_stats.errCount);
    tpf_serialized_update8(1, (tsug_t *)&my_group->group_stats.envErr);
    tpf_stpoc(TPF_STPOC_OFF, stpoc_value);  // done with modify for now

  }

  //
  // Main loop:
  //  The child ECB will loop and perform just the number of reads and writes requested.
  //  But if any error occurs, it will output an error message and exit.
  //

  boolean toggleNOREL = false; // used to toggle between FILUC and FILNC/UNFRC
  boolean held = false;  // will set to true if the record ends up getting held
  boolean readDone = false; // will set to true when the record has been read
                            //  (If not read then we got a core block and there is no minimal verify to do)

  while ( ((numberOfReads + numberOfWrites) > 0) && (!anyError) && (my_group->group_flag == QFNL_GROUP_STAT_STARTED) )

  { // While there are still reads or writes to do

     // generate a random ordinal to work on

     random_ordinal = qfnl_random(my_group->begin_ordinal, my_group->end_ordinal);

     if (debug) printMsg(9,'I',WTOPC_NO_CHAIN,"!!!!! child ECB %d hex random ordinal is %x !!!!", incoming_ecb_parm_ptr->ecbNum, random_ordinal);

     // find its file address

     if ( (anyError=qfnl_get_fixed_fileaddr(my_group,
                                            random_ordinal,
                                            decb,
                                            &fileaddr,
                                            FARFstring,
                                            QFNL_FARF_STRING_SIZE,
                                            true)) ) {

        // increment error counter(s)
        tpf_stpoc(TPF_STPOC_ON, stpoc_value);   // allow ourselves to modify key 9 storage gotten by tpf_gsysc
        tpf_serialized_update8(1, (tsug_t *)&my_group->group_stats.errCount);
        tpf_serialized_update8(1, (tsug_t *)&my_group->group_stats.envErr);
        tpf_stpoc(TPF_STPOC_OFF, stpoc_value);  // done with modify for now

        // bail out
        break;

     }

     //
     // If any reads, do a read, but decide whether or not you will do a lock:
     //    If we are not doing locking (LOCK-N was specified on the START)
     //      then don't hold.
     //    If we are doing locking, but there are no more writes to do
     //      then don't hold
     //    Otherwise, (we are locking, and there is still a write to do)
     //       do a hold
     //

     if (numberOfReads >= 1) { // if there is at least one read

       // If we are locking and there are still writes to do, then we must
       //   do a find with hold;
       // else just do a find without hold

       if ( (QFNL_LOCKING) && (numberOfWrites) ) {

           held = TRUE;
           findtype = HOLD_WAIT;

       } else {

           held = FALSE;
           findtype = NOHOLD_WAIT;

       }

       // read the record

       record_ptr = find_record_ext(decb,                               // decb
                                    &fileaddr,                          // file address of record to read
                                    incoming_ecb_parm_ptr->rec_ID,      // use record ID that was read from first record
                                    '\0',                               // bypass RCC
                                    findtype,                           // type of find controls hold or not
                                    FIND_DEFEXT);                       // use default non GDS

       if (record_ptr == NULL) { // if failure finding random ordinal

          anyError = true;
          sprintf(errMsg, "ERROR READING RECORD TYPE %s ORDINAL %x, FA=%s SUD = %2.2X",
                  my_group->rec_type, random_ordinal, FARFstring, decb->idecsud);
          serrc_op(SERRC_ALL_HEAP_RETURN,0x5555,errMsg, NULL);

          // increment error counter(s)
          tpf_stpoc(TPF_STPOC_ON, stpoc_value);   // allow ourselves to modify key 9 storage gotten by tpf_gsysc
          tpf_serialized_update8(1, (tsug_t *)&my_group->group_stats.errCount);
          tpf_serialized_update8(1, (tsug_t *)&my_group->group_stats.ioErr);
          tpf_stpoc(TPF_STPOC_OFF, stpoc_value);  // done with modify for now

          // For certain errors, the CP will automatically unhold the lock so the application should not attempt to
          //   unhold as well, to avoid extraneous CTL-24 dumps...

          if ( (decb->idecsud & CXSGHE) ||
               (decb->idecsud & CXSGAE) )
              held = false;   // no longer being held by my application, as the CP did the unhold for me

          break;  // get out of loop so we can bail out

       // end if failure finding random ordinal
       } else { // read worked


          readDone = TRUE;

          numberOfReads--;      // decrement the number of reads left

          // Also increment the running count of reads done

          tpf_stpoc(TPF_STPOC_ON, stpoc_value);   // allow ourselves to modify key 9 storage gotten by tpf_gsysc
          tpf_serialized_update8(1, (tsug_t *)&my_group->group_stats.readCount);
          tpf_stpoc(TPF_STPOC_OFF, stpoc_value);  // done with modify for now

          // Increment running count of holds when appropriate

          if (held) {
            tpf_stpoc(TPF_STPOC_ON, stpoc_value);   // allow ourselves to modify key 9 storage gotten by tpf_gsysc
            tpf_serialized_update8(1, (tsug_t *)&my_group->group_stats.holdCount);
            tpf_stpoc(TPF_STPOC_OFF, stpoc_value);  // done with modify for now
          }

       } // end else read worked


     // end if there is at least one read

     } else if (numberOfWrites >= 1) { // if there are no reads left but at least one write,
                                       //    we must get and initialize a core block to write
                                       //    to the record

       record_ptr = getcc(decb,
                          (enum t_getfmt) (GETCC_SIZE+GETCC_FILL),
                          my_group->record_size,
                          0x00);

       if (record_ptr == NULL) {  // if error getting storage

          sprintf(errMsg, "ERROR GETTING STORAGE TO FILE RECORD TYPE %s ORDINAL %x, FA=%s",
                  my_group->rec_type, random_ordinal, FARFstring);
          serrc_op(SERRC_ALL_HEAP_RETURN,0x5555,errMsg, NULL);
          anyError = true;

          // increment error counter(s)
          tpf_stpoc(TPF_STPOC_ON, stpoc_value);   // allow ourselves to modify key 9 storage gotten by tpf_gsysc
          tpf_serialized_update8(1, (tsug_t *)&my_group->group_stats.errCount);
          tpf_serialized_update8(1, (tsug_t *)&my_group->group_stats.mallocErr);
          tpf_stpoc(TPF_STPOC_OFF, stpoc_value);  // done with modify for now

          break;

       }

       readDone = FALSE;

     } // end if there are no reads left but at least one write


     if (readDone) { // if the record was read, perform minimial verification of its contents

       if ( !qfnl_is_recordOK(record_ptr, fileaddr, my_group) ) { // if record read is NOT OK

          sprintf(errMsg, "QUESTIONABLE RECORD CONTENTS RECORD TYPE %s ORDINAL %x, FA=%s",
                  my_group->rec_type, random_ordinal, FARFstring);
          serrc_op(SERRC_ALL_HEAP_RETURN,0x5555,errMsg, NULL);
          anyError = true;

          // increment error counter(s)
          tpf_stpoc(TPF_STPOC_ON, stpoc_value);   // allow ourselves to modify key 9 storage gotten by tpf_gsysc
          tpf_serialized_update8(1, (tsug_t *)&my_group->group_stats.errCount);
          tpf_serialized_update8(1, (tsug_t *)&my_group->group_stats.contentErr);
          tpf_stpoc(TPF_STPOC_OFF, stpoc_value);  // done with modify for now

          break;

       } // if record read is NOT OK

     } // end if the record was read


     if (numberOfWrites >= 1) { // if there is at least one write left

        // Before we can write the record out to the database, we must modify its contents

        if (qfnl_modify_fixed_record(record_ptr,
                                     fileaddr,
                                     my_group) == 0) {  // if modify fails quit now
          sprintf(errMsg, "ERROR MODIFYING RECORD TYPE %s ORDINAL %x, FA=%s",
                  my_group->rec_type, random_ordinal, FARFstring);
          serrc_op(SERRC_ALL_HEAP_RETURN,0x5555,errMsg, NULL);
          anyError = true;

          // increment error counter(s)
          tpf_stpoc(TPF_STPOC_ON, stpoc_value);   // allow ourselves to modify key 9 storage gotten by tpf_gsysc
          tpf_serialized_update8(1, (tsug_t *)&my_group->group_stats.errCount);
          tpf_serialized_update8(1, (tsug_t *)&my_group->group_stats.envErr);
          tpf_stpoc(TPF_STPOC_OFF, stpoc_value);  // done with modify for now

          break;
        }


        if (held) {

        // if the record was held, we have to unhold it, but toggle how this is done

           if (toggleNOREL) {

             filetype = NOREL;     // We will file with NOREL for this iteration
             toggleNOREL = FALSE;  // Next time we will use UNHOLD

           } else {

             filetype = UNHOLD;    // We will file with UNNHOLD for this iteration
             toggleNOREL = TRUE;   // Next time we will use NOREL

           }

        } else {  // else the record was not held, so we do not unhold it

           filetype = NOREL;       // we will file with NOREL
        }

        // file the record with or without unhold as was decided

        file_record_ext(decb,                               // decb
                        &fileaddr,                          // file address of record to read
                        incoming_ecb_parm_ptr->rec_ID,      // use record ID that was read from first record
                        '\0',                               // bypass RCC
                        filetype,                           // type of file controls unhold or not
                        FILE_DEFEXT);                       // use default non GDS

        // For file with NOREL, we must wait for completion and decide whether or not to unhold
        //   the record

        if (filetype == NOREL) {

          if ( waitc() ) {
            anyError = true;
            sprintf(errMsg, "ERROR FILING RECORD TYPE %s ORDINAL %x, FA=%s SUD = %2.2X",
                    my_group->rec_type, random_ordinal, FARFstring, decb->idecsud);
            serrc_op(SERRC_ALL_HEAP_RETURN,0x5555,errMsg, NULL);

          // increment error counter(s)
            tpf_stpoc(TPF_STPOC_ON, stpoc_value);   // allow ourselves to modify key 9 storage gotten by tpf_gsysc
            tpf_serialized_update8(1, (tsug_t *)&my_group->group_stats.errCount);
            tpf_serialized_update8(1, (tsug_t *)&my_group->group_stats.ioErr);
            tpf_stpoc(TPF_STPOC_OFF, stpoc_value);  // done with modify for now

            break;

          }

          if ( held )   // if it was read with hold need to undhold it
            unfrc_ext(decb,FIND_DEFEXT);

        }

        numberOfWrites--;    // decrement the number of writes left

        // update running count of writes done
        tpf_stpoc(TPF_STPOC_ON, stpoc_value);   // allow ourselves to modify key 9 storage gotten by tpf_gsysc
        tpf_serialized_update8(1, (tsug_t *)&my_group->group_stats.writeCount);
        tpf_stpoc(TPF_STPOC_OFF, stpoc_value);  // done with modify for now

     } // end if there is at least one write left

     // release the block that was used to read or write to the record

     if ( levtest(decb) ) {
       relcc(decb);
       record_ptr = NULL;
     }

  } // end while there are still reads or writes to do


  // If there was any error, stop all activity against this group

  if (anyError) {  // if any error, stop the group
                   //  and release the core block if it exists
                   //  and unhold if hold was done

    tpf_stpoc(TPF_STPOC_ON, stpoc_value);
    my_group->group_flag = QFNL_GROUP_STAT_STOPPED;
    tpf_stpoc(TPF_STPOC_OFF, stpoc_value);

    if ( levtest(decb) ) {
      relcc(decb);
      record_ptr = NULL;
    }

    if ( held )   // if it was read with hold and
                  //   and not already unheld by the CP,
                  //   we need to unhold it
      unfrc_ext(decb,FIND_DEFEXT);

  } // end if any error

  // release the decb if there is one

  if (decb)
    tpf_decb_release(decb);


  // increment the nuber of ECBs that have exited and decrement the number of ECBs that are currently active, as
  //   this child ECB is about to exit

  tpf_stpoc(TPF_STPOC_ON, stpoc_value);  // allow ourselves to modify key 9 storage gotten by tpf_gsysc
  tpf_serialized_update8(1, (tsug_t *)&my_group->group_stats.ecbExitCount); // increment total count ECBs exited
  tpf_serialized_update8(-1, (tsug_t *)&my_group->group_stats.ecbCount);    // decrement count ECBs active
  tpf_stpoc(TPF_STPOC_OFF, stpoc_value);  // done for now

 } // end if group found

 // exit the child ECB as its work is ended

 if (verbose) printMsg(15,'I',WTOPC_NO_CHAIN,"SWISC ECB Exiting !!!");
 exit(0);

} // end qfnl_fixed_child_ecb

boolean qfnl_get_fixed_fileaddr(qfnl_fixed_info_Ptr group_info_ptr,
                                unsigned int current_ordinal,
                                TPF_DECB *decb,
                                TPF_FA8 *fileaddr_ptr,
                                char * FARFstring,
                                int FARFstring_size,
                                boolean doDump)
{

  boolean anyError = false;  // assume no error unless it happens
  char errMsg[100] = ""; // use to generate error messages

/**************************************************************************************************************/
// NOTE:
// This function will determine the file address for a fixed file ordinal by calling FACS with the
//    ordinal number and information taken from the fixed file group.  It will store the file address
//    in the decb and also convert the file address to a string
// This function will return FALSE if there was no error obtaining the file address;
//    Otherwise it returns TRUE if there was an error.
/**************************************************************************************************************/

  TPF_FAC8 fac;

  fac.ifacord = current_ordinal;
  memcpy(fac.ifacrec, "        ", sizeof(fac.ifacrec));
  memcpy(fac.ifacrec, group_info_ptr->rec_type, strlen(group_info_ptr->rec_type));
  fac.ifactyp = IFAC8FCS;
  tpf_fac8c(&fac);

  if (fac.ifacret != TPF_FAC8_NRM) { // FACS_ERROR

    anyError = true;           // flag error
    // put out an appropriate error message depending on the nature of the FACS error
    switch (fac.ifacret)
    {
      case TPF_FAC8_NIU:
        sprintf(errMsg,"FACS ERROR PROCESSING ORDINAL %x, RECORD TYPE %s : NOT IN USE",
                fac.ifacord, group_info_ptr->rec_type);
        break;
      case TPF_FAC8_RTH:
        sprintf(errMsg,"FACS ERROR PROCESSING ORDINAL %x, RECORD TYPE %s : RECORD TYPE NONEXISTENT OR EXCEEDS LIMIT",
                fac.ifacord, group_info_ptr->rec_type);

        break;
      case TPF_FAC8_ROR:
        sprintf(errMsg,"FACS ERROR PROCESSING ORDINAL %x, RECORD TYPE %s : ORDINAL OUT OF RANGE",
                current_ordinal, group_info_ptr->rec_type);
        break;
      case TPF_FAC8_NSP:
        sprintf(errMsg,"FACS ERROR PROCESSING ORDINAL %x, RECORD TYPE %s : NO SPLIT CHAIN FOR RECORD",
                current_ordinal, group_info_ptr->rec_type);
        break;
      case TPF_FAC8_POR:
        sprintf(errMsg,"FACS ERROR PROCESSING ORDINAL %x, RECORD TYPE %s : INPUT PARAMETER NOT VALID",
                current_ordinal, group_info_ptr->rec_type);
        break;
      default:
        sprintf(errMsg,"FACS ERROR PROCESSING ORDINAL %x, RECORD TYPE %s : UNKNOWN ERROR",
                current_ordinal, group_info_ptr->rec_type);
        break;
    }  // end switch

    if (doDump)
        serrc_op(SERRC_ALL_HEAP_RETURN,0x5555,errMsg, NULL);
    else
        printMsg(28, 'E', WTOPC_NO_CHAIN, errMsg);


 // end FACS_ERROR

 } else { // file address retrieved from FACS

    decb->idecfa = fac.ifacadr;  // get file address retrieved from FACS
    *fileaddr_ptr = fac.ifacadr;      // ditto

    qfnl_FARF6_to_hex_string(FARFstring, FARFstring_size, *fileaddr_ptr);

 } // end else file address retrieved from FACS

 return(anyError);  // let caller know if there was an error or not

}

boolean qfnl_get_pool_fileaddr(qfnl_pool_info_Ptr group_info_ptr,
                               unsigned int index,
                               TPF_DECB *decb,
                               TPF_FA8 *fileaddr_ptr,
                               char * FARFstring,
                               int FARFstring_size,
                               boolean doDump)
{ // beginning qfnl_get_pool_fileaddr

  boolean anyError = false;  // assume no error unless it happens
  char errMsg[100] = ""; // use to generate error messages


  //  Get the pool file address for the index specified from the slot in the table.
  //     But first check if the index given is greater than the number of pools in the group.  If that is the case, return an error

  if (index > group_info_ptr->number_records) { // if index exceeds number of records

     anyError = true;    // flag error

     // put out informative error message

     sprintf(errMsg,"INDEX OF POOL REQUESTED, %d, IS GREATER THAN NUMBER OF POOLS, %d, IN POOL GROUP %d",
                     index, group_info_ptr->number_records, group_info_ptr->group_number);

     if (doDump)
         serrc_op(SERRC_ALL_HEAP_RETURN,0x5555,errMsg, NULL);
     else
         printMsg(91, 'E', WTOPC_NO_CHAIN, errMsg);

  // end if index exceeds number of records
  } else if (index < 1) { // if the index is 0 or less, no good

     anyError = true;    // flag error

     // put out informative error message

     sprintf(errMsg,"INDEX OF POOL REQUESTED, %d, FOR POOL GROUP %d CANNOT BE LESS THAN 1",
                     index, group_info_ptr->group_number);

     if (doDump)
         serrc_op(SERRC_ALL_HEAP_RETURN,0x5555,errMsg, NULL);
     else
         printMsg(91, 'E', WTOPC_NO_CHAIN, errMsg);

  // end if index is 0 or less, no good

  } else { // The index is within the range allowed for this group

    qfnl_pool_batch_Ptr pool_table = group_info_ptr->pool_batch_ptr;  // point at table of pools for this group

    if (pool_table == NULL ) { // This is unexpected, but if the pool table isn't available then bail out!!!
        anyError = true;  // flag error
        sprintf(errMsg,"PROCESSING REQUEST FOR INDEX=%d, TABLE OF POOLS NOT AVAILABLE FOR POOL GROUP %d",
                        index, group_info_ptr->group_number);
        if (doDump)
            serrc_op(SERRC_ALL_HEAP_RETURN,0x5555,errMsg, NULL);
        else
            printMsg(91, 'E', WTOPC_NO_CHAIN, errMsg);
    } else {  // pool table not NULL

       char * index_item = &(pool_table->first_item);  // point at first item in table

       // convert the input index to a zero-based index
       int zero_based = index;
       zero_based--;

       index_item = index_item + zero_based * sizeof(qfnl_pool_item); // point at item requested

       qfnl_pool_item_Ptr pool_item = (qfnl_pool_item_Ptr) index_item;  // access requested item as item in pool group

       // Now that we are pointing at the element we want, get its file address

       // ... but first check that it is still valid (i.e., has not expired) ...

       if (pool_item->flags == QFNL_POOL_ITEM_NOT_VALID) { // if item has expired

         TPF_FA8 fileaddr;  // used to hold pool file address when gotten
         char FARFstring[QFNL_FARF_STRING_SIZE];  // used for debugging purposes

         void * record_ptr = NULL;  // pointer for record read (and later filed)

         tpf_TOD_ext_type tod;  // used to store TOD which will be copied to the pool item

         corhc(pool_item);      // lock item in case we need to update it

         if (pool_item->flags == QFNL_POOL_ITEM_NOT_VALID) { // only get new one if still not valid

         /* ---------------- Beginning of Logic to Get and Initialize a New Pool Record when Current One Expired --------------- */
           char *state_ptr = (char*)cinfc_fast_ss(CINFC_CMMOPS, ecbptr()->ce1dbi);
           if (*state_ptr & GFS_INACTIVE) { // if GFS not active, forget about getting another pool and expiry, just use this one

             tpf_stpoc(TPF_STPOC_ON, stpoc_value);  // allow modify key 9
             pool_item->flags = QFNL_POOL_ITEM_VALID; // just pretend this one is valid so we don't keep getting CTL-E2's
             tpf_stpoc(TPF_STPOC_OFF, stpoc_value);  // done with modify

           } else { // get a new pool because this one expired and GFS is active

             fileaddr = getfc(decb,
                              GETFC_TYPE1,
                              group_info_ptr->handle.handle_id,
                              (int)(GETFC_BLOCK+GETFC_FILL),
                               GETFC_NOSERRC);

             if (!fileaddr) { // if GETFC failed

                anyError = true;
                sprintf(errMsg,"GETFC FOR %dTH POOL IN POOL GROUP %d FAILED AFTER IT EXPIRED",
                               index, group_info_ptr->group_number);
                printMsg(101, 'E', WTOPC_NO_CHAIN, errMsg);

               // end if GETFC failed
             } else { // succesfully got pool record

               record_ptr = decb->idecdad;  // address core block obtained

               int convertOK = qfnl_FARF6_to_hex_string(FARFstring, sizeof(FARFstring), fileaddr);
               if ( convertOK == 1 ) {  // converted ok
                  if (debug) printf("\n!!! Hooray... The %dTH pool in group %d replaced with file address %s !!\n",
                             index, group_info_ptr->group_number, FARFstring);
               }

               tpf_stpoc(TPF_STPOC_ON, stpoc_value);  // allow modify key 9
               pool_item->FA = fileaddr;  // copy pool file address to table
               tpf_stpoc(TPF_STPOC_OFF, stpoc_value);  // done modify for now

               // Leave the item flagged as invalid until we can successfully initialize it and file it out
               if (qfnl_modify_pool_record(record_ptr,
                                           fileaddr,
                                           group_info_ptr,
                                           &tod) == 0) {  // if modify fails quit now
                   anyError = true;
               } else { // otherwise copy the time that is stored in the record as time the pool was added
                   tpf_stpoc(TPF_STPOC_ON, stpoc_value); // allow modify again
                   memcpy(&pool_item->tod_added, &tod, sizeof(tod));
                   tpf_stpoc(TPF_STPOC_OFF, stpoc_value);  // done modify for now
               }

               if (!anyError) {  // if no failures yet, try to file the newly gotten pool record

                 file_record_ext(decb,
                                 &fileaddr,
                                 group_info_ptr->hex_rid,
                                 '\0',
                                 NOREL,
                                 FILE_DEFEXT);

                 if ( waitc() ) {  // if error on wait
                    anyError = true;
                    if (convertOK) {
                      sprintf(errMsg, "ERROR FILING NEW POOL FILE ADDRESS %s ON EXPIRATION OF OLD, SUD = %2.2X",
                                       FARFstring,  decb->idecsud);
                    } else {
                      sprintf(errMsg, "ERROR FILING NEW POOL FOR %dTH ITEM IN GROUP %d ON EXPIRE OF OLD, SUD = %2.2X",
                                       index, group_info_ptr->group_number, decb->idecsud);
                    }
                    printMsg(102, 'E', WTOPC_NO_CHAIN, errMsg);
                 } else { // the new pool record was filed out ok

                    tpf_stpoc(TPF_STPOC_ON, stpoc_value); // allow modify again

                    // update running count of writes done
                    tpf_serialized_update8(1, (tsug_t *)&group_info_ptr->group_stats.writeCount);

                    memcpy(&pool_item->tod_updated, &tod, sizeof(tod));  // store the time this item was last updated

                    // Flag the item as valid
                    pool_item->flags = QFNL_POOL_ITEM_VALID;

                    tpf_stpoc(TPF_STPOC_OFF, stpoc_value);  // done with modify

                 } // end else the new pool record was filed out ok

               } // end if no failures yet, try to file the record

             } // end else succesfully got pool record

           } // END else get a new pool because this one expired and GFS is active
         /* ----------------       END of Logic to Get and Initialize a New Pool Record when Current One Expired --------------- */

         } // end only get new one if still not valid

         coruc(pool_item);      // unlock item as we are done with it

         if (levtest(decb))     // release any core block used to replace expired pool record
            relcc(decb);

       } // end if item has expired

       decb->idecfa = pool_item->FA;   // get file address for this item
       *fileaddr_ptr = pool_item->FA;  // ditto

       qfnl_FARF6_to_hex_string(FARFstring, FARFstring_size, *fileaddr_ptr);

       if (debug) {
         printf("\n!!! GET POOL FILE ADDRESS CALLED FOR INDEX=%d, FA FOUND= %.16lX!!!\n",
             index, pool_item->FA);
         printf("\n!!! ADDRESS OF ITEM IS  %.16lX!!!\n", pool_item);
       }


    } // end else pool table not NULL

  } // end else The index is within the range allowed for this group

  return(anyError);  // let caller know if there was an error or not

} // end qfnl_get_pool_fileaddr

void qfnl_pool_set_expired(qfnl_pool_info_Ptr group_info_ptr)
{ // begin qfnl_pool_set_expired

    // Lock the pool group in case someone is trying to KILL it from under you

    corhc(group_info_ptr);

    qfnl_pool_batch_Ptr pool_table = group_info_ptr->pool_batch_ptr;  // point at table of pools for this group

    // This is unexpected, but if there isn't a table of pools any more then just bail out

    if (pool_table == NULL) {
       coruc(group_info_ptr);   // unlock the pool group
       return;                  // bail out
    }

    qfnl_pool_item_Ptr pool_item = (qfnl_pool_item_Ptr) &(group_info_ptr->pool_batch_ptr->first_item);  // point at first slot in pool table

    // Loop thru all items in the table, temporarily doing a corhc on each one, and mark them as no longer valid

    for (unsigned int pool_count = 1;
        pool_count <= group_info_ptr->number_records;
        pool_count++, pool_item++) { // for every pool in the group...

           // temporarily lock the item before potentially updating it
           corhc(pool_item);

           tpf_stpoc(TPF_STPOC_ON, stpoc_value);  // allow modify key 9

           // if the pool is valid, set it as invalid
           if (pool_item->flags == QFNL_POOL_ITEM_VALID)
              pool_item->flags = QFNL_POOL_ITEM_NOT_VALID;

           tpf_stpoc(TPF_STPOC_OFF, stpoc_value);  // done with modify for now

           if (debug) printMsg(1,'I',WTOPC_NO_CHAIN,"INVALIDATED %dTH ITEM OF POOL GROUP %d",
                               pool_count, group_info_ptr->group_number);

           // unlock the item now that it has been updated
           coruc(pool_item);

    } // end for every pool in the group...

    coruc(group_info_ptr);  // done updating this group so unlock it

    return;                 // return to caller

} // end qfnl_pool_set_expired

/*********************************************************************/
/*                                                                   */
/* Function:                                                         */
/*                                                                   */
/* Name:            convert_ebcdic_to_hex()                          */
/*                                                                   */
/* Abstract:        Convert string of EBCDIC data to hex             */
/*                                                                   */
/* Description:     Take a string of EBCDIC data and translate it    */
/*                  into a string of hexadecimal data.               */
/*                                                                   */
/*-------------------------------------------------------------------*/
/*                                                                   */
/* Interface Requirements:                                           */
/*                                                                   */
/* Input:                                                            */
/*                                                                   */
/*  Parameters:    input_data  :  Pointer to string of EBCDIC data   */
/*                 output_data :  Pointer to where string of hex data*/
/*                                will be put.                       */
/*                                                                   */
/* Returns/Exits:                                                    */
/*                                                                   */
/*  Function return value: Length of string contains hexadecimal data*/
/*                         If length is zero, data was not converted.*/
/*                         Converted string will be put in into      */
/*                         location specified by output_data.        */
/*                                                                   */
/*********************************************************************/

unsigned int convert_input_data_to_hex(char *input_data,
                          char *output_data)

{
   char              tmp_buffer[3];
   unsigned int      i;               /* index                      */
   unsigned int      j;               /* index                      */
   unsigned int      input_len;       /* input length               */
   unsigned int      output_len;      /* output length - return code*/

    /* verify that newdata variable must be an even  number of     */
    /* hexadecimal digits.                                         */

   input_len = strlen(input_data);

   if ((input_len%2) > 0)             /* check remainder > 0        */
      output_len = 0;
   else  /* conversion for newdata from string to characters */
   {
     for (i=0, j=0 ; i<input_len; i=i+2, j++)
     {
        tmp_buffer[0] = *((input_data)+i);
        tmp_buffer[1] = *((input_data)+i+1);
        tmp_buffer[2] = '\0';
        output_data[j] =
            (char) strtoul(tmp_buffer, (char**) NULL, HEX_BASE);
     } /* end for                     */

     output_data[j] = '\0';
     output_len = j;
    }  /* end conversion for newdata  */

   return output_len;
}

/**********************************************************************/
/* Start the CANNED case background execution.                        */
/**********************************************************************/
static void cannedStart(int force)
{
  char ctrltok[9];
  long size;
  unsigned int wait = 0;
  int istream = ecbptr()->ce1isn;
  drvr_cntl *ctrl_blk = NULL;
  drvr_ecbinfo ecbinfo;

  /********************************************************************/
  /* Get/Initialize the Control Block                                 */
  /********************************************************************/
  getSysHeapToken(ctrltok);
  while (ctrl_blk == NULL)
  {
    /******************************************************************/
    /* Get the control block                                          */
    /******************************************************************/
    ctrl_blk = (drvr_cntl *) tpf_fsysc(ctrltok, &size);

    /******************************************************************/
    /* Control block does not exist yet. Create one.                  */
    /******************************************************************/
    if (ctrl_blk == NULL)
    {
      /* No control block, get storage */
      ctrl_blk = (drvr_cntl *) tpf_gsysc(1, ctrltok, NULL,
                                         GSYSC_UNIQUE + GSYSC_64BIT);
      if (ctrl_blk == NULL)
      {
        if (errno == ETPFGSYS_NUNQTKN)
        {
          /* Storage with the token already exists; someone must have */
          /* allocated one just now. Start over and get the block.    */
          continue;
        }
        else
        {
          /* Error we can't recover from... */
          printErr(105, "ERROR GETTING CONTROL BLOCK");
          return;
        }
      }

      tpf_stpoc(TPF_STPOC_ON, NULL);
      lockc(&(ctrl_blk->lock), LOCK_O_SPIN);

      /* Initialize the control block */
      memcpy(&(ctrl_blk->hdr), DRVR_HDR, 4);
      ctrl_blk->instances = 0;
      memset(&(ctrl_blk->is), 0, sizeof(ctrl_blk->is));

      /* don't unlock as we're doing more after the while loop */
      break;
    } /* if (ctrl_blk == NULL) */

    /******************************************************************/
    /* Existing control block found. Use it.                          */
    /******************************************************************/
    else
    {
      /* We got some storage, make sure it's valid */
      tpf_stpoc(TPF_STPOC_ON, NULL);
      lockc(&(ctrl_blk->lock), LOCK_O_SPIN);

      /* We have a storage address, make sure it's valid */
      if (memcmp(&(ctrl_blk->hdr), DRVR_HDR, 4) != 0)
      {
        unlkc(&(ctrl_blk->lock));
        tpf_stpoc(TPF_STPOC_OFF, NULL);

        /* If we've waited a while, just quit */
        if (wait++ > DEFER_MAX)
          return;

        /* Header not initialized yet, wait for it to finish */
        sleep(1);
        defrc();

        ctrl_blk = NULL;
        continue;
      }

      /****************************************************************/
      /* Valid, check for STOP condition */
      /****************************************************************/
      if (ctrl_blk->is[istream - 1].status == STOP)
      {
        unlkc(&(ctrl_blk->lock));
        tpf_stpoc(TPF_STPOC_OFF, NULL);

        printErr(107, "UNABLE TO START CANNED TEST CASES - STOPPING ON ISTREAM %d",
                 istream);
        return;
      }

      /* don't unlock as we're doing more after the while loop */
      break;
    }
  } /* while (ctrl_blk == NULL) */

  /********************************************************************/
  /* Add an execution instance                                        */
  /********************************************************************/
  ctrl_blk->instances++;
  ctrl_blk->is[istream - 1].status = START;
  ctrl_blk->is[istream - 1].instances++;

  unlkc(&(ctrl_blk->lock));
  tpf_stpoc(TPF_STPOC_OFF, NULL);

  /********************************************************************/
  /* Instantiate a new ECB for CANNED test case mode processing       */
  /********************************************************************/
  printMsg(14, 'I', WTOPC_NO_CHAIN, "STARTING CANNED CASES ON ISTREAM %d",
           istream);
  printMsg(8,'W',WTOPC_NO_CHAIN,"CANNED CASES JUST A PLACE HOLDER: THEY DO NOTHING");

  memcpy(&ecbinfo.hdr, "canned", 6);
  ecbinfo.flag = force;
  ecbinfo.casenum = 0;
  ecbinfo.ecbroute = ecbptr()->ebrout;
  cremc(sizeof(drvr_ecbinfo), &ecbinfo, QFNL);
}

/**********************************************************************/
/* Run CANNED test cases                                              */
/**********************************************************************/
static void cannedRun()
{
  int rc;
  int done = 0;
  unsigned int wait = 0;
  char ctrltok[9];
  long size;
  int istream = ecbptr()->ce1isn;
  drvr_cntl *ctrl_blk = NULL;
  drvr_ecbinfo *ecbinfo = (drvr_ecbinfo *) &ecbptr()->ebw000;
  ecbptr()->ebrout = ecbinfo->ecbroute;

  /********************************************************************/
  /* Determine control token name                                     */
  /********************************************************************/
  getSysHeapToken(ctrltok);

  /********************************************************************/
  /* Force Continous Mode to only output error messages.              */
  /********************************************************************/
  verbose = 0;
  warnings = 0;
  repeat_canned = 1;

  /********************************************************************/
  /* Process test cases continually                                   */
  /********************************************************************/
  while (!done)
  {
    /******************************************************************/
    /* Driver repeated CANNED case LOGIC                              */
    /******************************************************************/
    // TODO: enter your repeated "canned" case logic here.
    //       In the sample, we run all test cases via the 'runApi'
    //       function, incrementing each pass. When an "DRVR_INVALID"
    //       return is met, we start over at case #0.
    rc = runApi(ecbinfo->casenum++);
    if (rc == DRVR_INVALID)
    {
      ecbinfo->casenum = 0;
    }

    /******************************************************************/
    /* Get the control block                                          */
    /******************************************************************/
    ctrl_blk = (drvr_cntl *) tpf_fsysc(ctrltok, &size);
    if (ctrl_blk == NULL)
    {
      /* No control block, quit */
      return;
    }
    tpf_stpoc(TPF_STPOC_ON, NULL);
    lockc(&(ctrl_blk->lock), LOCK_O_SPIN);

    /******************************************************************/
    /* Check for initialization by someone else                       */
    /******************************************************************/
    while (memcmp(&(ctrl_blk->hdr), DRVR_HDR, 4) != 0)
    {
      unlkc(&(ctrl_blk->lock));
      tpf_stpoc(TPF_STPOC_OFF, NULL);

      /* if we've waited for a long time, give up */
      if (wait++ > DEFER_MAX)
        return;

      sleep(1);
      defrc();

      /* re-get the control block, as it could possibly have
         been eliminated in the time-being */
      ctrl_blk = (drvr_cntl *) tpf_fsysc(ctrltok, &size);
      if (ctrl_blk == NULL)
        return;

      tpf_stpoc(TPF_STPOC_ON, NULL);
      lockc(&(ctrl_blk->lock), LOCK_O_SPIN);
    }

    /******************************************************************/
    /* Process CANNED case status information                         */
    /******************************************************************/
    if (rc != DRVR_INVALID)
    {
      ctrl_blk->is[istream - 1].tests_run++;
    }
    if (rc == DRVR_FAILURE)
    {
      ctrl_blk->is[istream - 1].tests_fail++;
    }

    /******************************************************************/
    /* Check control block for stop indicator                         */
    /******************************************************************/
    if (ctrl_blk->is[istream - 1].status == STOP)
    {
      ctrl_blk->is[istream - 1].instances--;
      ctrl_blk->instances--;

      unlkc(&(ctrl_blk->lock));
      tpf_stpoc(TPF_STPOC_OFF, NULL);

      return;
    }

    /******************************************************************/
    /* If not forced mode, we want to spawn an ECB for each test case */
    /******************************************************************/
    if (ecbinfo->flag != CANNED_FORCE)
    {
      /* done = 1 will quit while loop */
      done = 1;
    }

    unlkc(&(ctrl_blk->lock));
    tpf_stpoc(TPF_STPOC_OFF, NULL);

    /******************************************************************/
    /* Pause for a moment before running the next test case           */
    /******************************************************************/
    sleep(CANNED_DELAY);

  } /* while (!done) */

  /********************************************************************/
  /* Now start a new ECB to re-run tests                              */
  /********************************************************************/
  cremc(sizeof(drvr_ecbinfo), ecbinfo, QFNL);
}

/**********************************************************************/
/* Stop the CANNED case background execution.                         */
/**********************************************************************/
static void cannedStop(int force)
{
  char ctrltok[9];
  long size;
  unsigned int wait = 0;
  int istream = ecbptr()->ce1isn;
  drvr_cntl *ctrl_blk = NULL;

  /********************************************************************/
  /* Get the Control Block                                            */
  /********************************************************************/
  getSysHeapToken(ctrltok);
  ctrl_blk = (drvr_cntl *) tpf_fsysc(ctrltok, &size);
  if (ctrl_blk == NULL)
  {
    /* No control block, can't stop when not started */
    printErr(106, "CANNED TEST CASES NOT RUNNING");
    return;
  }
  tpf_stpoc(TPF_STPOC_ON, NULL);
  lockc(&(ctrl_blk->lock), LOCK_O_SPIN);

  /********************************************************************/
  /* Check for initialization by someone else                         */
  /********************************************************************/
  while (memcmp(&(ctrl_blk->hdr), DRVR_HDR, 4) != 0)
  {
    unlkc(&(ctrl_blk->lock));
    tpf_stpoc(TPF_STPOC_OFF, NULL);

    /* if we've waited for a long time, give up */
    if (wait++ > DEFER_MAX)
      return;

    sleep(1);
    defrc();

    /* re-get the control block, as it could possibly have
       been eliminated in the time-being */
    ctrl_blk = (drvr_cntl *) tpf_fsysc(ctrltok, &size);
    if (ctrl_blk == NULL)
      return;

    tpf_stpoc(TPF_STPOC_ON, NULL);
    lockc(&(ctrl_blk->lock), LOCK_O_SPIN);
  }

  /********************************************************************/
  /* Check if we can stop the driver                                  */
  /********************************************************************/
  if (ctrl_blk->is[istream - 1].status == INIT)
  {
    unlkc(&(ctrl_blk->lock));
    tpf_stpoc(TPF_STPOC_OFF, NULL);

    printErr(108, "CANNED TEST CASES NOT RUNNING ON ISTREAM %d", istream);
    return;
  }
  else if (force == 0 && ctrl_blk->is[istream - 1].status == STOP)
  {
    unlkc(&(ctrl_blk->lock));
    tpf_stpoc(TPF_STPOC_OFF, NULL);

    printErr(109, "UNABLE TO STOP CANNED CASES - ALREADY STOPPING ON ISTREAM %d",
             istream);
    return;
  }

  /********************************************************************/
  /* Stop the driver                                                  */
  /********************************************************************/
  ctrl_blk->is[istream - 1].status = STOP;

  printMsg(4, 'I', WTOPC_NO_CHAIN,
           "STOPPING ALL CANNED CASE INSTANCES ON ISTREAM %d", istream);

  /********************************************************************/
  /* Wait until all instances are stopped                             */
  /********************************************************************/
  while (force == 0 && ctrl_blk->is[istream - 1].instances > 0)
  {
    unlkc(&(ctrl_blk->lock));
    tpf_stpoc(TPF_STPOC_OFF, NULL);

    sleep(1);
    defrc();

    /* re-get the control block, as it could possibly have
       been eliminated in the time-being */
    ctrl_blk = (drvr_cntl *) tpf_fsysc(ctrltok, &size);
    if (ctrl_blk == NULL)
      return;

    /* Lock first thing */
    tpf_stpoc(TPF_STPOC_ON, NULL);
    lockc(&(ctrl_blk->lock), LOCK_O_SPIN);

    /* Check for a valid expected control block */
    if (memcmp(&(ctrl_blk->hdr), DRVR_HDR, 4) != 0 ||
        ctrl_blk->is[istream - 1].status != STOP)
    {
      unlkc(&(ctrl_blk->lock));
      tpf_stpoc(TPF_STPOC_OFF, NULL);

      return;
    }
  }

  /********************************************************************/
  /* Output some data                                                 */
  /********************************************************************/
  printMsg(6, 'I', WTOPC_NO_CHAIN,
           "COMPLETED %d TESTS (%d FAILED) ON ISTREAM %d",
           ctrl_blk->is[istream - 1].tests_run,
           ctrl_blk->is[istream - 1].tests_fail, istream);
  printMsg(8,'W',WTOPC_NO_CHAIN,"CANNED CASES JUST A PLACE HOLDER: THEY DO NOTHING");

  /********************************************************************/
  /* Reset the control block                                          */
  /********************************************************************/
  ctrl_blk->is[istream - 1].tests_run = 0;
  ctrl_blk->is[istream - 1].tests_fail = 0;
  ctrl_blk->is[istream - 1].instances = 0;
  ctrl_blk->is[istream - 1].status = INIT;

  /********************************************************************/
  /* If no driver instances on any istream, clean up sysheap          */
  /********************************************************************/
  if (force == 1 || ctrl_blk->instances <= 0)
  {
    memset(&(ctrl_blk->hdr), 0, 4);

    unlkc(&(ctrl_blk->lock));
    tpf_stpoc(TPF_STPOC_OFF, NULL);

    /* Release the storage block */
    tpf_rsysc(ctrl_blk, 1, ctrltok);
  }
  else
  {
    unlkc(&(ctrl_blk->lock));
    tpf_stpoc(TPF_STPOC_OFF, NULL);
  }

  /* Done! */
  printMsg(5, 'I', WTOPC_NO_CHAIN,
           "CANNED CASES STOPPED ON ISTREAM %d", istream);
}

/**********************************************************************/
/* Output the CANNED case background execution status.                */
/**********************************************************************/
static void cannedStatus()
{
  int i, tot_run, tot_fail, tot_inst;
  char *status;
  char ctrltok[9];
  long size;
  unsigned int wait = 0;
  drvr_cntl *ctrl_blk = NULL;

  /********************************************************************/
  /* Get the Control Block                                            */
  /********************************************************************/
  getSysHeapToken(ctrltok);
  ctrl_blk = (drvr_cntl *) tpf_fsysc(ctrltok, &size);
  if (ctrl_blk == NULL)
  {
    /* No control block, can't get status when not started */
    printErr(106, "CANNED TEST CASES NOT RUNNING");
    return;
  }
  tpf_stpoc(TPF_STPOC_ON, NULL);
  lockc(&(ctrl_blk->lock), LOCK_O_SPIN);

  /********************************************************************/
  /* Wait for initialization to complete                              */
  /********************************************************************/
  while (memcmp(&(ctrl_blk->hdr), DRVR_HDR, 4) != 0)
  {
    unlkc(&(ctrl_blk->lock));
    tpf_stpoc(TPF_STPOC_OFF, NULL);

    /* If we've waited too long, just quit */
    if (wait++ > DEFER_MAX)
    {
      printMsg(7, 0, WTOPC_CHAIN_END, "END OF DISPLAY");
      return;
    }

    sleep(1);
    defrc();

    ctrl_blk = (drvr_cntl *) tpf_fsysc(ctrltok, &size);
    if (ctrl_blk == NULL)
    {
      printMsg(7, 0, WTOPC_CHAIN_END, "END OF DISPLAY");
      return;
    }

    /* Lock first thing */
    tpf_stpoc(TPF_STPOC_ON, NULL);
    lockc(&(ctrl_blk->lock), LOCK_O_SPIN);
  }

  /********************************************************************/
  /* Output status information                                        */
  /********************************************************************/
  printMsg(7, 'I', WTOPC_CHAIN, "QFNL STATUS DISPLAY");
  printMsg(7, 0,  WTOPC_CHAIN,
           "   I-S   STATUS     INST    CASES   FAILED\n"
           "   ---   --------   ----   ------   ------");

  {
    /* Output I-stream information */
    tot_inst = tot_run = tot_fail = 0;
    for (i = 0; i < SAISTRM; i++)
    {
      if (ctrl_blk->is[i].status > INIT)
      {
        switch (ctrl_blk->is[i].status)
        {
        case STOP:
          status = "STOPPING";
          break;
        case START:
          status = "RUNNING";
          break;
        case INIT:
          status = "INACTIVE";
          break;
        default:
          status = "UNKNOWN";
        }

        printMsg(7, 0, WTOPC_CHAIN,
                 "   %3d   %-8s   %4d   %6d   %6d",
                 i + 1, status, ctrl_blk->is[i].instances,
                 ctrl_blk->is[i].tests_run,
                 ctrl_blk->is[i].tests_fail);

        tot_inst += ctrl_blk->is[i].instances;
        tot_run += ctrl_blk->is[i].tests_run;
        tot_fail += ctrl_blk->is[i].tests_fail;
      } /* if ctrl_blk->is[i].status > INIT */

    } /* for (i < SAISTRM) */
  }

  /* Unlock; done with control block */
  unlkc(&(ctrl_blk->lock));
  tpf_stpoc(TPF_STPOC_OFF, NULL);

  printMsg(7, 0, WTOPC_CHAIN,
           "\n            TOTALS: %4d   %6d   %6d",
           tot_inst, tot_run, tot_fail);
  printMsg(7, 0, WTOPC_CHAIN_END, "END OF DISPLAY");
}


/**********************************************************************/
/* Get the system Heap Token                                          */
/**********************************************************************/
char * getSysHeapToken(char *buffer)
{
  struct uatbc_area sua;

  if (buffer == NULL)
    return NULL;

  sua.uat_call = UAT_ESSI;
  uatbc(&sua);
  memcpy(buffer, DRVR_HDR, 4);
  memcpy(buffer + 4, sua.uat_ss_name, 4);
  buffer[8] = '\0';

  return buffer;
}

/**********************************************************************/
/* Output a message                                                   */
/**********************************************************************/
void printMsg(int num, char code, enum t_wtopc_chain chain,
              char *fmt, ...)
{
  va_list arg;
  char buffer[256];

  va_start(arg, fmt);
  vsprintf(buffer, fmt, arg);
  va_end(arg);

  if (code != 0)
  {
    struct wtopc_header msg_hdr;
    wtopc_insert_header(&msg_hdr, DRVR_HDR, num, code, WTOPC_SYS_TIME);
    wtopc(buffer, WTOPC_EBROUT, chain, &msg_hdr);
  }
  else
  {
    wtopc(buffer, WTOPC_EBROUT, chain, WTOPC_NO_HEADER);
  }
}

/**********************************************************************/
/* Output an informational message, taking account the global verbose */
/* flag. Messages are chained based on parameter.                     */
/**********************************************************************/
void printInfo(infochain chain, char *fmt, ...)
{
  struct wtopc_header msg_hdr;
  va_list arg;
  char buffer[256];

  if (!verbose)
  {
    return;
  }

  va_start(arg, fmt);
  vsprintf(buffer, fmt, arg);
  va_end(arg);

  switch (chain)
  {
  case DISPHEAD:
    wtopc_insert_header(&msg_hdr, DRVR_HDR, 1, 'I', WTOPC_SYS_TIME);
    wtopc(buffer, WTOPC_EBROUT, WTOPC_CHAIN, &msg_hdr);
    break;
  case DISPSTOP:
    wtopc(buffer, WTOPC_EBROUT, WTOPC_CHAIN_END, WTOPC_NO_HEADER);
    break;
  case DISPCONT:
  default:
    wtopc(buffer, WTOPC_EBROUT, WTOPC_CHAIN, WTOPC_NO_HEADER);
    break;
  }
}

/**********************************************************************/
/* Output a message based on 'warning' var. This always is displayed. */
/**********************************************************************/
void printWarn(infochain chain, char *fmt, ...)
{
  struct wtopc_header msg_hdr;
  va_list arg;
  char buffer[256];

  if (warnings == 0)
  {
    return;
  }

  va_start(arg, fmt);
  vsprintf(buffer, fmt, arg);
  va_end(arg);

  switch (chain)
  {
  case DISPHEAD:
    wtopc_insert_header(&msg_hdr, DRVR_HDR, 1, 'I', WTOPC_SYS_TIME);
    wtopc(buffer, WTOPC_EBROUT, WTOPC_CHAIN, &msg_hdr);
    break;
  case DISPSTOP:
    wtopc(buffer, WTOPC_EBROUT, WTOPC_CHAIN_END, WTOPC_NO_HEADER);
    break;
  case DISPCONT:
  default:
    wtopc(buffer, WTOPC_EBROUT, WTOPC_CHAIN, WTOPC_NO_HEADER);
    break;
  }
}

/**********************************************************************/
/* Output an error message. This always is displayed.                 */
/**********************************************************************/
void printErr(int num, char *fmt, ...)
{
  va_list arg;
  struct wtopc_header msg_hdr;
  char buffer[256];

  va_start(arg, fmt);
  vsprintf(buffer, fmt, arg);
  va_end(arg);

  wtopc_insert_header(&msg_hdr, DRVR_HDR, num, 'E', WTOPC_SYS_TIME);
  wtopc(buffer, WTOPC_EBROUT, WTOPC_NO_CHAIN, &msg_hdr);
}
